/*
 * Copyright (C) 2010-2019 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#include "StorageManagerSet.h"

#include "ArgumentCoders.h"
#include "Decoder.h"
#include "HandleMessage.h"
#include "StorageAreaIdentifier.h"
#include "StorageAreaImplIdentifier.h"
#include "StorageManagerSetMessages.h"
#include "StorageNamespaceIdentifier.h"
#include "WebCoreArgumentCoders.h"
#include <WebCore/SecurityOriginData.h>
#include <pal/SessionID.h>
#include <wtf/HashMap.h>
#include <wtf/Optional.h>
#include <wtf/text/WTFString.h>

namespace Messages {

namespace StorageManagerSet {

void ConnectToLocalStorageArea::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Optional<WebKit::StorageAreaIdentifier>& storageAreaID)
{
    *encoder << storageAreaID;
    connection.sendSyncReply(WTFMove(encoder));
}

void ConnectToTransientLocalStorageArea::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Optional<WebKit::StorageAreaIdentifier>& storageAreaID)
{
    *encoder << storageAreaID;
    connection.sendSyncReply(WTFMove(encoder));
}

void ConnectToSessionStorageArea::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Optional<WebKit::StorageAreaIdentifier>& storageAreaID)
{
    *encoder << storageAreaID;
    connection.sendSyncReply(WTFMove(encoder));
}

void GetValues::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const HashMap<String, String>& values)
{
    *encoder << values;
    connection.sendSyncReply(WTFMove(encoder));
}

} // namespace StorageManagerSet

} // namespace Messages

namespace WebKit {

void StorageManagerSet::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    auto protectedThis = makeRef(*this);
    if (decoder.messageName() == Messages::StorageManagerSet::DisconnectFromStorageArea::name()) {
        IPC::handleMessage<Messages::StorageManagerSet::DisconnectFromStorageArea>(connection, decoder, this, &StorageManagerSet::disconnectFromStorageArea);
        return;
    }
    if (decoder.messageName() == Messages::StorageManagerSet::CloneSessionStorageNamespace::name()) {
        IPC::handleMessage<Messages::StorageManagerSet::CloneSessionStorageNamespace>(connection, decoder, this, &StorageManagerSet::cloneSessionStorageNamespace);
        return;
    }
    if (decoder.messageName() == Messages::StorageManagerSet::SetItem::name()) {
        IPC::handleMessage<Messages::StorageManagerSet::SetItem>(connection, decoder, this, &StorageManagerSet::setItem);
        return;
    }
    if (decoder.messageName() == Messages::StorageManagerSet::RemoveItem::name()) {
        IPC::handleMessage<Messages::StorageManagerSet::RemoveItem>(connection, decoder, this, &StorageManagerSet::removeItem);
        return;
    }
    if (decoder.messageName() == Messages::StorageManagerSet::Clear::name()) {
        IPC::handleMessage<Messages::StorageManagerSet::Clear>(connection, decoder, this, &StorageManagerSet::clear);
        return;
    }
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED();
}

void StorageManagerSet::didReceiveSyncMessage(IPC::Connection& connection, IPC::Decoder& decoder, std::unique_ptr<IPC::Encoder>& replyEncoder)
{
    auto protectedThis = makeRef(*this);
    if (decoder.messageName() == Messages::StorageManagerSet::ConnectToLocalStorageArea::name()) {
        IPC::handleMessageSynchronousWantsConnection<Messages::StorageManagerSet::ConnectToLocalStorageArea>(connection, decoder, replyEncoder, this, &StorageManagerSet::connectToLocalStorageArea);
        return;
    }
    if (decoder.messageName() == Messages::StorageManagerSet::ConnectToTransientLocalStorageArea::name()) {
        IPC::handleMessageSynchronousWantsConnection<Messages::StorageManagerSet::ConnectToTransientLocalStorageArea>(connection, decoder, replyEncoder, this, &StorageManagerSet::connectToTransientLocalStorageArea);
        return;
    }
    if (decoder.messageName() == Messages::StorageManagerSet::ConnectToSessionStorageArea::name()) {
        IPC::handleMessageSynchronousWantsConnection<Messages::StorageManagerSet::ConnectToSessionStorageArea>(connection, decoder, replyEncoder, this, &StorageManagerSet::connectToSessionStorageArea);
        return;
    }
    if (decoder.messageName() == Messages::StorageManagerSet::GetValues::name()) {
        IPC::handleMessageSynchronousWantsConnection<Messages::StorageManagerSet::GetValues>(connection, decoder, replyEncoder, this, &StorageManagerSet::getValues);
        return;
    }
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    UNUSED_PARAM(replyEncoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit

