/*
 * Copyright (C) 2010-2019 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#include "WebAutomationSessionProxy.h"

#include "ArgumentCoders.h"
#include "CoordinateSystem.h"
#include "Decoder.h"
#include "HandleMessage.h"
#include "WebAutomationSessionProxyMessages.h"
#include "WebCoreArgumentCoders.h"
#include <WebCore/Cookie.h>
#include <WebCore/FrameIdentifier.h>
#include <WebCore/IntPoint.h>
#include <WebCore/IntRect.h>
#include <WebCore/PageIdentifier.h>
#include <wtf/Optional.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace Messages {

namespace WebAutomationSessionProxy {

void ResolveChildFrameWithOrdinal::callReply(IPC::Decoder& decoder, CompletionHandler<void(Optional<String>&&, Optional<WebCore::FrameIdentifier>&&)>&& completionHandler)
{
    Optional<Optional<String>> errorType;
    decoder >> errorType;
    if (!errorType) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    Optional<Optional<WebCore::FrameIdentifier>> frameID;
    decoder >> frameID;
    if (!frameID) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*errorType), WTFMove(*frameID));
}

void ResolveChildFrameWithOrdinal::cancelReply(CompletionHandler<void(Optional<String>&&, Optional<WebCore::FrameIdentifier>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Optional<String>>::create(), IPC::AsyncReplyError<Optional<WebCore::FrameIdentifier>>::create());
}

void ResolveChildFrameWithOrdinal::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Optional<String>& errorType, const Optional<WebCore::FrameIdentifier>& frameID)
{
    *encoder << errorType;
    *encoder << frameID;
    connection.sendSyncReply(WTFMove(encoder));
}

void ResolveChildFrameWithNodeHandle::callReply(IPC::Decoder& decoder, CompletionHandler<void(Optional<String>&&, Optional<WebCore::FrameIdentifier>&&)>&& completionHandler)
{
    Optional<Optional<String>> errorType;
    decoder >> errorType;
    if (!errorType) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    Optional<Optional<WebCore::FrameIdentifier>> frameID;
    decoder >> frameID;
    if (!frameID) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*errorType), WTFMove(*frameID));
}

void ResolveChildFrameWithNodeHandle::cancelReply(CompletionHandler<void(Optional<String>&&, Optional<WebCore::FrameIdentifier>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Optional<String>>::create(), IPC::AsyncReplyError<Optional<WebCore::FrameIdentifier>>::create());
}

void ResolveChildFrameWithNodeHandle::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Optional<String>& errorType, const Optional<WebCore::FrameIdentifier>& frameID)
{
    *encoder << errorType;
    *encoder << frameID;
    connection.sendSyncReply(WTFMove(encoder));
}

void ResolveChildFrameWithName::callReply(IPC::Decoder& decoder, CompletionHandler<void(Optional<String>&&, Optional<WebCore::FrameIdentifier>&&)>&& completionHandler)
{
    Optional<Optional<String>> errorType;
    decoder >> errorType;
    if (!errorType) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    Optional<Optional<WebCore::FrameIdentifier>> frameID;
    decoder >> frameID;
    if (!frameID) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*errorType), WTFMove(*frameID));
}

void ResolveChildFrameWithName::cancelReply(CompletionHandler<void(Optional<String>&&, Optional<WebCore::FrameIdentifier>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Optional<String>>::create(), IPC::AsyncReplyError<Optional<WebCore::FrameIdentifier>>::create());
}

void ResolveChildFrameWithName::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Optional<String>& errorType, const Optional<WebCore::FrameIdentifier>& frameID)
{
    *encoder << errorType;
    *encoder << frameID;
    connection.sendSyncReply(WTFMove(encoder));
}

void ResolveParentFrame::callReply(IPC::Decoder& decoder, CompletionHandler<void(Optional<String>&&, Optional<WebCore::FrameIdentifier>&&)>&& completionHandler)
{
    Optional<Optional<String>> errorType;
    decoder >> errorType;
    if (!errorType) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    Optional<Optional<WebCore::FrameIdentifier>> frameID;
    decoder >> frameID;
    if (!frameID) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*errorType), WTFMove(*frameID));
}

void ResolveParentFrame::cancelReply(CompletionHandler<void(Optional<String>&&, Optional<WebCore::FrameIdentifier>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Optional<String>>::create(), IPC::AsyncReplyError<Optional<WebCore::FrameIdentifier>>::create());
}

void ResolveParentFrame::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Optional<String>& errorType, const Optional<WebCore::FrameIdentifier>& frameID)
{
    *encoder << errorType;
    *encoder << frameID;
    connection.sendSyncReply(WTFMove(encoder));
}

void ComputeElementLayout::callReply(IPC::Decoder& decoder, CompletionHandler<void(Optional<String>&&, WebCore::IntRect&&, Optional<WebCore::IntPoint>&&, bool&&)>&& completionHandler)
{
    Optional<Optional<String>> errorType;
    decoder >> errorType;
    if (!errorType) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    Optional<WebCore::IntRect> rect;
    decoder >> rect;
    if (!rect) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    Optional<Optional<WebCore::IntPoint>> inViewCenterPoint;
    decoder >> inViewCenterPoint;
    if (!inViewCenterPoint) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    Optional<bool> isObscured;
    decoder >> isObscured;
    if (!isObscured) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*errorType), WTFMove(*rect), WTFMove(*inViewCenterPoint), WTFMove(*isObscured));
}

void ComputeElementLayout::cancelReply(CompletionHandler<void(Optional<String>&&, WebCore::IntRect&&, Optional<WebCore::IntPoint>&&, bool&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Optional<String>>::create(), IPC::AsyncReplyError<WebCore::IntRect>::create(), IPC::AsyncReplyError<Optional<WebCore::IntPoint>>::create(), IPC::AsyncReplyError<bool>::create());
}

void ComputeElementLayout::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Optional<String>& errorType, const WebCore::IntRect& rect, const Optional<WebCore::IntPoint>& inViewCenterPoint, bool isObscured)
{
    *encoder << errorType;
    *encoder << rect;
    *encoder << inViewCenterPoint;
    *encoder << isObscured;
    connection.sendSyncReply(WTFMove(encoder));
}

void SelectOptionElement::callReply(IPC::Decoder& decoder, CompletionHandler<void(Optional<String>&&)>&& completionHandler)
{
    Optional<Optional<String>> errorType;
    decoder >> errorType;
    if (!errorType) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*errorType));
}

void SelectOptionElement::cancelReply(CompletionHandler<void(Optional<String>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Optional<String>>::create());
}

void SelectOptionElement::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Optional<String>& errorType)
{
    *encoder << errorType;
    connection.sendSyncReply(WTFMove(encoder));
}

void SetFilesForInputFileUpload::callReply(IPC::Decoder& decoder, CompletionHandler<void(Optional<String>&&)>&& completionHandler)
{
    Optional<Optional<String>> errorType;
    decoder >> errorType;
    if (!errorType) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*errorType));
}

void SetFilesForInputFileUpload::cancelReply(CompletionHandler<void(Optional<String>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Optional<String>>::create());
}

void SetFilesForInputFileUpload::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Optional<String>& errorType)
{
    *encoder << errorType;
    connection.sendSyncReply(WTFMove(encoder));
}

void SnapshotRectForScreenshot::callReply(IPC::Decoder& decoder, CompletionHandler<void(Optional<String>&&, WebCore::IntRect&&)>&& completionHandler)
{
    Optional<Optional<String>> errorType;
    decoder >> errorType;
    if (!errorType) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    Optional<WebCore::IntRect> rect;
    decoder >> rect;
    if (!rect) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*errorType), WTFMove(*rect));
}

void SnapshotRectForScreenshot::cancelReply(CompletionHandler<void(Optional<String>&&, WebCore::IntRect&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Optional<String>>::create(), IPC::AsyncReplyError<WebCore::IntRect>::create());
}

void SnapshotRectForScreenshot::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Optional<String>& errorType, const WebCore::IntRect& rect)
{
    *encoder << errorType;
    *encoder << rect;
    connection.sendSyncReply(WTFMove(encoder));
}

void GetCookiesForFrame::callReply(IPC::Decoder& decoder, CompletionHandler<void(Optional<String>&&, Vector<WebCore::Cookie>&&)>&& completionHandler)
{
    Optional<Optional<String>> errorType;
    decoder >> errorType;
    if (!errorType) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    Optional<Vector<WebCore::Cookie>> cookies;
    decoder >> cookies;
    if (!cookies) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*errorType), WTFMove(*cookies));
}

void GetCookiesForFrame::cancelReply(CompletionHandler<void(Optional<String>&&, Vector<WebCore::Cookie>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Optional<String>>::create(), IPC::AsyncReplyError<Vector<WebCore::Cookie>>::create());
}

void GetCookiesForFrame::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Optional<String>& errorType, const Vector<WebCore::Cookie>& cookies)
{
    *encoder << errorType;
    *encoder << cookies;
    connection.sendSyncReply(WTFMove(encoder));
}

void DeleteCookie::callReply(IPC::Decoder& decoder, CompletionHandler<void(Optional<String>&&)>&& completionHandler)
{
    Optional<Optional<String>> errorType;
    decoder >> errorType;
    if (!errorType) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*errorType));
}

void DeleteCookie::cancelReply(CompletionHandler<void(Optional<String>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Optional<String>>::create());
}

void DeleteCookie::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Optional<String>& errorType)
{
    *encoder << errorType;
    connection.sendSyncReply(WTFMove(encoder));
}

} // namespace WebAutomationSessionProxy

} // namespace Messages

namespace WebKit {

void WebAutomationSessionProxy::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    if (decoder.messageName() == Messages::WebAutomationSessionProxy::EvaluateJavaScriptFunction::name()) {
        IPC::handleMessage<Messages::WebAutomationSessionProxy::EvaluateJavaScriptFunction>(decoder, this, &WebAutomationSessionProxy::evaluateJavaScriptFunction);
        return;
    }
    if (decoder.messageName() == Messages::WebAutomationSessionProxy::ResolveChildFrameWithOrdinal::name()) {
        IPC::handleMessageAsync<Messages::WebAutomationSessionProxy::ResolveChildFrameWithOrdinal>(connection, decoder, this, &WebAutomationSessionProxy::resolveChildFrameWithOrdinal);
        return;
    }
    if (decoder.messageName() == Messages::WebAutomationSessionProxy::ResolveChildFrameWithNodeHandle::name()) {
        IPC::handleMessageAsync<Messages::WebAutomationSessionProxy::ResolveChildFrameWithNodeHandle>(connection, decoder, this, &WebAutomationSessionProxy::resolveChildFrameWithNodeHandle);
        return;
    }
    if (decoder.messageName() == Messages::WebAutomationSessionProxy::ResolveChildFrameWithName::name()) {
        IPC::handleMessageAsync<Messages::WebAutomationSessionProxy::ResolveChildFrameWithName>(connection, decoder, this, &WebAutomationSessionProxy::resolveChildFrameWithName);
        return;
    }
    if (decoder.messageName() == Messages::WebAutomationSessionProxy::ResolveParentFrame::name()) {
        IPC::handleMessageAsync<Messages::WebAutomationSessionProxy::ResolveParentFrame>(connection, decoder, this, &WebAutomationSessionProxy::resolveParentFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebAutomationSessionProxy::FocusFrame::name()) {
        IPC::handleMessage<Messages::WebAutomationSessionProxy::FocusFrame>(decoder, this, &WebAutomationSessionProxy::focusFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebAutomationSessionProxy::ComputeElementLayout::name()) {
        IPC::handleMessageAsync<Messages::WebAutomationSessionProxy::ComputeElementLayout>(connection, decoder, this, &WebAutomationSessionProxy::computeElementLayout);
        return;
    }
    if (decoder.messageName() == Messages::WebAutomationSessionProxy::SelectOptionElement::name()) {
        IPC::handleMessageAsync<Messages::WebAutomationSessionProxy::SelectOptionElement>(connection, decoder, this, &WebAutomationSessionProxy::selectOptionElement);
        return;
    }
    if (decoder.messageName() == Messages::WebAutomationSessionProxy::SetFilesForInputFileUpload::name()) {
        IPC::handleMessageAsync<Messages::WebAutomationSessionProxy::SetFilesForInputFileUpload>(connection, decoder, this, &WebAutomationSessionProxy::setFilesForInputFileUpload);
        return;
    }
    if (decoder.messageName() == Messages::WebAutomationSessionProxy::TakeScreenshot::name()) {
        IPC::handleMessage<Messages::WebAutomationSessionProxy::TakeScreenshot>(decoder, this, &WebAutomationSessionProxy::takeScreenshot);
        return;
    }
    if (decoder.messageName() == Messages::WebAutomationSessionProxy::SnapshotRectForScreenshot::name()) {
        IPC::handleMessageAsync<Messages::WebAutomationSessionProxy::SnapshotRectForScreenshot>(connection, decoder, this, &WebAutomationSessionProxy::snapshotRectForScreenshot);
        return;
    }
    if (decoder.messageName() == Messages::WebAutomationSessionProxy::GetCookiesForFrame::name()) {
        IPC::handleMessageAsync<Messages::WebAutomationSessionProxy::GetCookiesForFrame>(connection, decoder, this, &WebAutomationSessionProxy::getCookiesForFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebAutomationSessionProxy::DeleteCookie::name()) {
        IPC::handleMessageAsync<Messages::WebAutomationSessionProxy::DeleteCookie>(connection, decoder, this, &WebAutomationSessionProxy::deleteCookie);
        return;
    }
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit

