/*
 * Copyright (C) 2010-2019 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ArgumentCoders.h"
#include "Connection.h"
#include "WebAutomationSessionProxyMessagesReplies.h"
#include <WebCore/Cookie.h>
#include <WebCore/FrameIdentifier.h>
#include <WebCore/IntPoint.h>
#include <WebCore/PageIdentifier.h>
#include <wtf/Forward.h>
#include <wtf/Optional.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace WebCore {
class IntRect;
}

namespace WebKit {
enum class CoordinateSystem : uint8_t;
}

namespace Messages {
namespace WebAutomationSessionProxy {

static inline IPC::StringReference messageReceiverName()
{
    return IPC::StringReference("WebAutomationSessionProxy");
}

class EvaluateJavaScriptFunction {
public:
    typedef std::tuple<const WebCore::PageIdentifier&, const Optional<WebCore::FrameIdentifier>&, const String&, const Vector<String>&, bool, const Optional<double>&, uint64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("EvaluateJavaScriptFunction"); }
    static const bool isSync = false;

    EvaluateJavaScriptFunction(const WebCore::PageIdentifier& pageID, const Optional<WebCore::FrameIdentifier>& frameID, const String& function, const Vector<String>& arguments, bool expectsImplicitCallbackArgument, const Optional<double>& callbackTimeout, uint64_t callbackID)
        : m_arguments(pageID, frameID, function, arguments, expectsImplicitCallbackArgument, callbackTimeout, callbackID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ResolveChildFrameWithOrdinal {
public:
    typedef std::tuple<const WebCore::PageIdentifier&, const Optional<WebCore::FrameIdentifier>&, uint32_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("ResolveChildFrameWithOrdinal"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Optional<String>&&, Optional<WebCore::FrameIdentifier>&&)>&&);
    static void cancelReply(CompletionHandler<void(Optional<String>&&, Optional<WebCore::FrameIdentifier>&&)>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "ResolveChildFrameWithOrdinalReply" }; }
    using AsyncReply = ResolveChildFrameWithOrdinalAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Optional<String>& errorType, const Optional<WebCore::FrameIdentifier>& frameID);
    using Reply = std::tuple<Optional<String>&, Optional<WebCore::FrameIdentifier>&>;
    using ReplyArguments = std::tuple<Optional<String>, Optional<WebCore::FrameIdentifier>>;
    ResolveChildFrameWithOrdinal(const WebCore::PageIdentifier& pageID, const Optional<WebCore::FrameIdentifier>& frameID, uint32_t ordinal)
        : m_arguments(pageID, frameID, ordinal)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ResolveChildFrameWithNodeHandle {
public:
    typedef std::tuple<const WebCore::PageIdentifier&, const Optional<WebCore::FrameIdentifier>&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("ResolveChildFrameWithNodeHandle"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Optional<String>&&, Optional<WebCore::FrameIdentifier>&&)>&&);
    static void cancelReply(CompletionHandler<void(Optional<String>&&, Optional<WebCore::FrameIdentifier>&&)>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "ResolveChildFrameWithNodeHandleReply" }; }
    using AsyncReply = ResolveChildFrameWithNodeHandleAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Optional<String>& errorType, const Optional<WebCore::FrameIdentifier>& frameID);
    using Reply = std::tuple<Optional<String>&, Optional<WebCore::FrameIdentifier>&>;
    using ReplyArguments = std::tuple<Optional<String>, Optional<WebCore::FrameIdentifier>>;
    ResolveChildFrameWithNodeHandle(const WebCore::PageIdentifier& pageID, const Optional<WebCore::FrameIdentifier>& frameID, const String& nodeHandle)
        : m_arguments(pageID, frameID, nodeHandle)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ResolveChildFrameWithName {
public:
    typedef std::tuple<const WebCore::PageIdentifier&, const Optional<WebCore::FrameIdentifier>&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("ResolveChildFrameWithName"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Optional<String>&&, Optional<WebCore::FrameIdentifier>&&)>&&);
    static void cancelReply(CompletionHandler<void(Optional<String>&&, Optional<WebCore::FrameIdentifier>&&)>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "ResolveChildFrameWithNameReply" }; }
    using AsyncReply = ResolveChildFrameWithNameAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Optional<String>& errorType, const Optional<WebCore::FrameIdentifier>& frameID);
    using Reply = std::tuple<Optional<String>&, Optional<WebCore::FrameIdentifier>&>;
    using ReplyArguments = std::tuple<Optional<String>, Optional<WebCore::FrameIdentifier>>;
    ResolveChildFrameWithName(const WebCore::PageIdentifier& pageID, const Optional<WebCore::FrameIdentifier>& frameID, const String& name)
        : m_arguments(pageID, frameID, name)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ResolveParentFrame {
public:
    typedef std::tuple<const WebCore::PageIdentifier&, const Optional<WebCore::FrameIdentifier>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("ResolveParentFrame"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Optional<String>&&, Optional<WebCore::FrameIdentifier>&&)>&&);
    static void cancelReply(CompletionHandler<void(Optional<String>&&, Optional<WebCore::FrameIdentifier>&&)>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "ResolveParentFrameReply" }; }
    using AsyncReply = ResolveParentFrameAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Optional<String>& errorType, const Optional<WebCore::FrameIdentifier>& frameID);
    using Reply = std::tuple<Optional<String>&, Optional<WebCore::FrameIdentifier>&>;
    using ReplyArguments = std::tuple<Optional<String>, Optional<WebCore::FrameIdentifier>>;
    ResolveParentFrame(const WebCore::PageIdentifier& pageID, const Optional<WebCore::FrameIdentifier>& frameID)
        : m_arguments(pageID, frameID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class FocusFrame {
public:
    typedef std::tuple<const WebCore::PageIdentifier&, const Optional<WebCore::FrameIdentifier>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("FocusFrame"); }
    static const bool isSync = false;

    FocusFrame(const WebCore::PageIdentifier& pageID, const Optional<WebCore::FrameIdentifier>& frameID)
        : m_arguments(pageID, frameID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ComputeElementLayout {
public:
    typedef std::tuple<const WebCore::PageIdentifier&, const Optional<WebCore::FrameIdentifier>&, const String&, bool, WebKit::CoordinateSystem> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("ComputeElementLayout"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Optional<String>&&, WebCore::IntRect&&, Optional<WebCore::IntPoint>&&, bool&&)>&&);
    static void cancelReply(CompletionHandler<void(Optional<String>&&, WebCore::IntRect&&, Optional<WebCore::IntPoint>&&, bool&&)>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "ComputeElementLayoutReply" }; }
    using AsyncReply = ComputeElementLayoutAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Optional<String>& errorType, const WebCore::IntRect& rect, const Optional<WebCore::IntPoint>& inViewCenterPoint, bool isObscured);
    using Reply = std::tuple<Optional<String>&, WebCore::IntRect&, Optional<WebCore::IntPoint>&, bool&>;
    using ReplyArguments = std::tuple<Optional<String>, WebCore::IntRect, Optional<WebCore::IntPoint>, bool>;
    ComputeElementLayout(const WebCore::PageIdentifier& pageID, const Optional<WebCore::FrameIdentifier>& frameID, const String& nodeHandle, bool scrollIntoViewIfNeeded, WebKit::CoordinateSystem coordinateSystem)
        : m_arguments(pageID, frameID, nodeHandle, scrollIntoViewIfNeeded, coordinateSystem)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SelectOptionElement {
public:
    typedef std::tuple<const WebCore::PageIdentifier&, const Optional<WebCore::FrameIdentifier>&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SelectOptionElement"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Optional<String>&&)>&&);
    static void cancelReply(CompletionHandler<void(Optional<String>&&)>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "SelectOptionElementReply" }; }
    using AsyncReply = SelectOptionElementAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Optional<String>& errorType);
    using Reply = std::tuple<Optional<String>&>;
    using ReplyArguments = std::tuple<Optional<String>>;
    SelectOptionElement(const WebCore::PageIdentifier& pageID, const Optional<WebCore::FrameIdentifier>& frameID, const String& nodeHandle)
        : m_arguments(pageID, frameID, nodeHandle)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetFilesForInputFileUpload {
public:
    typedef std::tuple<const WebCore::PageIdentifier&, const Optional<WebCore::FrameIdentifier>&, const String&, const Vector<String>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SetFilesForInputFileUpload"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Optional<String>&&)>&&);
    static void cancelReply(CompletionHandler<void(Optional<String>&&)>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "SetFilesForInputFileUploadReply" }; }
    using AsyncReply = SetFilesForInputFileUploadAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Optional<String>& errorType);
    using Reply = std::tuple<Optional<String>&>;
    using ReplyArguments = std::tuple<Optional<String>>;
    SetFilesForInputFileUpload(const WebCore::PageIdentifier& pageID, const Optional<WebCore::FrameIdentifier>& frameID, const String& nodeHandle, const Vector<String>& filenames)
        : m_arguments(pageID, frameID, nodeHandle, filenames)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class TakeScreenshot {
public:
    typedef std::tuple<const WebCore::PageIdentifier&, const Optional<WebCore::FrameIdentifier>&, const String&, bool, bool, uint64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("TakeScreenshot"); }
    static const bool isSync = false;

    TakeScreenshot(const WebCore::PageIdentifier& pageID, const Optional<WebCore::FrameIdentifier>& frameID, const String& nodeHandle, bool scrollIntoViewIfNeeded, bool clipToViewport, uint64_t callbackID)
        : m_arguments(pageID, frameID, nodeHandle, scrollIntoViewIfNeeded, clipToViewport, callbackID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SnapshotRectForScreenshot {
public:
    typedef std::tuple<const WebCore::PageIdentifier&, const Optional<WebCore::FrameIdentifier>&, const String&, bool, bool> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SnapshotRectForScreenshot"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Optional<String>&&, WebCore::IntRect&&)>&&);
    static void cancelReply(CompletionHandler<void(Optional<String>&&, WebCore::IntRect&&)>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "SnapshotRectForScreenshotReply" }; }
    using AsyncReply = SnapshotRectForScreenshotAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Optional<String>& errorType, const WebCore::IntRect& rect);
    using Reply = std::tuple<Optional<String>&, WebCore::IntRect&>;
    using ReplyArguments = std::tuple<Optional<String>, WebCore::IntRect>;
    SnapshotRectForScreenshot(const WebCore::PageIdentifier& pageID, const Optional<WebCore::FrameIdentifier>& frameID, const String& nodeHandle, bool scrollIntoViewIfNeeded, bool clipToViewport)
        : m_arguments(pageID, frameID, nodeHandle, scrollIntoViewIfNeeded, clipToViewport)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetCookiesForFrame {
public:
    typedef std::tuple<const WebCore::PageIdentifier&, const Optional<WebCore::FrameIdentifier>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetCookiesForFrame"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Optional<String>&&, Vector<WebCore::Cookie>&&)>&&);
    static void cancelReply(CompletionHandler<void(Optional<String>&&, Vector<WebCore::Cookie>&&)>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "GetCookiesForFrameReply" }; }
    using AsyncReply = GetCookiesForFrameAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Optional<String>& errorType, const Vector<WebCore::Cookie>& cookies);
    using Reply = std::tuple<Optional<String>&, Vector<WebCore::Cookie>&>;
    using ReplyArguments = std::tuple<Optional<String>, Vector<WebCore::Cookie>>;
    GetCookiesForFrame(const WebCore::PageIdentifier& pageID, const Optional<WebCore::FrameIdentifier>& frameID)
        : m_arguments(pageID, frameID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DeleteCookie {
public:
    typedef std::tuple<const WebCore::PageIdentifier&, const Optional<WebCore::FrameIdentifier>&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("DeleteCookie"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Optional<String>&&)>&&);
    static void cancelReply(CompletionHandler<void(Optional<String>&&)>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "DeleteCookieReply" }; }
    using AsyncReply = DeleteCookieAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Optional<String>& errorType);
    using Reply = std::tuple<Optional<String>&>;
    using ReplyArguments = std::tuple<Optional<String>>;
    DeleteCookie(const WebCore::PageIdentifier& pageID, const Optional<WebCore::FrameIdentifier>& frameID, const String& cookieName)
        : m_arguments(pageID, frameID, cookieName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

} // namespace WebAutomationSessionProxy
} // namespace Messages
