/*
 * Copyright (C) 2010-2019 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ArgumentCoders.h"
#include "Connection.h"
#include "WebCookieManagerMessagesReplies.h"
#include <WebCore/Cookie.h>
#include <wtf/Forward.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace PAL {
class SessionID;
}

namespace WebCore {
enum class HTTPCookieAcceptPolicy : uint8_t;
struct Cookie;
}

namespace WebKit {
enum class SoupCookiePersistentStorageType : bool;
}

namespace Messages {
namespace WebCookieManager {

static inline IPC::StringReference messageReceiverName()
{
    return IPC::StringReference("WebCookieManager");
}

class GetHostnamesWithCookies {
public:
    typedef std::tuple<const PAL::SessionID&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetHostnamesWithCookies"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Vector<String>&&)>&&);
    static void cancelReply(CompletionHandler<void(Vector<String>&&)>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "GetHostnamesWithCookiesReply" }; }
    using AsyncReply = GetHostnamesWithCookiesAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<String>& hostnames);
    using Reply = std::tuple<Vector<String>&>;
    using ReplyArguments = std::tuple<Vector<String>>;
    explicit GetHostnamesWithCookies(const PAL::SessionID& sessionID)
        : m_arguments(sessionID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DeleteCookiesForHostnames {
public:
    typedef std::tuple<const PAL::SessionID&, const Vector<String>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("DeleteCookiesForHostnames"); }
    static const bool isSync = false;

    DeleteCookiesForHostnames(const PAL::SessionID& sessionID, const Vector<String>& hostnames)
        : m_arguments(sessionID, hostnames)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DeleteAllCookies {
public:
    typedef std::tuple<const PAL::SessionID&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("DeleteAllCookies"); }
    static const bool isSync = false;

    explicit DeleteAllCookies(const PAL::SessionID& sessionID)
        : m_arguments(sessionID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetCookie {
public:
    typedef std::tuple<const PAL::SessionID&, const Vector<WebCore::Cookie>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SetCookie"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "SetCookieReply" }; }
    using AsyncReply = SetCookieAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    SetCookie(const PAL::SessionID& sessionID, const Vector<WebCore::Cookie>& cookie)
        : m_arguments(sessionID, cookie)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetCookies {
public:
    typedef std::tuple<const PAL::SessionID&, const Vector<WebCore::Cookie>&, const URL&, const URL&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SetCookies"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "SetCookiesReply" }; }
    using AsyncReply = SetCookiesAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    SetCookies(const PAL::SessionID& sessionID, const Vector<WebCore::Cookie>& cookies, const URL& url, const URL& mainDocumentURL)
        : m_arguments(sessionID, cookies, url, mainDocumentURL)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetAllCookies {
public:
    typedef std::tuple<const PAL::SessionID&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetAllCookies"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Vector<WebCore::Cookie>&&)>&&);
    static void cancelReply(CompletionHandler<void(Vector<WebCore::Cookie>&&)>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "GetAllCookiesReply" }; }
    using AsyncReply = GetAllCookiesAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<WebCore::Cookie>& cookies);
    using Reply = std::tuple<Vector<WebCore::Cookie>&>;
    using ReplyArguments = std::tuple<Vector<WebCore::Cookie>>;
    explicit GetAllCookies(const PAL::SessionID& sessionID)
        : m_arguments(sessionID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetCookies {
public:
    typedef std::tuple<const PAL::SessionID&, const URL&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetCookies"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Vector<WebCore::Cookie>&&)>&&);
    static void cancelReply(CompletionHandler<void(Vector<WebCore::Cookie>&&)>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "GetCookiesReply" }; }
    using AsyncReply = GetCookiesAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<WebCore::Cookie>& cookies);
    using Reply = std::tuple<Vector<WebCore::Cookie>&>;
    using ReplyArguments = std::tuple<Vector<WebCore::Cookie>>;
    GetCookies(const PAL::SessionID& sessionID, const URL& url)
        : m_arguments(sessionID, url)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DeleteCookie {
public:
    typedef std::tuple<const PAL::SessionID&, const WebCore::Cookie&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("DeleteCookie"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "DeleteCookieReply" }; }
    using AsyncReply = DeleteCookieAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    DeleteCookie(const PAL::SessionID& sessionID, const WebCore::Cookie& cookie)
        : m_arguments(sessionID, cookie)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DeleteAllCookiesModifiedSince {
public:
    typedef std::tuple<const PAL::SessionID&, const WallTime&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("DeleteAllCookiesModifiedSince"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "DeleteAllCookiesModifiedSinceReply" }; }
    using AsyncReply = DeleteAllCookiesModifiedSinceAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    DeleteAllCookiesModifiedSince(const PAL::SessionID& sessionID, const WallTime& time)
        : m_arguments(sessionID, time)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetHTTPCookieAcceptPolicy {
public:
    typedef std::tuple<WebCore::HTTPCookieAcceptPolicy> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SetHTTPCookieAcceptPolicy"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "SetHTTPCookieAcceptPolicyReply" }; }
    using AsyncReply = SetHTTPCookieAcceptPolicyAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    explicit SetHTTPCookieAcceptPolicy(WebCore::HTTPCookieAcceptPolicy policy)
        : m_arguments(policy)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetHTTPCookieAcceptPolicy {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetHTTPCookieAcceptPolicy"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebCore::HTTPCookieAcceptPolicy&&)>&&);
    static void cancelReply(CompletionHandler<void(WebCore::HTTPCookieAcceptPolicy&&)>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "GetHTTPCookieAcceptPolicyReply" }; }
    using AsyncReply = GetHTTPCookieAcceptPolicyAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, WebCore::HTTPCookieAcceptPolicy policy);
    using Reply = std::tuple<WebCore::HTTPCookieAcceptPolicy&>;
    using ReplyArguments = std::tuple<WebCore::HTTPCookieAcceptPolicy>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class StartObservingCookieChanges {
public:
    typedef std::tuple<const PAL::SessionID&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("StartObservingCookieChanges"); }
    static const bool isSync = false;

    explicit StartObservingCookieChanges(const PAL::SessionID& sessionID)
        : m_arguments(sessionID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class StopObservingCookieChanges {
public:
    typedef std::tuple<const PAL::SessionID&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("StopObservingCookieChanges"); }
    static const bool isSync = false;

    explicit StopObservingCookieChanges(const PAL::SessionID& sessionID)
        : m_arguments(sessionID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if USE(SOUP)
class SetCookiePersistentStorage {
public:
    typedef std::tuple<const PAL::SessionID&, const String&, WebKit::SoupCookiePersistentStorageType> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SetCookiePersistentStorage"); }
    static const bool isSync = false;

    SetCookiePersistentStorage(const PAL::SessionID& sessionID, const String& storagePath, WebKit::SoupCookiePersistentStorageType storageType)
        : m_arguments(sessionID, storagePath, storageType)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

} // namespace WebCookieManager
} // namespace Messages
