/*
 * Copyright (C) 2010-2019 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#if ENABLE(INDEXED_DATABASE)

#include "WebIDBServer.h"

#include "ArgumentCoders.h"
#include "Decoder.h"
#include "HandleMessage.h"
#include "WebCoreArgumentCoders.h"
#include "WebIDBServerMessages.h"
#include <WebCore/IDBCursorInfo.h>
#include <WebCore/IDBGetAllRecordsData.h>
#include <WebCore/IDBGetRecordData.h>
#include <WebCore/IDBIndexInfo.h>
#include <WebCore/IDBIterateCursorData.h>
#include <WebCore/IDBKeyData.h>
#include <WebCore/IDBKeyRangeData.h>
#include <WebCore/IDBObjectStoreInfo.h>
#include <WebCore/IDBRequestData.h>
#include <WebCore/IDBResourceIdentifier.h>
#include <WebCore/IDBTransactionInfo.h>
#include <WebCore/IDBValue.h>
#include <WebCore/IndexedDB.h>
#include <WebCore/SecurityOriginData.h>
#include <wtf/text/WTFString.h>

namespace WebKit {

void WebIDBServer::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    auto protectedThis = makeRef(*this);
    if (decoder.messageName() == Messages::WebIDBServer::DeleteDatabase::name()) {
        IPC::handleMessage<Messages::WebIDBServer::DeleteDatabase>(decoder, this, &WebIDBServer::deleteDatabase);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::OpenDatabase::name()) {
        IPC::handleMessage<Messages::WebIDBServer::OpenDatabase>(decoder, this, &WebIDBServer::openDatabase);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::AbortTransaction::name()) {
        IPC::handleMessage<Messages::WebIDBServer::AbortTransaction>(decoder, this, &WebIDBServer::abortTransaction);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::CommitTransaction::name()) {
        IPC::handleMessage<Messages::WebIDBServer::CommitTransaction>(decoder, this, &WebIDBServer::commitTransaction);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::DidFinishHandlingVersionChangeTransaction::name()) {
        IPC::handleMessage<Messages::WebIDBServer::DidFinishHandlingVersionChangeTransaction>(decoder, this, &WebIDBServer::didFinishHandlingVersionChangeTransaction);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::CreateObjectStore::name()) {
        IPC::handleMessage<Messages::WebIDBServer::CreateObjectStore>(decoder, this, &WebIDBServer::createObjectStore);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::DeleteObjectStore::name()) {
        IPC::handleMessage<Messages::WebIDBServer::DeleteObjectStore>(decoder, this, &WebIDBServer::deleteObjectStore);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::RenameObjectStore::name()) {
        IPC::handleMessage<Messages::WebIDBServer::RenameObjectStore>(decoder, this, &WebIDBServer::renameObjectStore);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::ClearObjectStore::name()) {
        IPC::handleMessage<Messages::WebIDBServer::ClearObjectStore>(decoder, this, &WebIDBServer::clearObjectStore);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::CreateIndex::name()) {
        IPC::handleMessage<Messages::WebIDBServer::CreateIndex>(decoder, this, &WebIDBServer::createIndex);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::DeleteIndex::name()) {
        IPC::handleMessage<Messages::WebIDBServer::DeleteIndex>(decoder, this, &WebIDBServer::deleteIndex);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::RenameIndex::name()) {
        IPC::handleMessage<Messages::WebIDBServer::RenameIndex>(decoder, this, &WebIDBServer::renameIndex);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::PutOrAdd::name()) {
        IPC::handleMessage<Messages::WebIDBServer::PutOrAdd>(decoder, this, &WebIDBServer::putOrAdd);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::GetRecord::name()) {
        IPC::handleMessage<Messages::WebIDBServer::GetRecord>(decoder, this, &WebIDBServer::getRecord);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::GetAllRecords::name()) {
        IPC::handleMessage<Messages::WebIDBServer::GetAllRecords>(decoder, this, &WebIDBServer::getAllRecords);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::GetCount::name()) {
        IPC::handleMessage<Messages::WebIDBServer::GetCount>(decoder, this, &WebIDBServer::getCount);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::DeleteRecord::name()) {
        IPC::handleMessage<Messages::WebIDBServer::DeleteRecord>(decoder, this, &WebIDBServer::deleteRecord);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::OpenCursor::name()) {
        IPC::handleMessage<Messages::WebIDBServer::OpenCursor>(decoder, this, &WebIDBServer::openCursor);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::IterateCursor::name()) {
        IPC::handleMessage<Messages::WebIDBServer::IterateCursor>(decoder, this, &WebIDBServer::iterateCursor);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::EstablishTransaction::name()) {
        IPC::handleMessage<Messages::WebIDBServer::EstablishTransaction>(decoder, this, &WebIDBServer::establishTransaction);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::DatabaseConnectionPendingClose::name()) {
        IPC::handleMessage<Messages::WebIDBServer::DatabaseConnectionPendingClose>(decoder, this, &WebIDBServer::databaseConnectionPendingClose);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::DatabaseConnectionClosed::name()) {
        IPC::handleMessage<Messages::WebIDBServer::DatabaseConnectionClosed>(decoder, this, &WebIDBServer::databaseConnectionClosed);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::AbortOpenAndUpgradeNeeded::name()) {
        IPC::handleMessage<Messages::WebIDBServer::AbortOpenAndUpgradeNeeded>(decoder, this, &WebIDBServer::abortOpenAndUpgradeNeeded);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::DidFireVersionChangeEvent::name()) {
        IPC::handleMessage<Messages::WebIDBServer::DidFireVersionChangeEvent>(decoder, this, &WebIDBServer::didFireVersionChangeEvent);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::OpenDBRequestCancelled::name()) {
        IPC::handleMessage<Messages::WebIDBServer::OpenDBRequestCancelled>(decoder, this, &WebIDBServer::openDBRequestCancelled);
        return;
    }
    if (decoder.messageName() == Messages::WebIDBServer::GetAllDatabaseNames::name()) {
        IPC::handleMessage<Messages::WebIDBServer::GetAllDatabaseNames>(connection, decoder, this, &WebIDBServer::getAllDatabaseNames);
        return;
    }
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit


#endif // ENABLE(INDEXED_DATABASE)
