/*
 * Copyright (C) 2010-2019 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#include "WebPage.h"

#include "ArgumentCoders.h"
#if PLATFORM(COCOA)
#include "Attachment.h"
#endif
#if PLATFORM(COCOA)
#include "AttributedString.h"
#endif
#include "CallbackID.h"
#include "ContentWorldShared.h"
#if ENABLE(DATA_DETECTION)
#include "DataDetectionResult.h"
#endif
#include "DataReference.h"
#include "Decoder.h"
#if PLATFORM(IOS_FAMILY)
#include "DocumentEditingContext.h"
#endif
#include "DownloadID.h"
#if !PLATFORM(GTK) && ENABLE(DRAG_SUPPORT) || PLATFORM(GTK) && ENABLE(DRAG_SUPPORT)
#include "DragControllerAction.h"
#endif
#include "DrawingAreaInfo.h"
#if PLATFORM(COCOA) || PLATFORM(MAC)
#include "EditingRange.h"
#endif
#include "FormDataReference.h"
#if ENABLE(GAMEPAD)
#include "GamepadData.h"
#endif
#include "HandleMessage.h"
#if PLATFORM(COCOA)
#include "InsertTextOptions.h"
#endif
#if PLATFORM(IOS_FAMILY)
#include "InteractionInformationAtPosition.h"
#endif
#if PLATFORM(IOS_FAMILY)
#include "InteractionInformationRequest.h"
#endif
#include "LayerTreeContext.h"
#include "LoadParameters.h"
#include "OptionalCallbackID.h"
#include "PrintInfo.h"
#include "SandboxExtension.h"
#include "SessionState.h"
#include "SharedBufferDataReference.h"
#if PLATFORM(COCOA)
#include "SharedMemory.h"
#endif
#if PLATFORM(IOS_FAMILY)
#include "SyntheticEditingCommandType.h"
#endif
#if PLATFORM(IOS_FAMILY)
#include "TransactionID.h"
#endif
#include "UserData.h"
#if PLATFORM(GTK) || PLATFORM(WPE)
#include "UserMessage.h"
#endif
#if PLATFORM(IOS_FAMILY)
#include "WebAutocorrectionData.h"
#endif
#if ENABLE(CONTEXT_MENUS)
#include "WebContextMenuItemData.h"
#endif
#include "WebCoreArgumentCoders.h"
#include "WebEvent.h"
#include "WebPageMessages.h"
#include "WebPageProxyIdentifier.h"
#include "WebPreferencesStore.h"
#if PLATFORM(GTK) && ENABLE(DRAG_SUPPORT)
#include "WebSelectionData.h"
#endif
#include "WebsitePoliciesData.h"
#include <JavaScriptCore/ConsoleTypes.h>
#include <JavaScriptCore/InspectorFrontendChannel.h>
#include <WebCore/ActivityState.h>
#include <WebCore/BackForwardItemIdentifier.h>
#if ENABLE(MEDIA_STREAM)
#include <WebCore/CaptureDevice.h>
#endif
#include <WebCore/Color.h>
#if PLATFORM(COCOA)
#include <WebCore/CompositionHighlight.h>
#endif
#if PLATFORM(COCOA)
#include <WebCore/CompositionUnderline.h>
#endif
#if PLATFORM(MAC)
#include <WebCore/DictationAlternative.h>
#endif
#if !PLATFORM(GTK) && ENABLE(DRAG_SUPPORT)
#include <WebCore/DragData.h>
#endif
#include <WebCore/ElementContext.h>
#if PLATFORM(COCOA) || PLATFORM(IOS_FAMILY) || PLATFORM(MAC)
#include <WebCore/FloatPoint.h>
#endif
#include <WebCore/FloatRect.h>
#if PLATFORM(IOS_FAMILY)
#include <WebCore/FloatSize.h>
#endif
#include <WebCore/FontAttributeChanges.h>
#include <WebCore/FrameIdentifier.h>
#include <WebCore/FrameLoaderTypes.h>
#include <WebCore/GlobalFrameIdentifier.h>
#include <WebCore/GlobalWindowIdentifier.h>
#include <WebCore/IntPoint.h>
#include <WebCore/IntRect.h>
#include <WebCore/IntSize.h>
#include <WebCore/LayoutMilestone.h>
#if ENABLE(WIRELESS_PLAYBACK_TARGET) && !PLATFORM(IOS_FAMILY)
#include <WebCore/MediaPlaybackTargetContext.h>
#endif
#include <WebCore/MediaProducer.h>
#if PLATFORM(MAC)
#include <WebCore/PageOverlay.h>
#endif
#include <WebCore/PlatformEvent.h>
#if ENABLE(POINTER_EVENTS) || PLATFORM(IOS_FAMILY)
#include <WebCore/PointerID.h>
#endif
#if PLATFORM(IOS_FAMILY)
#include <WebCore/RectEdges.h>
#endif
#include <WebCore/ResourceError.h>
#include <WebCore/ResourceRequest.h>
#include <WebCore/ResourceResponse.h>
#include <WebCore/RunJavaScriptParameters.h>
#include <WebCore/SecurityPolicyViolationEvent.h>
#include <WebCore/ShouldTreatAsContinuingLoad.h>
#include <WebCore/TextCheckerClient.h>
#if PLATFORM(IOS_FAMILY) && ENABLE(DRAG_SUPPORT)
#include <WebCore/TextIndicator.h>
#endif
#include <WebCore/TextManipulationController.h>
#if PLATFORM(IOS_FAMILY)
#include <WebCore/ViewportArguments.h>
#endif
#include <WebCore/WritingDirection.h>
#include <utility>
#include <wtf/OptionSet.h>
#include <wtf/Optional.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace Messages {

namespace WebPage {

void TestProcessIncomingSyncMessagesWhenWaitingForSyncReply::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool handled)
{
    *encoder << handled;
    connection.sendSyncReply(WTFMove(encoder));
}

#if PLATFORM(IOS_FAMILY)

void RequestAutocorrectionData::callReply(IPC::Decoder& decoder, CompletionHandler<void(WebKit::WebAutocorrectionData&&)>&& completionHandler)
{
    Optional<WebKit::WebAutocorrectionData> data;
    decoder >> data;
    if (!data) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*data));
}

void RequestAutocorrectionData::cancelReply(CompletionHandler<void(WebKit::WebAutocorrectionData&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<WebKit::WebAutocorrectionData>::create());
}

void RequestAutocorrectionData::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::WebAutocorrectionData& data)
{
    *encoder << data;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(IOS_FAMILY)

void SyncApplyAutocorrection::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool autocorrectionApplied)
{
    *encoder << autocorrectionApplied;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(IOS_FAMILY)

void RequestEvasionRectsAboveSelection::callReply(IPC::Decoder& decoder, CompletionHandler<void(Vector<WebCore::FloatRect>&&)>&& completionHandler)
{
    Optional<Vector<WebCore::FloatRect>> rects;
    decoder >> rects;
    if (!rects) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*rects));
}

void RequestEvasionRectsAboveSelection::cancelReply(CompletionHandler<void(Vector<WebCore::FloatRect>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Vector<WebCore::FloatRect>>::create());
}

void RequestEvasionRectsAboveSelection::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<WebCore::FloatRect>& rects)
{
    *encoder << rects;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(IOS_FAMILY)

void GetPositionInformation::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::InteractionInformationAtPosition& information)
{
    *encoder << information;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(IOS_FAMILY)

void UpdateSelectionWithDelta::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void UpdateSelectionWithDelta::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void UpdateSelectionWithDelta::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(IOS_FAMILY)

void RequestDocumentEditingContext::callReply(IPC::Decoder& decoder, CompletionHandler<void(WebKit::DocumentEditingContext&&)>&& completionHandler)
{
    Optional<WebKit::DocumentEditingContext> response;
    decoder >> response;
    if (!response) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*response));
}

void RequestDocumentEditingContext::cancelReply(CompletionHandler<void(WebKit::DocumentEditingContext&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<WebKit::DocumentEditingContext>::create());
}

void RequestDocumentEditingContext::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::DocumentEditingContext& response)
{
    *encoder << response;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(IOS_TOUCH_EVENTS)

void TouchEventSync::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool handled)
{
    *encoder << handled;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void GetContentsAsAttributedString::callReply(IPC::Decoder& decoder, CompletionHandler<void(WebKit::AttributedString&&)>&& completionHandler)
{
    Optional<WebKit::AttributedString> result;
    decoder >> result;
    if (!result) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*result));
}

void GetContentsAsAttributedString::cancelReply(CompletionHandler<void(WebKit::AttributedString&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<WebKit::AttributedString>::create());
}

void GetContentsAsAttributedString::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::AttributedString& result)
{
    *encoder << result;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(DATA_DETECTION)

void DetectDataInAllFrames::callReply(IPC::Decoder& decoder, CompletionHandler<void(WebKit::DataDetectionResult&&)>&& completionHandler)
{
    Optional<WebKit::DataDetectionResult> result;
    decoder >> result;
    if (!result) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*result));
}

void DetectDataInAllFrames::cancelReply(CompletionHandler<void(WebKit::DataDetectionResult&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<WebKit::DataDetectionResult>::create());
}

void DetectDataInAllFrames::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::DataDetectionResult& result)
{
    *encoder << result;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(DATA_DETECTION)

void RemoveDataDetectedLinks::callReply(IPC::Decoder& decoder, CompletionHandler<void(WebKit::DataDetectionResult&&)>&& completionHandler)
{
    Optional<WebKit::DataDetectionResult> result;
    decoder >> result;
    if (!result) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*result));
}

void RemoveDataDetectedLinks::cancelReply(CompletionHandler<void(WebKit::DataDetectionResult&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<WebKit::DataDetectionResult>::create());
}

void RemoveDataDetectedLinks::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::DataDetectionResult& result)
{
    *encoder << result;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

void TryClose::callReply(IPC::Decoder& decoder, CompletionHandler<void(bool&&)>&& completionHandler)
{
    Optional<bool> shouldClose;
    decoder >> shouldClose;
    if (!shouldClose) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*shouldClose));
}

void TryClose::cancelReply(CompletionHandler<void(bool&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<bool>::create());
}

void TryClose::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool shouldClose)
{
    *encoder << shouldClose;
    connection.sendSyncReply(WTFMove(encoder));
}

#if PLATFORM(IOS_FAMILY) && ENABLE(DRAG_SUPPORT)

void InsertDroppedImagePlaceholders::callReply(IPC::Decoder& decoder, CompletionHandler<void(Vector<WebCore::IntRect>&&, Optional<WebCore::TextIndicatorData>&&)>&& completionHandler)
{
    Optional<Vector<WebCore::IntRect>> dropPreviewRects;
    decoder >> dropPreviewRects;
    if (!dropPreviewRects) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    Optional<Optional<WebCore::TextIndicatorData>> textIndicator;
    decoder >> textIndicator;
    if (!textIndicator) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*dropPreviewRects), WTFMove(*textIndicator));
}

void InsertDroppedImagePlaceholders::cancelReply(CompletionHandler<void(Vector<WebCore::IntRect>&&, Optional<WebCore::TextIndicatorData>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Vector<WebCore::IntRect>>::create(), IPC::AsyncReplyError<Optional<WebCore::TextIndicatorData>>::create());
}

void InsertDroppedImagePlaceholders::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<WebCore::IntRect>& dropPreviewRects, const Optional<WebCore::TextIndicatorData>& textIndicator)
{
    *encoder << dropPreviewRects;
    *encoder << textIndicator;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(MEDIA_STREAM)

void UserMediaAccessWasGranted::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void UserMediaAccessWasGranted::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void UserMediaAccessWasGranted::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if (PLATFORM(COCOA) && PLATFORM(IOS_FAMILY))

void ComputePagesForPrintingAndDrawToPDF::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, uint32_t pageCount)
{
    *encoder << pageCount;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(GTK)

void GetCenterForZoomGesture::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebCore::IntPoint& center)
{
    *encoder << center;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void GetStringSelectionForPasteboard::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const String& stringValue)
{
    *encoder << stringValue;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void GetDataSelectionForPasteboard::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::SharedMemory::Handle& handle, uint64_t size)
{
    *encoder << handle;
    *encoder << size;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void ReadSelectionFromPasteboard::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool result)
{
    *encoder << result;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void ShouldDelayWindowOrderingEvent::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool result)
{
    *encoder << result;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void AcceptsFirstMouse::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool result)
{
    *encoder << result;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void HasMarkedText::callReply(IPC::Decoder& decoder, CompletionHandler<void(bool&&)>&& completionHandler)
{
    Optional<bool> hasMarkedText;
    decoder >> hasMarkedText;
    if (!hasMarkedText) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*hasMarkedText));
}

void HasMarkedText::cancelReply(CompletionHandler<void(bool&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<bool>::create());
}

void HasMarkedText::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool hasMarkedText)
{
    *encoder << hasMarkedText;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

void TextInputContextsInRect::callReply(IPC::Decoder& decoder, CompletionHandler<void(Vector<struct WebCore::ElementContext>&&)>&& completionHandler)
{
    Optional<Vector<struct WebCore::ElementContext>> contexts;
    decoder >> contexts;
    if (!contexts) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*contexts));
}

void TextInputContextsInRect::cancelReply(CompletionHandler<void(Vector<struct WebCore::ElementContext>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Vector<struct WebCore::ElementContext>>::create());
}

void TextInputContextsInRect::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<struct WebCore::ElementContext>& contexts)
{
    *encoder << contexts;
    connection.sendSyncReply(WTFMove(encoder));
}

void FocusTextInputContext::callReply(IPC::Decoder& decoder, CompletionHandler<void(bool&&)>&& completionHandler)
{
    Optional<bool> success;
    decoder >> success;
    if (!success) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*success));
}

void FocusTextInputContext::cancelReply(CompletionHandler<void(bool&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<bool>::create());
}

void FocusTextInputContext::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool success)
{
    *encoder << success;
    connection.sendSyncReply(WTFMove(encoder));
}

#if PLATFORM(GTK) || PLATFORM(WPE)

void SendMessageToWebExtensionWithReply::callReply(IPC::Decoder& decoder, CompletionHandler<void(WebKit::UserMessage&&)>&& completionHandler)
{
    Optional<WebKit::UserMessage> replyMessage;
    decoder >> replyMessage;
    if (!replyMessage) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*replyMessage));
}

void SendMessageToWebExtensionWithReply::cancelReply(CompletionHandler<void(WebKit::UserMessage&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<WebKit::UserMessage>::create());
}

void SendMessageToWebExtensionWithReply::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::UserMessage& replyMessage)
{
    *encoder << replyMessage;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

void StartTextManipulations::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void StartTextManipulations::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void StartTextManipulations::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

void CompleteTextManipulation::callReply(IPC::Decoder& decoder, CompletionHandler<void(WebCore::TextManipulationController::ManipulationResult&&)>&& completionHandler)
{
    Optional<WebCore::TextManipulationController::ManipulationResult> result;
    decoder >> result;
    if (!result) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*result));
}

void CompleteTextManipulation::cancelReply(CompletionHandler<void(WebCore::TextManipulationController::ManipulationResult&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<WebCore::TextManipulationController::ManipulationResult>::create());
}

void CompleteTextManipulation::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, WebCore::TextManipulationController::ManipulationResult result)
{
    *encoder << result;
    connection.sendSyncReply(WTFMove(encoder));
}

} // namespace WebPage

} // namespace Messages

namespace WebKit {

void WebPage::didReceiveWebPageMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    auto protectedThis = makeRef(*this);
    if (decoder.messageName() == Messages::WebPage::SetInitialFocus::name()) {
        IPC::handleMessage<Messages::WebPage::SetInitialFocus>(decoder, this, &WebPage::setInitialFocus);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetActivityState::name()) {
        IPC::handleMessage<Messages::WebPage::SetActivityState>(decoder, this, &WebPage::setActivityState);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetLayerHostingMode::name()) {
        IPC::handleMessage<Messages::WebPage::SetLayerHostingMode>(decoder, this, &WebPage::setLayerHostingMode);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetBackgroundColor::name()) {
        IPC::handleMessage<Messages::WebPage::SetBackgroundColor>(decoder, this, &WebPage::setBackgroundColor);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::AddConsoleMessage::name()) {
        IPC::handleMessage<Messages::WebPage::AddConsoleMessage>(decoder, this, &WebPage::addConsoleMessage);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SendCSPViolationReport::name()) {
        IPC::handleMessage<Messages::WebPage::SendCSPViolationReport>(decoder, this, &WebPage::sendCSPViolationReport);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::EnqueueSecurityPolicyViolationEvent::name()) {
        IPC::handleMessage<Messages::WebPage::EnqueueSecurityPolicyViolationEvent>(decoder, this, &WebPage::enqueueSecurityPolicyViolationEvent);
        return;
    }
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::SetTopContentInsetFenced::name()) {
        IPC::handleMessage<Messages::WebPage::SetTopContentInsetFenced>(decoder, this, &WebPage::setTopContentInsetFenced);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPage::SetTopContentInset::name()) {
        IPC::handleMessage<Messages::WebPage::SetTopContentInset>(decoder, this, &WebPage::setTopContentInset);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetUnderlayColor::name()) {
        IPC::handleMessage<Messages::WebPage::SetUnderlayColor>(decoder, this, &WebPage::setUnderlayColor);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::ViewWillStartLiveResize::name()) {
        IPC::handleMessage<Messages::WebPage::ViewWillStartLiveResize>(decoder, this, &WebPage::viewWillStartLiveResize);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::ViewWillEndLiveResize::name()) {
        IPC::handleMessage<Messages::WebPage::ViewWillEndLiveResize>(decoder, this, &WebPage::viewWillEndLiveResize);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::ExecuteEditCommandWithCallback::name()) {
        IPC::handleMessage<Messages::WebPage::ExecuteEditCommandWithCallback>(decoder, this, &WebPage::executeEditCommandWithCallback);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::KeyEvent::name()) {
        IPC::handleMessage<Messages::WebPage::KeyEvent>(decoder, this, &WebPage::keyEvent);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::MouseEvent::name()) {
        IPC::handleMessage<Messages::WebPage::MouseEvent>(decoder, this, &WebPage::mouseEvent);
        return;
    }
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::SetViewportConfigurationViewLayoutSize::name()) {
        IPC::handleMessage<Messages::WebPage::SetViewportConfigurationViewLayoutSize>(decoder, this, &WebPage::setViewportConfigurationViewLayoutSize);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::SetMaximumUnobscuredSize::name()) {
        IPC::handleMessage<Messages::WebPage::SetMaximumUnobscuredSize>(decoder, this, &WebPage::setMaximumUnobscuredSize);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::SetDeviceOrientation::name()) {
        IPC::handleMessage<Messages::WebPage::SetDeviceOrientation>(decoder, this, &WebPage::setDeviceOrientation);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::SetOverrideViewportArguments::name()) {
        IPC::handleMessage<Messages::WebPage::SetOverrideViewportArguments>(decoder, this, &WebPage::setOverrideViewportArguments);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::DynamicViewportSizeUpdate::name()) {
        IPC::handleMessage<Messages::WebPage::DynamicViewportSizeUpdate>(decoder, this, &WebPage::dynamicViewportSizeUpdate);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::HandleTap::name()) {
        IPC::handleMessage<Messages::WebPage::HandleTap>(decoder, this, &WebPage::handleTap);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::PotentialTapAtPosition::name()) {
        IPC::handleMessage<Messages::WebPage::PotentialTapAtPosition>(decoder, this, &WebPage::potentialTapAtPosition);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::CommitPotentialTap::name()) {
        IPC::handleMessage<Messages::WebPage::CommitPotentialTap>(decoder, this, &WebPage::commitPotentialTap);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::CancelPotentialTap::name()) {
        IPC::handleMessage<Messages::WebPage::CancelPotentialTap>(decoder, this, &WebPage::cancelPotentialTap);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::TapHighlightAtPosition::name()) {
        IPC::handleMessage<Messages::WebPage::TapHighlightAtPosition>(decoder, this, &WebPage::tapHighlightAtPosition);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::DidRecognizeLongPress::name()) {
        IPC::handleMessage<Messages::WebPage::DidRecognizeLongPress>(decoder, this, &WebPage::didRecognizeLongPress);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::HandleDoubleTapForDoubleClickAtPoint::name()) {
        IPC::handleMessage<Messages::WebPage::HandleDoubleTapForDoubleClickAtPoint>(decoder, this, &WebPage::handleDoubleTapForDoubleClickAtPoint);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::InspectorNodeSearchMovedToPosition::name()) {
        IPC::handleMessage<Messages::WebPage::InspectorNodeSearchMovedToPosition>(decoder, this, &WebPage::inspectorNodeSearchMovedToPosition);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::InspectorNodeSearchEndedAtPosition::name()) {
        IPC::handleMessage<Messages::WebPage::InspectorNodeSearchEndedAtPosition>(decoder, this, &WebPage::inspectorNodeSearchEndedAtPosition);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::BlurFocusedElement::name()) {
        IPC::handleMessage<Messages::WebPage::BlurFocusedElement>(decoder, this, &WebPage::blurFocusedElement);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::SelectWithGesture::name()) {
        IPC::handleMessage<Messages::WebPage::SelectWithGesture>(decoder, this, &WebPage::selectWithGesture);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::UpdateSelectionWithTouches::name()) {
        IPC::handleMessage<Messages::WebPage::UpdateSelectionWithTouches>(decoder, this, &WebPage::updateSelectionWithTouches);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::SelectWithTwoTouches::name()) {
        IPC::handleMessage<Messages::WebPage::SelectWithTwoTouches>(decoder, this, &WebPage::selectWithTwoTouches);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::ExtendSelection::name()) {
        IPC::handleMessage<Messages::WebPage::ExtendSelection>(decoder, this, &WebPage::extendSelection);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::SelectWordBackward::name()) {
        IPC::handleMessage<Messages::WebPage::SelectWordBackward>(decoder, this, &WebPage::selectWordBackward);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::MoveSelectionByOffset::name()) {
        IPC::handleMessage<Messages::WebPage::MoveSelectionByOffset>(decoder, this, &WebPage::moveSelectionByOffset);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::SelectTextWithGranularityAtPoint::name()) {
        IPC::handleMessage<Messages::WebPage::SelectTextWithGranularityAtPoint>(decoder, this, &WebPage::selectTextWithGranularityAtPoint);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::SelectPositionAtBoundaryWithDirection::name()) {
        IPC::handleMessage<Messages::WebPage::SelectPositionAtBoundaryWithDirection>(decoder, this, &WebPage::selectPositionAtBoundaryWithDirection);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::MoveSelectionAtBoundaryWithDirection::name()) {
        IPC::handleMessage<Messages::WebPage::MoveSelectionAtBoundaryWithDirection>(decoder, this, &WebPage::moveSelectionAtBoundaryWithDirection);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::SelectPositionAtPoint::name()) {
        IPC::handleMessage<Messages::WebPage::SelectPositionAtPoint>(decoder, this, &WebPage::selectPositionAtPoint);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::BeginSelectionInDirection::name()) {
        IPC::handleMessage<Messages::WebPage::BeginSelectionInDirection>(decoder, this, &WebPage::beginSelectionInDirection);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::UpdateSelectionWithExtentPoint::name()) {
        IPC::handleMessage<Messages::WebPage::UpdateSelectionWithExtentPoint>(decoder, this, &WebPage::updateSelectionWithExtentPoint);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::UpdateSelectionWithExtentPointAndBoundary::name()) {
        IPC::handleMessage<Messages::WebPage::UpdateSelectionWithExtentPointAndBoundary>(decoder, this, &WebPage::updateSelectionWithExtentPointAndBoundary);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::RequestDictationContext::name()) {
        IPC::handleMessage<Messages::WebPage::RequestDictationContext>(decoder, this, &WebPage::requestDictationContext);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::ReplaceDictatedText::name()) {
        IPC::handleMessage<Messages::WebPage::ReplaceDictatedText>(decoder, this, &WebPage::replaceDictatedText);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::ReplaceSelectedText::name()) {
        IPC::handleMessage<Messages::WebPage::ReplaceSelectedText>(decoder, this, &WebPage::replaceSelectedText);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::RequestAutocorrectionData::name()) {
        IPC::handleMessageAsync<Messages::WebPage::RequestAutocorrectionData>(connection, decoder, this, &WebPage::requestAutocorrectionData);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::ApplyAutocorrection::name()) {
        IPC::handleMessage<Messages::WebPage::ApplyAutocorrection>(decoder, this, &WebPage::applyAutocorrection);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::RequestAutocorrectionContext::name()) {
        IPC::handleMessage<Messages::WebPage::RequestAutocorrectionContext>(decoder, this, &WebPage::requestAutocorrectionContext);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::RequestEvasionRectsAboveSelection::name()) {
        IPC::handleMessageAsync<Messages::WebPage::RequestEvasionRectsAboveSelection>(connection, decoder, this, &WebPage::requestEvasionRectsAboveSelection);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::RequestPositionInformation::name()) {
        IPC::handleMessage<Messages::WebPage::RequestPositionInformation>(decoder, this, &WebPage::requestPositionInformation);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::StartInteractionWithElementContextOrPosition::name()) {
        IPC::handleMessage<Messages::WebPage::StartInteractionWithElementContextOrPosition>(decoder, this, &WebPage::startInteractionWithElementContextOrPosition);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::StopInteraction::name()) {
        IPC::handleMessage<Messages::WebPage::StopInteraction>(decoder, this, &WebPage::stopInteraction);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::PerformActionOnElement::name()) {
        IPC::handleMessage<Messages::WebPage::PerformActionOnElement>(decoder, this, &WebPage::performActionOnElement);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::FocusNextFocusedElement::name()) {
        IPC::handleMessage<Messages::WebPage::FocusNextFocusedElement>(decoder, this, &WebPage::focusNextFocusedElement);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::SetFocusedElementValue::name()) {
        IPC::handleMessage<Messages::WebPage::SetFocusedElementValue>(decoder, this, &WebPage::setFocusedElementValue);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::AutofillLoginCredentials::name()) {
        IPC::handleMessage<Messages::WebPage::AutofillLoginCredentials>(decoder, this, &WebPage::autofillLoginCredentials);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::SetFocusedElementValueAsNumber::name()) {
        IPC::handleMessage<Messages::WebPage::SetFocusedElementValueAsNumber>(decoder, this, &WebPage::setFocusedElementValueAsNumber);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::SetFocusedElementSelectedIndex::name()) {
        IPC::handleMessage<Messages::WebPage::SetFocusedElementSelectedIndex>(decoder, this, &WebPage::setFocusedElementSelectedIndex);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::ApplicationWillResignActive::name()) {
        IPC::handleMessage<Messages::WebPage::ApplicationWillResignActive>(decoder, this, &WebPage::applicationWillResignActive);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::ApplicationDidEnterBackground::name()) {
        IPC::handleMessage<Messages::WebPage::ApplicationDidEnterBackground>(decoder, this, &WebPage::applicationDidEnterBackground);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::ApplicationDidFinishSnapshottingAfterEnteringBackground::name()) {
        IPC::handleMessage<Messages::WebPage::ApplicationDidFinishSnapshottingAfterEnteringBackground>(decoder, this, &WebPage::applicationDidFinishSnapshottingAfterEnteringBackground);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::ApplicationWillEnterForeground::name()) {
        IPC::handleMessage<Messages::WebPage::ApplicationWillEnterForeground>(decoder, this, &WebPage::applicationWillEnterForeground);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::ApplicationDidBecomeActive::name()) {
        IPC::handleMessage<Messages::WebPage::ApplicationDidBecomeActive>(decoder, this, &WebPage::applicationDidBecomeActive);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::ApplicationDidEnterBackgroundForMedia::name()) {
        IPC::handleMessage<Messages::WebPage::ApplicationDidEnterBackgroundForMedia>(decoder, this, &WebPage::applicationDidEnterBackgroundForMedia);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::ApplicationWillEnterForegroundForMedia::name()) {
        IPC::handleMessage<Messages::WebPage::ApplicationWillEnterForegroundForMedia>(decoder, this, &WebPage::applicationWillEnterForegroundForMedia);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::ContentSizeCategoryDidChange::name()) {
        IPC::handleMessage<Messages::WebPage::ContentSizeCategoryDidChange>(decoder, this, &WebPage::contentSizeCategoryDidChange);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::GetSelectionContext::name()) {
        IPC::handleMessage<Messages::WebPage::GetSelectionContext>(decoder, this, &WebPage::getSelectionContext);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::SetAllowsMediaDocumentInlinePlayback::name()) {
        IPC::handleMessage<Messages::WebPage::SetAllowsMediaDocumentInlinePlayback>(decoder, this, &WebPage::setAllowsMediaDocumentInlinePlayback);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::HandleTwoFingerTapAtPoint::name()) {
        IPC::handleMessage<Messages::WebPage::HandleTwoFingerTapAtPoint>(decoder, this, &WebPage::handleTwoFingerTapAtPoint);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::HandleStylusSingleTapAtPoint::name()) {
        IPC::handleMessage<Messages::WebPage::HandleStylusSingleTapAtPoint>(decoder, this, &WebPage::handleStylusSingleTapAtPoint);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::SetForceAlwaysUserScalable::name()) {
        IPC::handleMessage<Messages::WebPage::SetForceAlwaysUserScalable>(decoder, this, &WebPage::setForceAlwaysUserScalable);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::GetRectsForGranularityWithSelectionOffset::name()) {
        IPC::handleMessage<Messages::WebPage::GetRectsForGranularityWithSelectionOffset>(decoder, this, &WebPage::getRectsForGranularityWithSelectionOffset);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::GetRectsAtSelectionOffsetWithText::name()) {
        IPC::handleMessage<Messages::WebPage::GetRectsAtSelectionOffsetWithText>(decoder, this, &WebPage::getRectsAtSelectionOffsetWithText);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::StoreSelectionForAccessibility::name()) {
        IPC::handleMessage<Messages::WebPage::StoreSelectionForAccessibility>(decoder, this, &WebPage::storeSelectionForAccessibility);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::StartAutoscrollAtPosition::name()) {
        IPC::handleMessage<Messages::WebPage::StartAutoscrollAtPosition>(decoder, this, &WebPage::startAutoscrollAtPosition);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::CancelAutoscroll::name()) {
        IPC::handleMessage<Messages::WebPage::CancelAutoscroll>(decoder, this, &WebPage::cancelAutoscroll);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::RequestFocusedElementInformation::name()) {
        IPC::handleMessage<Messages::WebPage::RequestFocusedElementInformation>(decoder, this, &WebPage::requestFocusedElementInformation);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::HardwareKeyboardAvailabilityChanged::name()) {
        IPC::handleMessage<Messages::WebPage::HardwareKeyboardAvailabilityChanged>(decoder, this, &WebPage::hardwareKeyboardAvailabilityChanged);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::SetIsShowingInputViewForFocusedElement::name()) {
        IPC::handleMessage<Messages::WebPage::SetIsShowingInputViewForFocusedElement>(decoder, this, &WebPage::setIsShowingInputViewForFocusedElement);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::UpdateSelectionWithDelta::name()) {
        IPC::handleMessageAsync<Messages::WebPage::UpdateSelectionWithDelta>(connection, decoder, this, &WebPage::updateSelectionWithDelta);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::RequestDocumentEditingContext::name()) {
        IPC::handleMessageAsync<Messages::WebPage::RequestDocumentEditingContext>(connection, decoder, this, &WebPage::requestDocumentEditingContext);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::GenerateSyntheticEditingCommand::name()) {
        IPC::handleMessage<Messages::WebPage::GenerateSyntheticEditingCommand>(decoder, this, &WebPage::generateSyntheticEditingCommand);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPage::SetControlledByAutomation::name()) {
        IPC::handleMessage<Messages::WebPage::SetControlledByAutomation>(decoder, this, &WebPage::setControlledByAutomation);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::ConnectInspector::name()) {
        IPC::handleMessage<Messages::WebPage::ConnectInspector>(decoder, this, &WebPage::connectInspector);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::DisconnectInspector::name()) {
        IPC::handleMessage<Messages::WebPage::DisconnectInspector>(decoder, this, &WebPage::disconnectInspector);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SendMessageToTargetBackend::name()) {
        IPC::handleMessage<Messages::WebPage::SendMessageToTargetBackend>(decoder, this, &WebPage::sendMessageToTargetBackend);
        return;
    }
#if ENABLE(REMOTE_INSPECTOR)
    if (decoder.messageName() == Messages::WebPage::SetIndicating::name()) {
        IPC::handleMessage<Messages::WebPage::SetIndicating>(decoder, this, &WebPage::setIndicating);
        return;
    }
#endif
#if ENABLE(IOS_TOUCH_EVENTS)
    if (decoder.messageName() == Messages::WebPage::ResetPotentialTapSecurityOrigin::name()) {
        IPC::handleMessage<Messages::WebPage::ResetPotentialTapSecurityOrigin>(decoder, this, &WebPage::resetPotentialTapSecurityOrigin);
        return;
    }
#endif
#if !ENABLE(IOS_TOUCH_EVENTS) && ENABLE(TOUCH_EVENTS)
    if (decoder.messageName() == Messages::WebPage::TouchEvent::name()) {
        IPC::handleMessage<Messages::WebPage::TouchEvent>(decoder, this, &WebPage::touchEvent);
        return;
    }
#endif
#if ENABLE(POINTER_EVENTS)
    if (decoder.messageName() == Messages::WebPage::CancelPointer::name()) {
        IPC::handleMessage<Messages::WebPage::CancelPointer>(decoder, this, &WebPage::cancelPointer);
        return;
    }
#endif
#if ENABLE(POINTER_EVENTS)
    if (decoder.messageName() == Messages::WebPage::TouchWithIdentifierWasRemoved::name()) {
        IPC::handleMessage<Messages::WebPage::TouchWithIdentifierWasRemoved>(decoder, this, &WebPage::touchWithIdentifierWasRemoved);
        return;
    }
#endif
#if ENABLE(INPUT_TYPE_COLOR)
    if (decoder.messageName() == Messages::WebPage::DidEndColorPicker::name()) {
        IPC::handleMessage<Messages::WebPage::DidEndColorPicker>(decoder, this, &WebPage::didEndColorPicker);
        return;
    }
#endif
#if ENABLE(INPUT_TYPE_COLOR)
    if (decoder.messageName() == Messages::WebPage::DidChooseColor::name()) {
        IPC::handleMessage<Messages::WebPage::DidChooseColor>(decoder, this, &WebPage::didChooseColor);
        return;
    }
#endif
#if ENABLE(DATALIST_ELEMENT)
    if (decoder.messageName() == Messages::WebPage::DidSelectDataListOption::name()) {
        IPC::handleMessage<Messages::WebPage::DidSelectDataListOption>(decoder, this, &WebPage::didSelectDataListOption);
        return;
    }
#endif
#if ENABLE(DATALIST_ELEMENT)
    if (decoder.messageName() == Messages::WebPage::DidCloseSuggestions::name()) {
        IPC::handleMessage<Messages::WebPage::DidCloseSuggestions>(decoder, this, &WebPage::didCloseSuggestions);
        return;
    }
#endif
#if ENABLE(CONTEXT_MENUS)
    if (decoder.messageName() == Messages::WebPage::ContextMenuHidden::name()) {
        IPC::handleMessage<Messages::WebPage::ContextMenuHidden>(decoder, this, &WebPage::contextMenuHidden);
        return;
    }
#endif
#if ENABLE(CONTEXT_MENUS)
    if (decoder.messageName() == Messages::WebPage::ContextMenuForKeyEvent::name()) {
        IPC::handleMessage<Messages::WebPage::ContextMenuForKeyEvent>(decoder, this, &WebPage::contextMenuForKeyEvent);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPage::ScrollBy::name()) {
        IPC::handleMessage<Messages::WebPage::ScrollBy>(decoder, this, &WebPage::scrollBy);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::CenterSelectionInVisibleArea::name()) {
        IPC::handleMessage<Messages::WebPage::CenterSelectionInVisibleArea>(decoder, this, &WebPage::centerSelectionInVisibleArea);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::GoToBackForwardItem::name()) {
        IPC::handleMessage<Messages::WebPage::GoToBackForwardItem>(decoder, this, &WebPage::goToBackForwardItem);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::TryRestoreScrollPosition::name()) {
        IPC::handleMessage<Messages::WebPage::TryRestoreScrollPosition>(decoder, this, &WebPage::tryRestoreScrollPosition);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::LoadURLInFrame::name()) {
        IPC::handleMessage<Messages::WebPage::LoadURLInFrame>(decoder, this, &WebPage::loadURLInFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::LoadDataInFrame::name()) {
        IPC::handleMessage<Messages::WebPage::LoadDataInFrame>(decoder, this, &WebPage::loadDataInFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::LoadRequest::name()) {
        IPC::handleMessage<Messages::WebPage::LoadRequest>(decoder, this, &WebPage::loadRequest);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::LoadRequestWaitingForProcessLaunch::name()) {
        IPC::handleMessage<Messages::WebPage::LoadRequestWaitingForProcessLaunch>(decoder, this, &WebPage::loadRequestWaitingForProcessLaunch);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::LoadData::name()) {
        IPC::handleMessage<Messages::WebPage::LoadData>(decoder, this, &WebPage::loadData);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::LoadAlternateHTML::name()) {
        IPC::handleMessage<Messages::WebPage::LoadAlternateHTML>(decoder, this, &WebPage::loadAlternateHTML);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::NavigateToPDFLinkWithSimulatedClick::name()) {
        IPC::handleMessage<Messages::WebPage::NavigateToPDFLinkWithSimulatedClick>(decoder, this, &WebPage::navigateToPDFLinkWithSimulatedClick);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::Reload::name()) {
        IPC::handleMessage<Messages::WebPage::Reload>(decoder, this, &WebPage::reload);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::StopLoading::name()) {
        IPC::handleMessage<Messages::WebPage::StopLoading>(decoder, this, &WebPage::stopLoading);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::StopLoadingFrame::name()) {
        IPC::handleMessage<Messages::WebPage::StopLoadingFrame>(decoder, this, &WebPage::stopLoadingFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::RestoreSession::name()) {
        IPC::handleMessage<Messages::WebPage::RestoreSession>(decoder, this, &WebPage::restoreSession);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::UpdateBackForwardListForReattach::name()) {
        IPC::handleMessage<Messages::WebPage::UpdateBackForwardListForReattach>(decoder, this, &WebPage::updateBackForwardListForReattach);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetCurrentHistoryItemForReattach::name()) {
        IPC::handleMessage<Messages::WebPage::SetCurrentHistoryItemForReattach>(decoder, this, &WebPage::setCurrentHistoryItemForReattach);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::DidRemoveBackForwardItem::name()) {
        IPC::handleMessage<Messages::WebPage::DidRemoveBackForwardItem>(decoder, this, &WebPage::didRemoveBackForwardItem);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::UpdateWebsitePolicies::name()) {
        IPC::handleMessage<Messages::WebPage::UpdateWebsitePolicies>(decoder, this, &WebPage::updateWebsitePolicies);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::DidReceivePolicyDecision::name()) {
        IPC::handleMessage<Messages::WebPage::DidReceivePolicyDecision>(decoder, this, &WebPage::didReceivePolicyDecision);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::ContinueWillSubmitForm::name()) {
        IPC::handleMessage<Messages::WebPage::ContinueWillSubmitForm>(decoder, this, &WebPage::continueWillSubmitForm);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::ClearSelection::name()) {
        IPC::handleMessage<Messages::WebPage::ClearSelection>(decoder, this, &WebPage::clearSelection);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::RestoreSelectionInFocusedEditableElement::name()) {
        IPC::handleMessage<Messages::WebPage::RestoreSelectionInFocusedEditableElement>(decoder, this, &WebPage::restoreSelectionInFocusedEditableElement);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::GetContentsAsString::name()) {
        IPC::handleMessage<Messages::WebPage::GetContentsAsString>(decoder, this, &WebPage::getContentsAsString);
        return;
    }
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::GetContentsAsAttributedString::name()) {
        IPC::handleMessageAsync<Messages::WebPage::GetContentsAsAttributedString>(connection, decoder, this, &WebPage::getContentsAsAttributedString);
        return;
    }
#endif
#if ENABLE(MHTML)
    if (decoder.messageName() == Messages::WebPage::GetContentsAsMHTMLData::name()) {
        IPC::handleMessage<Messages::WebPage::GetContentsAsMHTMLData>(decoder, this, &WebPage::getContentsAsMHTMLData);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPage::GetMainResourceDataOfFrame::name()) {
        IPC::handleMessage<Messages::WebPage::GetMainResourceDataOfFrame>(decoder, this, &WebPage::getMainResourceDataOfFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::GetResourceDataFromFrame::name()) {
        IPC::handleMessage<Messages::WebPage::GetResourceDataFromFrame>(decoder, this, &WebPage::getResourceDataFromFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::GetRenderTreeExternalRepresentation::name()) {
        IPC::handleMessage<Messages::WebPage::GetRenderTreeExternalRepresentation>(decoder, this, &WebPage::getRenderTreeExternalRepresentation);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::GetSelectionOrContentsAsString::name()) {
        IPC::handleMessage<Messages::WebPage::GetSelectionOrContentsAsString>(decoder, this, &WebPage::getSelectionOrContentsAsString);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::GetSelectionAsWebArchiveData::name()) {
        IPC::handleMessage<Messages::WebPage::GetSelectionAsWebArchiveData>(decoder, this, &WebPage::getSelectionAsWebArchiveData);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::GetSourceForFrame::name()) {
        IPC::handleMessage<Messages::WebPage::GetSourceForFrame>(decoder, this, &WebPage::getSourceForFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::GetWebArchiveOfFrame::name()) {
        IPC::handleMessage<Messages::WebPage::GetWebArchiveOfFrame>(decoder, this, &WebPage::getWebArchiveOfFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::RunJavaScriptInMainFrameScriptWorld::name()) {
        IPC::handleMessage<Messages::WebPage::RunJavaScriptInMainFrameScriptWorld>(decoder, this, &WebPage::runJavaScriptInMainFrameScriptWorld);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::RunJavaScriptInFrame::name()) {
        IPC::handleMessage<Messages::WebPage::RunJavaScriptInFrame>(decoder, this, &WebPage::runJavaScriptInFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::ForceRepaint::name()) {
        IPC::handleMessage<Messages::WebPage::ForceRepaint>(decoder, this, &WebPage::forceRepaint);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SelectAll::name()) {
        IPC::handleMessage<Messages::WebPage::SelectAll>(decoder, this, &WebPage::selectAll);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::ScheduleFullEditorStateUpdate::name()) {
        IPC::handleMessage<Messages::WebPage::ScheduleFullEditorStateUpdate>(decoder, this, &WebPage::scheduleFullEditorStateUpdate);
        return;
    }
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::PerformDictionaryLookupOfCurrentSelection::name()) {
        IPC::handleMessage<Messages::WebPage::PerformDictionaryLookupOfCurrentSelection>(decoder, this, &WebPage::performDictionaryLookupOfCurrentSelection);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::PerformDictionaryLookupAtLocation::name()) {
        IPC::handleMessage<Messages::WebPage::PerformDictionaryLookupAtLocation>(decoder, this, &WebPage::performDictionaryLookupAtLocation);
        return;
    }
#endif
#if ENABLE(DATA_DETECTION)
    if (decoder.messageName() == Messages::WebPage::DetectDataInAllFrames::name()) {
        IPC::handleMessageAsync<Messages::WebPage::DetectDataInAllFrames>(connection, decoder, this, &WebPage::detectDataInAllFrames);
        return;
    }
#endif
#if ENABLE(DATA_DETECTION)
    if (decoder.messageName() == Messages::WebPage::RemoveDataDetectedLinks::name()) {
        IPC::handleMessageAsync<Messages::WebPage::RemoveDataDetectedLinks>(connection, decoder, this, &WebPage::removeDataDetectedLinks);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPage::ChangeFont::name()) {
        IPC::handleMessage<Messages::WebPage::ChangeFont>(decoder, this, &WebPage::changeFont);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::ChangeFontAttributes::name()) {
        IPC::handleMessage<Messages::WebPage::ChangeFontAttributes>(decoder, this, &WebPage::changeFontAttributes);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::PreferencesDidChange::name()) {
        IPC::handleMessage<Messages::WebPage::PreferencesDidChange>(decoder, this, &WebPage::preferencesDidChange);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetUserAgent::name()) {
        IPC::handleMessage<Messages::WebPage::SetUserAgent>(decoder, this, &WebPage::setUserAgent);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetCustomTextEncodingName::name()) {
        IPC::handleMessage<Messages::WebPage::SetCustomTextEncodingName>(decoder, this, &WebPage::setCustomTextEncodingName);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SuspendActiveDOMObjectsAndAnimations::name()) {
        IPC::handleMessage<Messages::WebPage::SuspendActiveDOMObjectsAndAnimations>(decoder, this, &WebPage::suspendActiveDOMObjectsAndAnimations);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::ResumeActiveDOMObjectsAndAnimations::name()) {
        IPC::handleMessage<Messages::WebPage::ResumeActiveDOMObjectsAndAnimations>(decoder, this, &WebPage::resumeActiveDOMObjectsAndAnimations);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::Close::name()) {
        IPC::handleMessage<Messages::WebPage::Close>(decoder, this, &WebPage::close);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::TryClose::name()) {
        IPC::handleMessageAsync<Messages::WebPage::TryClose>(connection, decoder, this, &WebPage::tryClose);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetEditable::name()) {
        IPC::handleMessage<Messages::WebPage::SetEditable>(decoder, this, &WebPage::setEditable);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::ValidateCommand::name()) {
        IPC::handleMessage<Messages::WebPage::ValidateCommand>(decoder, this, &WebPage::validateCommand);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::ExecuteEditCommand::name()) {
        IPC::handleMessage<Messages::WebPage::ExecuteEditCommand>(decoder, this, &WebPage::executeEditCommand);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::IncreaseListLevel::name()) {
        IPC::handleMessage<Messages::WebPage::IncreaseListLevel>(decoder, this, &WebPage::increaseListLevel);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::DecreaseListLevel::name()) {
        IPC::handleMessage<Messages::WebPage::DecreaseListLevel>(decoder, this, &WebPage::decreaseListLevel);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::ChangeListType::name()) {
        IPC::handleMessage<Messages::WebPage::ChangeListType>(decoder, this, &WebPage::changeListType);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetBaseWritingDirection::name()) {
        IPC::handleMessage<Messages::WebPage::SetBaseWritingDirection>(decoder, this, &WebPage::setBaseWritingDirection);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetNeedsFontAttributes::name()) {
        IPC::handleMessage<Messages::WebPage::SetNeedsFontAttributes>(decoder, this, &WebPage::setNeedsFontAttributes);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::RequestFontAttributesAtSelectionStart::name()) {
        IPC::handleMessage<Messages::WebPage::RequestFontAttributesAtSelectionStart>(decoder, this, &WebPage::requestFontAttributesAtSelectionStart);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::DidRemoveEditCommand::name()) {
        IPC::handleMessage<Messages::WebPage::DidRemoveEditCommand>(decoder, this, &WebPage::didRemoveEditCommand);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::ReapplyEditCommand::name()) {
        IPC::handleMessage<Messages::WebPage::ReapplyEditCommand>(decoder, this, &WebPage::reapplyEditCommand);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::UnapplyEditCommand::name()) {
        IPC::handleMessage<Messages::WebPage::UnapplyEditCommand>(decoder, this, &WebPage::unapplyEditCommand);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetPageAndTextZoomFactors::name()) {
        IPC::handleMessage<Messages::WebPage::SetPageAndTextZoomFactors>(decoder, this, &WebPage::setPageAndTextZoomFactors);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetPageZoomFactor::name()) {
        IPC::handleMessage<Messages::WebPage::SetPageZoomFactor>(decoder, this, &WebPage::setPageZoomFactor);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetTextZoomFactor::name()) {
        IPC::handleMessage<Messages::WebPage::SetTextZoomFactor>(decoder, this, &WebPage::setTextZoomFactor);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::WindowScreenDidChange::name()) {
        IPC::handleMessage<Messages::WebPage::WindowScreenDidChange>(decoder, this, &WebPage::windowScreenDidChange);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::AccessibilitySettingsDidChange::name()) {
        IPC::handleMessage<Messages::WebPage::AccessibilitySettingsDidChange>(decoder, this, &WebPage::accessibilitySettingsDidChange);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::ScalePage::name()) {
        IPC::handleMessage<Messages::WebPage::ScalePage>(decoder, this, &WebPage::scalePage);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::ScalePageInViewCoordinates::name()) {
        IPC::handleMessage<Messages::WebPage::ScalePageInViewCoordinates>(decoder, this, &WebPage::scalePageInViewCoordinates);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::ScaleView::name()) {
        IPC::handleMessage<Messages::WebPage::ScaleView>(decoder, this, &WebPage::scaleView);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetUseFixedLayout::name()) {
        IPC::handleMessage<Messages::WebPage::SetUseFixedLayout>(decoder, this, &WebPage::setUseFixedLayout);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetFixedLayoutSize::name()) {
        IPC::handleMessage<Messages::WebPage::SetFixedLayoutSize>(decoder, this, &WebPage::setFixedLayoutSize);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::ListenForLayoutMilestones::name()) {
        IPC::handleMessage<Messages::WebPage::ListenForLayoutMilestones>(decoder, this, &WebPage::listenForLayoutMilestones);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetSuppressScrollbarAnimations::name()) {
        IPC::handleMessage<Messages::WebPage::SetSuppressScrollbarAnimations>(decoder, this, &WebPage::setSuppressScrollbarAnimations);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetEnableVerticalRubberBanding::name()) {
        IPC::handleMessage<Messages::WebPage::SetEnableVerticalRubberBanding>(decoder, this, &WebPage::setEnableVerticalRubberBanding);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetEnableHorizontalRubberBanding::name()) {
        IPC::handleMessage<Messages::WebPage::SetEnableHorizontalRubberBanding>(decoder, this, &WebPage::setEnableHorizontalRubberBanding);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetBackgroundExtendsBeyondPage::name()) {
        IPC::handleMessage<Messages::WebPage::SetBackgroundExtendsBeyondPage>(decoder, this, &WebPage::setBackgroundExtendsBeyondPage);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetPaginationMode::name()) {
        IPC::handleMessage<Messages::WebPage::SetPaginationMode>(decoder, this, &WebPage::setPaginationMode);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetPaginationBehavesLikeColumns::name()) {
        IPC::handleMessage<Messages::WebPage::SetPaginationBehavesLikeColumns>(decoder, this, &WebPage::setPaginationBehavesLikeColumns);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetPageLength::name()) {
        IPC::handleMessage<Messages::WebPage::SetPageLength>(decoder, this, &WebPage::setPageLength);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetGapBetweenPages::name()) {
        IPC::handleMessage<Messages::WebPage::SetGapBetweenPages>(decoder, this, &WebPage::setGapBetweenPages);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetPaginationLineGridEnabled::name()) {
        IPC::handleMessage<Messages::WebPage::SetPaginationLineGridEnabled>(decoder, this, &WebPage::setPaginationLineGridEnabled);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::PostInjectedBundleMessage::name()) {
        IPC::handleMessage<Messages::WebPage::PostInjectedBundleMessage>(decoder, this, &WebPage::postInjectedBundleMessage);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::FindString::name()) {
        IPC::handleMessage<Messages::WebPage::FindString>(decoder, this, &WebPage::findString);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::FindStringMatches::name()) {
        IPC::handleMessage<Messages::WebPage::FindStringMatches>(decoder, this, &WebPage::findStringMatches);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::GetImageForFindMatch::name()) {
        IPC::handleMessage<Messages::WebPage::GetImageForFindMatch>(decoder, this, &WebPage::getImageForFindMatch);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SelectFindMatch::name()) {
        IPC::handleMessage<Messages::WebPage::SelectFindMatch>(decoder, this, &WebPage::selectFindMatch);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::IndicateFindMatch::name()) {
        IPC::handleMessage<Messages::WebPage::IndicateFindMatch>(decoder, this, &WebPage::indicateFindMatch);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::HideFindUI::name()) {
        IPC::handleMessage<Messages::WebPage::HideFindUI>(decoder, this, &WebPage::hideFindUI);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::CountStringMatches::name()) {
        IPC::handleMessage<Messages::WebPage::CountStringMatches>(decoder, this, &WebPage::countStringMatches);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::ReplaceMatches::name()) {
        IPC::handleMessage<Messages::WebPage::ReplaceMatches>(decoder, this, &WebPage::replaceMatches);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::AddMIMETypeWithCustomContentProvider::name()) {
        IPC::handleMessage<Messages::WebPage::AddMIMETypeWithCustomContentProvider>(decoder, this, &WebPage::addMIMETypeWithCustomContentProvider);
        return;
    }
#if PLATFORM(GTK) && ENABLE(DRAG_SUPPORT)
    if (decoder.messageName() == Messages::WebPage::PerformDragControllerAction::name()) {
        IPC::handleMessage<Messages::WebPage::PerformDragControllerAction>(decoder, this, &WebPage::performDragControllerAction);
        return;
    }
#endif
#if !PLATFORM(GTK) && ENABLE(DRAG_SUPPORT)
    if (decoder.messageName() == Messages::WebPage::PerformDragControllerAction::name()) {
        IPC::handleMessage<Messages::WebPage::PerformDragControllerAction>(decoder, this, &WebPage::performDragControllerAction);
        return;
    }
#endif
#if ENABLE(DRAG_SUPPORT)
    if (decoder.messageName() == Messages::WebPage::DidStartDrag::name()) {
        IPC::handleMessage<Messages::WebPage::DidStartDrag>(decoder, this, &WebPage::didStartDrag);
        return;
    }
#endif
#if ENABLE(DRAG_SUPPORT)
    if (decoder.messageName() == Messages::WebPage::DragEnded::name()) {
        IPC::handleMessage<Messages::WebPage::DragEnded>(decoder, this, &WebPage::dragEnded);
        return;
    }
#endif
#if ENABLE(DRAG_SUPPORT)
    if (decoder.messageName() == Messages::WebPage::DragCancelled::name()) {
        IPC::handleMessage<Messages::WebPage::DragCancelled>(decoder, this, &WebPage::dragCancelled);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(DRAG_SUPPORT)
    if (decoder.messageName() == Messages::WebPage::RequestDragStart::name()) {
        IPC::handleMessage<Messages::WebPage::RequestDragStart>(decoder, this, &WebPage::requestDragStart);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(DRAG_SUPPORT)
    if (decoder.messageName() == Messages::WebPage::RequestAdditionalItemsForDragSession::name()) {
        IPC::handleMessage<Messages::WebPage::RequestAdditionalItemsForDragSession>(decoder, this, &WebPage::requestAdditionalItemsForDragSession);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(DRAG_SUPPORT)
    if (decoder.messageName() == Messages::WebPage::InsertDroppedImagePlaceholders::name()) {
        IPC::handleMessageAsync<Messages::WebPage::InsertDroppedImagePlaceholders>(connection, decoder, this, &WebPage::insertDroppedImagePlaceholders);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(DRAG_SUPPORT)
    if (decoder.messageName() == Messages::WebPage::DidConcludeDrop::name()) {
        IPC::handleMessage<Messages::WebPage::DidConcludeDrop>(decoder, this, &WebPage::didConcludeDrop);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPage::DidChangeSelectedIndexForActivePopupMenu::name()) {
        IPC::handleMessage<Messages::WebPage::DidChangeSelectedIndexForActivePopupMenu>(decoder, this, &WebPage::didChangeSelectedIndexForActivePopupMenu);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetTextForActivePopupMenu::name()) {
        IPC::handleMessage<Messages::WebPage::SetTextForActivePopupMenu>(decoder, this, &WebPage::setTextForActivePopupMenu);
        return;
    }
#if PLATFORM(GTK)
    if (decoder.messageName() == Messages::WebPage::FailedToShowPopupMenu::name()) {
        IPC::handleMessage<Messages::WebPage::FailedToShowPopupMenu>(decoder, this, &WebPage::failedToShowPopupMenu);
        return;
    }
#endif
#if ENABLE(CONTEXT_MENUS)
    if (decoder.messageName() == Messages::WebPage::DidSelectItemFromActiveContextMenu::name()) {
        IPC::handleMessage<Messages::WebPage::DidSelectItemFromActiveContextMenu>(decoder, this, &WebPage::didSelectItemFromActiveContextMenu);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::DidChooseFilesForOpenPanelWithDisplayStringAndIcon::name()) {
        IPC::handleMessage<Messages::WebPage::DidChooseFilesForOpenPanelWithDisplayStringAndIcon>(decoder, this, &WebPage::didChooseFilesForOpenPanelWithDisplayStringAndIcon);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPage::DidChooseFilesForOpenPanel::name()) {
        IPC::handleMessage<Messages::WebPage::DidChooseFilesForOpenPanel>(decoder, this, &WebPage::didChooseFilesForOpenPanel);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::DidCancelForOpenPanel::name()) {
        IPC::handleMessage<Messages::WebPage::DidCancelForOpenPanel>(decoder, this, &WebPage::didCancelForOpenPanel);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::DidCompleteShareSheet::name()) {
        IPC::handleMessage<Messages::WebPage::DidCompleteShareSheet>(decoder, this, &WebPage::didCompleteShareSheet);
        return;
    }
#if ENABLE(SANDBOX_EXTENSIONS)
    if (decoder.messageName() == Messages::WebPage::ExtendSandboxForFilesFromOpenPanel::name()) {
        IPC::handleMessage<Messages::WebPage::ExtendSandboxForFilesFromOpenPanel>(decoder, this, &WebPage::extendSandboxForFilesFromOpenPanel);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPage::AdvanceToNextMisspelling::name()) {
        IPC::handleMessage<Messages::WebPage::AdvanceToNextMisspelling>(decoder, this, &WebPage::advanceToNextMisspelling);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::ChangeSpellingToWord::name()) {
        IPC::handleMessage<Messages::WebPage::ChangeSpellingToWord>(decoder, this, &WebPage::changeSpellingToWord);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::DidFinishCheckingText::name()) {
        IPC::handleMessage<Messages::WebPage::DidFinishCheckingText>(decoder, this, &WebPage::didFinishCheckingText);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::DidCancelCheckingText::name()) {
        IPC::handleMessage<Messages::WebPage::DidCancelCheckingText>(decoder, this, &WebPage::didCancelCheckingText);
        return;
    }
#if USE(APPKIT)
    if (decoder.messageName() == Messages::WebPage::UppercaseWord::name()) {
        IPC::handleMessage<Messages::WebPage::UppercaseWord>(decoder, this, &WebPage::uppercaseWord);
        return;
    }
#endif
#if USE(APPKIT)
    if (decoder.messageName() == Messages::WebPage::LowercaseWord::name()) {
        IPC::handleMessage<Messages::WebPage::LowercaseWord>(decoder, this, &WebPage::lowercaseWord);
        return;
    }
#endif
#if USE(APPKIT)
    if (decoder.messageName() == Messages::WebPage::CapitalizeWord::name()) {
        IPC::handleMessage<Messages::WebPage::CapitalizeWord>(decoder, this, &WebPage::capitalizeWord);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::SetSmartInsertDeleteEnabled::name()) {
        IPC::handleMessage<Messages::WebPage::SetSmartInsertDeleteEnabled>(decoder, this, &WebPage::setSmartInsertDeleteEnabled);
        return;
    }
#endif
#if ENABLE(GEOLOCATION)
    if (decoder.messageName() == Messages::WebPage::DidReceiveGeolocationPermissionDecision::name()) {
        IPC::handleMessage<Messages::WebPage::DidReceiveGeolocationPermissionDecision>(decoder, this, &WebPage::didReceiveGeolocationPermissionDecision);
        return;
    }
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::WebPage::UserMediaAccessWasGranted::name()) {
        IPC::handleMessageAsync<Messages::WebPage::UserMediaAccessWasGranted>(connection, decoder, this, &WebPage::userMediaAccessWasGranted);
        return;
    }
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::WebPage::UserMediaAccessWasDenied::name()) {
        IPC::handleMessage<Messages::WebPage::UserMediaAccessWasDenied>(decoder, this, &WebPage::userMediaAccessWasDenied);
        return;
    }
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::WebPage::CaptureDevicesChanged::name()) {
        IPC::handleMessage<Messages::WebPage::CaptureDevicesChanged>(decoder, this, &WebPage::captureDevicesChanged);
        return;
    }
#endif
#if ENABLE(WEB_RTC)
    if (decoder.messageName() == Messages::WebPage::DisableICECandidateFiltering::name()) {
        IPC::handleMessage<Messages::WebPage::DisableICECandidateFiltering>(decoder, this, &WebPage::disableICECandidateFiltering);
        return;
    }
#endif
#if ENABLE(WEB_RTC)
    if (decoder.messageName() == Messages::WebPage::EnableICECandidateFiltering::name()) {
        IPC::handleMessage<Messages::WebPage::EnableICECandidateFiltering>(decoder, this, &WebPage::enableICECandidateFiltering);
        return;
    }
#endif
#if ENABLE(WEB_RTC) && USE(LIBWEBRTC)
    if (decoder.messageName() == Messages::WebPage::EnableEnumeratingAllNetworkInterfaces::name()) {
        IPC::handleMessage<Messages::WebPage::EnableEnumeratingAllNetworkInterfaces>(decoder, this, &WebPage::enableEnumeratingAllNetworkInterfaces);
        return;
    }
#endif
#if ENABLE(WEB_RTC) && USE(LIBWEBRTC)
    if (decoder.messageName() == Messages::WebPage::DisableEnumeratingAllNetworkInterfaces::name()) {
        IPC::handleMessage<Messages::WebPage::DisableEnumeratingAllNetworkInterfaces>(decoder, this, &WebPage::disableEnumeratingAllNetworkInterfaces);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPage::StopAllMediaPlayback::name()) {
        IPC::handleMessage<Messages::WebPage::StopAllMediaPlayback>(decoder, this, &WebPage::stopAllMediaPlayback);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SuspendAllMediaPlayback::name()) {
        IPC::handleMessage<Messages::WebPage::SuspendAllMediaPlayback>(decoder, this, &WebPage::suspendAllMediaPlayback);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::ResumeAllMediaPlayback::name()) {
        IPC::handleMessage<Messages::WebPage::ResumeAllMediaPlayback>(decoder, this, &WebPage::resumeAllMediaPlayback);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::DidReceiveNotificationPermissionDecision::name()) {
        IPC::handleMessage<Messages::WebPage::DidReceiveNotificationPermissionDecision>(decoder, this, &WebPage::didReceiveNotificationPermissionDecision);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::FreezeLayerTreeDueToSwipeAnimation::name()) {
        IPC::handleMessage<Messages::WebPage::FreezeLayerTreeDueToSwipeAnimation>(decoder, this, &WebPage::freezeLayerTreeDueToSwipeAnimation);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::UnfreezeLayerTreeDueToSwipeAnimation::name()) {
        IPC::handleMessage<Messages::WebPage::UnfreezeLayerTreeDueToSwipeAnimation>(decoder, this, &WebPage::unfreezeLayerTreeDueToSwipeAnimation);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::BeginPrinting::name()) {
        IPC::handleMessage<Messages::WebPage::BeginPrinting>(decoder, this, &WebPage::beginPrinting);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::EndPrinting::name()) {
        IPC::handleMessage<Messages::WebPage::EndPrinting>(decoder, this, &WebPage::endPrinting);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::ComputePagesForPrinting::name()) {
        IPC::handleMessage<Messages::WebPage::ComputePagesForPrinting>(decoder, this, &WebPage::computePagesForPrinting);
        return;
    }
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::DrawRectToImage::name()) {
        IPC::handleMessage<Messages::WebPage::DrawRectToImage>(decoder, this, &WebPage::drawRectToImage);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::DrawPagesToPDF::name()) {
        IPC::handleMessage<Messages::WebPage::DrawPagesToPDF>(decoder, this, &WebPage::drawPagesToPDF);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::DrawToPDF::name()) {
        IPC::handleMessage<Messages::WebPage::DrawToPDF>(decoder, this, &WebPage::drawToPDF);
        return;
    }
#endif
#if PLATFORM(GTK)
    if (decoder.messageName() == Messages::WebPage::DrawPagesForPrinting::name()) {
        IPC::handleMessage<Messages::WebPage::DrawPagesForPrinting>(decoder, this, &WebPage::drawPagesForPrinting);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPage::SetMediaVolume::name()) {
        IPC::handleMessage<Messages::WebPage::SetMediaVolume>(decoder, this, &WebPage::setMediaVolume);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetMuted::name()) {
        IPC::handleMessage<Messages::WebPage::SetMuted>(decoder, this, &WebPage::setMuted);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetMayStartMediaWhenInWindow::name()) {
        IPC::handleMessage<Messages::WebPage::SetMayStartMediaWhenInWindow>(decoder, this, &WebPage::setMayStartMediaWhenInWindow);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::StopMediaCapture::name()) {
        IPC::handleMessage<Messages::WebPage::StopMediaCapture>(decoder, this, &WebPage::stopMediaCapture);
        return;
    }
#if ENABLE(MEDIA_SESSION)
    if (decoder.messageName() == Messages::WebPage::HandleMediaEvent::name()) {
        IPC::handleMessage<Messages::WebPage::HandleMediaEvent>(decoder, this, &WebPage::handleMediaEvent);
        return;
    }
#endif
#if ENABLE(MEDIA_SESSION)
    if (decoder.messageName() == Messages::WebPage::SetVolumeOfMediaElement::name()) {
        IPC::handleMessage<Messages::WebPage::SetVolumeOfMediaElement>(decoder, this, &WebPage::setVolumeOfMediaElement);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPage::SetCanRunBeforeUnloadConfirmPanel::name()) {
        IPC::handleMessage<Messages::WebPage::SetCanRunBeforeUnloadConfirmPanel>(decoder, this, &WebPage::setCanRunBeforeUnloadConfirmPanel);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetCanRunModal::name()) {
        IPC::handleMessage<Messages::WebPage::SetCanRunModal>(decoder, this, &WebPage::setCanRunModal);
        return;
    }
#if PLATFORM(GTK) || PLATFORM(WPE)
    if (decoder.messageName() == Messages::WebPage::CancelComposition::name()) {
        IPC::handleMessage<Messages::WebPage::CancelComposition>(decoder, this, &WebPage::cancelComposition);
        return;
    }
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    if (decoder.messageName() == Messages::WebPage::DeleteSurrounding::name()) {
        IPC::handleMessage<Messages::WebPage::DeleteSurrounding>(decoder, this, &WebPage::deleteSurrounding);
        return;
    }
#endif
#if PLATFORM(GTK)
    if (decoder.messageName() == Messages::WebPage::CollapseSelectionInFrame::name()) {
        IPC::handleMessage<Messages::WebPage::CollapseSelectionInFrame>(decoder, this, &WebPage::collapseSelectionInFrame);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::SendComplexTextInputToPlugin::name()) {
        IPC::handleMessage<Messages::WebPage::SendComplexTextInputToPlugin>(decoder, this, &WebPage::sendComplexTextInputToPlugin);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::WindowAndViewFramesChanged::name()) {
        IPC::handleMessage<Messages::WebPage::WindowAndViewFramesChanged>(decoder, this, &WebPage::windowAndViewFramesChanged);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::SetMainFrameIsScrollable::name()) {
        IPC::handleMessage<Messages::WebPage::SetMainFrameIsScrollable>(decoder, this, &WebPage::setMainFrameIsScrollable);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::RegisterUIProcessAccessibilityTokens::name()) {
        IPC::handleMessage<Messages::WebPage::RegisterUIProcessAccessibilityTokens>(decoder, this, &WebPage::registerUIProcessAccessibilityTokens);
        return;
    }
#endif
#if (PLATFORM(COCOA) && ENABLE(SERVICE_CONTROLS))
    if (decoder.messageName() == Messages::WebPage::ReplaceSelectionWithPasteboardData::name()) {
        IPC::handleMessage<Messages::WebPage::ReplaceSelectionWithPasteboardData>(decoder, this, &WebPage::replaceSelectionWithPasteboardData);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::SetTextAsync::name()) {
        IPC::handleMessage<Messages::WebPage::SetTextAsync>(decoder, this, &WebPage::setTextAsync);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::InsertTextAsync::name()) {
        IPC::handleMessage<Messages::WebPage::InsertTextAsync>(decoder, this, &WebPage::insertTextAsync);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::HasMarkedText::name()) {
        IPC::handleMessageAsync<Messages::WebPage::HasMarkedText>(connection, decoder, this, &WebPage::hasMarkedText);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::GetMarkedRangeAsync::name()) {
        IPC::handleMessage<Messages::WebPage::GetMarkedRangeAsync>(decoder, this, &WebPage::getMarkedRangeAsync);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::GetSelectedRangeAsync::name()) {
        IPC::handleMessage<Messages::WebPage::GetSelectedRangeAsync>(decoder, this, &WebPage::getSelectedRangeAsync);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::CharacterIndexForPointAsync::name()) {
        IPC::handleMessage<Messages::WebPage::CharacterIndexForPointAsync>(decoder, this, &WebPage::characterIndexForPointAsync);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::FirstRectForCharacterRangeAsync::name()) {
        IPC::handleMessage<Messages::WebPage::FirstRectForCharacterRangeAsync>(decoder, this, &WebPage::firstRectForCharacterRangeAsync);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::SetCompositionAsync::name()) {
        IPC::handleMessage<Messages::WebPage::SetCompositionAsync>(decoder, this, &WebPage::setCompositionAsync);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::ConfirmCompositionAsync::name()) {
        IPC::handleMessage<Messages::WebPage::ConfirmCompositionAsync>(decoder, this, &WebPage::confirmCompositionAsync);
        return;
    }
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebPage::InsertDictatedTextAsync::name()) {
        IPC::handleMessage<Messages::WebPage::InsertDictatedTextAsync>(decoder, this, &WebPage::insertDictatedTextAsync);
        return;
    }
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebPage::AttributedSubstringForCharacterRangeAsync::name()) {
        IPC::handleMessage<Messages::WebPage::AttributedSubstringForCharacterRangeAsync>(decoder, this, &WebPage::attributedSubstringForCharacterRangeAsync);
        return;
    }
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebPage::FontAtSelection::name()) {
        IPC::handleMessage<Messages::WebPage::FontAtSelection>(decoder, this, &WebPage::fontAtSelection);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPage::SetAlwaysShowsHorizontalScroller::name()) {
        IPC::handleMessage<Messages::WebPage::SetAlwaysShowsHorizontalScroller>(decoder, this, &WebPage::setAlwaysShowsHorizontalScroller);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetAlwaysShowsVerticalScroller::name()) {
        IPC::handleMessage<Messages::WebPage::SetAlwaysShowsVerticalScroller>(decoder, this, &WebPage::setAlwaysShowsVerticalScroller);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetMinimumSizeForAutoLayout::name()) {
        IPC::handleMessage<Messages::WebPage::SetMinimumSizeForAutoLayout>(decoder, this, &WebPage::setMinimumSizeForAutoLayout);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetAutoSizingShouldExpandToViewHeight::name()) {
        IPC::handleMessage<Messages::WebPage::SetAutoSizingShouldExpandToViewHeight>(decoder, this, &WebPage::setAutoSizingShouldExpandToViewHeight);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetViewportSizeForCSSViewportUnits::name()) {
        IPC::handleMessage<Messages::WebPage::SetViewportSizeForCSSViewportUnits>(decoder, this, &WebPage::setViewportSizeForCSSViewportUnits);
        return;
    }
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::HandleAlternativeTextUIResult::name()) {
        IPC::handleMessage<Messages::WebPage::HandleAlternativeTextUIResult>(decoder, this, &WebPage::handleAlternativeTextUIResult);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::WillStartUserTriggeredZooming::name()) {
        IPC::handleMessage<Messages::WebPage::WillStartUserTriggeredZooming>(decoder, this, &WebPage::willStartUserTriggeredZooming);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPage::SetScrollPinningBehavior::name()) {
        IPC::handleMessage<Messages::WebPage::SetScrollPinningBehavior>(decoder, this, &WebPage::setScrollPinningBehavior);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetScrollbarOverlayStyle::name()) {
        IPC::handleMessage<Messages::WebPage::SetScrollbarOverlayStyle>(decoder, this, &WebPage::setScrollbarOverlayStyle);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::GetBytecodeProfile::name()) {
        IPC::handleMessage<Messages::WebPage::GetBytecodeProfile>(decoder, this, &WebPage::getBytecodeProfile);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::GetSamplingProfilerOutput::name()) {
        IPC::handleMessage<Messages::WebPage::GetSamplingProfilerOutput>(decoder, this, &WebPage::getSamplingProfilerOutput);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::TakeSnapshot::name()) {
        IPC::handleMessage<Messages::WebPage::TakeSnapshot>(decoder, this, &WebPage::takeSnapshot);
        return;
    }
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebPage::PerformImmediateActionHitTestAtLocation::name()) {
        IPC::handleMessage<Messages::WebPage::PerformImmediateActionHitTestAtLocation>(decoder, this, &WebPage::performImmediateActionHitTestAtLocation);
        return;
    }
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebPage::ImmediateActionDidUpdate::name()) {
        IPC::handleMessage<Messages::WebPage::ImmediateActionDidUpdate>(decoder, this, &WebPage::immediateActionDidUpdate);
        return;
    }
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebPage::ImmediateActionDidCancel::name()) {
        IPC::handleMessage<Messages::WebPage::ImmediateActionDidCancel>(decoder, this, &WebPage::immediateActionDidCancel);
        return;
    }
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebPage::ImmediateActionDidComplete::name()) {
        IPC::handleMessage<Messages::WebPage::ImmediateActionDidComplete>(decoder, this, &WebPage::immediateActionDidComplete);
        return;
    }
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebPage::DataDetectorsDidPresentUI::name()) {
        IPC::handleMessage<Messages::WebPage::DataDetectorsDidPresentUI>(decoder, this, &WebPage::dataDetectorsDidPresentUI);
        return;
    }
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebPage::DataDetectorsDidChangeUI::name()) {
        IPC::handleMessage<Messages::WebPage::DataDetectorsDidChangeUI>(decoder, this, &WebPage::dataDetectorsDidChangeUI);
        return;
    }
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebPage::DataDetectorsDidHideUI::name()) {
        IPC::handleMessage<Messages::WebPage::DataDetectorsDidHideUI>(decoder, this, &WebPage::dataDetectorsDidHideUI);
        return;
    }
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebPage::HandleAcceptedCandidate::name()) {
        IPC::handleMessage<Messages::WebPage::HandleAcceptedCandidate>(decoder, this, &WebPage::handleAcceptedCandidate);
        return;
    }
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebPage::SetUseSystemAppearance::name()) {
        IPC::handleMessage<Messages::WebPage::SetUseSystemAppearance>(decoder, this, &WebPage::setUseSystemAppearance);
        return;
    }
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebPage::SetHeaderBannerHeightForTesting::name()) {
        IPC::handleMessage<Messages::WebPage::SetHeaderBannerHeightForTesting>(decoder, this, &WebPage::setHeaderBannerHeightForTesting);
        return;
    }
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebPage::SetFooterBannerHeightForTesting::name()) {
        IPC::handleMessage<Messages::WebPage::SetFooterBannerHeightForTesting>(decoder, this, &WebPage::setFooterBannerHeightForTesting);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPage::EffectiveAppearanceDidChange::name()) {
        IPC::handleMessage<Messages::WebPage::EffectiveAppearanceDidChange>(decoder, this, &WebPage::effectiveAppearanceDidChange);
        return;
    }
#if PLATFORM(GTK)
    if (decoder.messageName() == Messages::WebPage::ThemeDidChange::name()) {
        IPC::handleMessage<Messages::WebPage::ThemeDidChange>(decoder, this, &WebPage::themeDidChange);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::RequestActiveNowPlayingSessionInfo::name()) {
        IPC::handleMessage<Messages::WebPage::RequestActiveNowPlayingSessionInfo>(decoder, this, &WebPage::requestActiveNowPlayingSessionInfo);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPage::SetShouldDispatchFakeMouseMoveEvents::name()) {
        IPC::handleMessage<Messages::WebPage::SetShouldDispatchFakeMouseMoveEvents>(decoder, this, &WebPage::setShouldDispatchFakeMouseMoveEvents);
        return;
    }
#if ENABLE(WIRELESS_PLAYBACK_TARGET) && !PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::PlaybackTargetSelected::name()) {
        IPC::handleMessage<Messages::WebPage::PlaybackTargetSelected>(decoder, this, &WebPage::playbackTargetSelected);
        return;
    }
#endif
#if ENABLE(WIRELESS_PLAYBACK_TARGET) && !PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::PlaybackTargetAvailabilityDidChange::name()) {
        IPC::handleMessage<Messages::WebPage::PlaybackTargetAvailabilityDidChange>(decoder, this, &WebPage::playbackTargetAvailabilityDidChange);
        return;
    }
#endif
#if ENABLE(WIRELESS_PLAYBACK_TARGET) && !PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::SetShouldPlayToPlaybackTarget::name()) {
        IPC::handleMessage<Messages::WebPage::SetShouldPlayToPlaybackTarget>(decoder, this, &WebPage::setShouldPlayToPlaybackTarget);
        return;
    }
#endif
#if ENABLE(WIRELESS_PLAYBACK_TARGET) && !PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::PlaybackTargetPickerWasDismissed::name()) {
        IPC::handleMessage<Messages::WebPage::PlaybackTargetPickerWasDismissed>(decoder, this, &WebPage::playbackTargetPickerWasDismissed);
        return;
    }
#endif
#if ENABLE(POINTER_LOCK)
    if (decoder.messageName() == Messages::WebPage::DidAcquirePointerLock::name()) {
        IPC::handleMessage<Messages::WebPage::DidAcquirePointerLock>(decoder, this, &WebPage::didAcquirePointerLock);
        return;
    }
#endif
#if ENABLE(POINTER_LOCK)
    if (decoder.messageName() == Messages::WebPage::DidNotAcquirePointerLock::name()) {
        IPC::handleMessage<Messages::WebPage::DidNotAcquirePointerLock>(decoder, this, &WebPage::didNotAcquirePointerLock);
        return;
    }
#endif
#if ENABLE(POINTER_LOCK)
    if (decoder.messageName() == Messages::WebPage::DidLosePointerLock::name()) {
        IPC::handleMessage<Messages::WebPage::DidLosePointerLock>(decoder, this, &WebPage::didLosePointerLock);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPage::clearWheelEventTestMonitor::name()) {
        IPC::handleMessage<Messages::WebPage::clearWheelEventTestMonitor>(decoder, this, &WebPage::clearWheelEventTestMonitor);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetShouldScaleViewToFitDocument::name()) {
        IPC::handleMessage<Messages::WebPage::SetShouldScaleViewToFitDocument>(decoder, this, &WebPage::setShouldScaleViewToFitDocument);
        return;
    }
#if ENABLE(VIDEO) && USE(GSTREAMER)
    if (decoder.messageName() == Messages::WebPage::DidEndRequestInstallMissingMediaPlugins::name()) {
        IPC::handleMessage<Messages::WebPage::DidEndRequestInstallMissingMediaPlugins>(decoder, this, &WebPage::didEndRequestInstallMissingMediaPlugins);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPage::SetUserInterfaceLayoutDirection::name()) {
        IPC::handleMessage<Messages::WebPage::SetUserInterfaceLayoutDirection>(decoder, this, &WebPage::setUserInterfaceLayoutDirection);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::DidGetLoadDecisionForIcon::name()) {
        IPC::handleMessage<Messages::WebPage::DidGetLoadDecisionForIcon>(decoder, this, &WebPage::didGetLoadDecisionForIcon);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetUseIconLoadingClient::name()) {
        IPC::handleMessage<Messages::WebPage::SetUseIconLoadingClient>(decoder, this, &WebPage::setUseIconLoadingClient);
        return;
    }
#if ENABLE(GAMEPAD)
    if (decoder.messageName() == Messages::WebPage::GamepadActivity::name()) {
        IPC::handleMessage<Messages::WebPage::GamepadActivity>(decoder, this, &WebPage::gamepadActivity);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPage::FrameBecameRemote::name()) {
        IPC::handleMessage<Messages::WebPage::FrameBecameRemote>(decoder, this, &WebPage::frameBecameRemote);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::RegisterURLSchemeHandler::name()) {
        IPC::handleMessage<Messages::WebPage::RegisterURLSchemeHandler>(decoder, this, &WebPage::registerURLSchemeHandler);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::URLSchemeTaskDidPerformRedirection::name()) {
        IPC::handleMessage<Messages::WebPage::URLSchemeTaskDidPerformRedirection>(decoder, this, &WebPage::urlSchemeTaskDidPerformRedirection);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::URLSchemeTaskDidReceiveResponse::name()) {
        IPC::handleMessage<Messages::WebPage::URLSchemeTaskDidReceiveResponse>(decoder, this, &WebPage::urlSchemeTaskDidReceiveResponse);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::URLSchemeTaskDidReceiveData::name()) {
        IPC::handleMessage<Messages::WebPage::URLSchemeTaskDidReceiveData>(decoder, this, &WebPage::urlSchemeTaskDidReceiveData);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::URLSchemeTaskDidComplete::name()) {
        IPC::handleMessage<Messages::WebPage::URLSchemeTaskDidComplete>(decoder, this, &WebPage::urlSchemeTaskDidComplete);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetIsSuspended::name()) {
        IPC::handleMessage<Messages::WebPage::SetIsSuspended>(decoder, this, &WebPage::setIsSuspended);
        return;
    }
#if ENABLE(ATTACHMENT_ELEMENT)
    if (decoder.messageName() == Messages::WebPage::InsertAttachment::name()) {
        IPC::handleMessage<Messages::WebPage::InsertAttachment>(decoder, this, &WebPage::insertAttachment);
        return;
    }
#endif
#if ENABLE(ATTACHMENT_ELEMENT)
    if (decoder.messageName() == Messages::WebPage::UpdateAttachmentAttributes::name()) {
        IPC::handleMessage<Messages::WebPage::UpdateAttachmentAttributes>(decoder, this, &WebPage::updateAttachmentAttributes);
        return;
    }
#endif
#if ENABLE(APPLICATION_MANIFEST)
    if (decoder.messageName() == Messages::WebPage::GetApplicationManifest::name()) {
        IPC::handleMessage<Messages::WebPage::GetApplicationManifest>(decoder, this, &WebPage::getApplicationManifest);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPage::SetDefersLoading::name()) {
        IPC::handleMessage<Messages::WebPage::SetDefersLoading>(decoder, this, &WebPage::setDefersLoading);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::UpdateCurrentModifierState::name()) {
        IPC::handleMessage<Messages::WebPage::UpdateCurrentModifierState>(decoder, this, &WebPage::updateCurrentModifierState);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SimulateDeviceOrientationChange::name()) {
        IPC::handleMessage<Messages::WebPage::SimulateDeviceOrientationChange>(decoder, this, &WebPage::simulateDeviceOrientationChange);
        return;
    }
#if ENABLE(SPEECH_SYNTHESIS)
    if (decoder.messageName() == Messages::WebPage::SpeakingErrorOccurred::name()) {
        IPC::handleMessage<Messages::WebPage::SpeakingErrorOccurred>(decoder, this, &WebPage::speakingErrorOccurred);
        return;
    }
#endif
#if ENABLE(SPEECH_SYNTHESIS)
    if (decoder.messageName() == Messages::WebPage::BoundaryEventOccurred::name()) {
        IPC::handleMessage<Messages::WebPage::BoundaryEventOccurred>(decoder, this, &WebPage::boundaryEventOccurred);
        return;
    }
#endif
#if ENABLE(SPEECH_SYNTHESIS)
    if (decoder.messageName() == Messages::WebPage::VoicesDidChange::name()) {
        IPC::handleMessage<Messages::WebPage::VoicesDidChange>(decoder, this, &WebPage::voicesDidChange);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPage::TextInputContextsInRect::name()) {
        IPC::handleMessageAsync<Messages::WebPage::TextInputContextsInRect>(connection, decoder, this, &WebPage::textInputContextsInRect);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::FocusTextInputContext::name()) {
        IPC::handleMessageAsync<Messages::WebPage::FocusTextInputContext>(connection, decoder, this, &WebPage::focusTextInputContext);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetCanShowPlaceholder::name()) {
        IPC::handleMessage<Messages::WebPage::SetCanShowPlaceholder>(decoder, this, &WebPage::setCanShowPlaceholder);
        return;
    }
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::WebPage::WasLoadedWithDataTransferFromPrevalentResource::name()) {
        IPC::handleMessage<Messages::WebPage::WasLoadedWithDataTransferFromPrevalentResource>(decoder, this, &WebPage::wasLoadedWithDataTransferFromPrevalentResource);
        return;
    }
#endif
#if USE(SYSTEM_PREVIEW)
    if (decoder.messageName() == Messages::WebPage::SystemPreviewActionTriggered::name()) {
        IPC::handleMessage<Messages::WebPage::SystemPreviewActionTriggered>(decoder, this, &WebPage::systemPreviewActionTriggered);
        return;
    }
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    if (decoder.messageName() == Messages::WebPage::SendMessageToWebExtension::name()) {
        IPC::handleMessage<Messages::WebPage::SendMessageToWebExtension>(decoder, this, &WebPage::sendMessageToWebExtension);
        return;
    }
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    if (decoder.messageName() == Messages::WebPage::SendMessageToWebExtensionWithReply::name()) {
        IPC::handleMessageAsync<Messages::WebPage::SendMessageToWebExtensionWithReply>(connection, decoder, this, &WebPage::sendMessageToWebExtensionWithReply);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPage::StartTextManipulations::name()) {
        IPC::handleMessageAsync<Messages::WebPage::StartTextManipulations>(connection, decoder, this, &WebPage::startTextManipulations);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::CompleteTextManipulation::name()) {
        IPC::handleMessageAsync<Messages::WebPage::CompleteTextManipulation>(connection, decoder, this, &WebPage::completeTextManipulation);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetOverriddenMediaType::name()) {
        IPC::handleMessage<Messages::WebPage::SetOverriddenMediaType>(decoder, this, &WebPage::setOverriddenMediaType);
        return;
    }
    if (decoder.messageName() == Messages::WebPage::SetHasResourceLoadClient::name()) {
        IPC::handleMessage<Messages::WebPage::SetHasResourceLoadClient>(decoder, this, &WebPage::setHasResourceLoadClient);
        return;
    }
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED();
}

void WebPage::didReceiveSyncWebPageMessage(IPC::Connection& connection, IPC::Decoder& decoder, std::unique_ptr<IPC::Encoder>& replyEncoder)
{
    auto protectedThis = makeRef(*this);
    if (decoder.messageName() == Messages::WebPage::TestProcessIncomingSyncMessagesWhenWaitingForSyncReply::name()) {
        IPC::handleMessageSynchronous<Messages::WebPage::TestProcessIncomingSyncMessagesWhenWaitingForSyncReply>(connection, decoder, replyEncoder, this, &WebPage::testProcessIncomingSyncMessagesWhenWaitingForSyncReply);
        return;
    }
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::SyncApplyAutocorrection::name()) {
        IPC::handleMessageSynchronous<Messages::WebPage::SyncApplyAutocorrection>(connection, decoder, replyEncoder, this, &WebPage::syncApplyAutocorrection);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPage::GetPositionInformation::name()) {
        IPC::handleMessageSynchronous<Messages::WebPage::GetPositionInformation>(connection, decoder, replyEncoder, this, &WebPage::getPositionInformation);
        return;
    }
#endif
#if ENABLE(IOS_TOUCH_EVENTS)
    if (decoder.messageName() == Messages::WebPage::TouchEventSync::name()) {
        IPC::handleMessageSynchronous<Messages::WebPage::TouchEventSync>(connection, decoder, replyEncoder, this, &WebPage::touchEventSync);
        return;
    }
#endif
#if (PLATFORM(COCOA) && PLATFORM(IOS_FAMILY))
    if (decoder.messageName() == Messages::WebPage::ComputePagesForPrintingAndDrawToPDF::name()) {
        IPC::handleMessageSynchronous<Messages::WebPage::ComputePagesForPrintingAndDrawToPDF>(connection, decoder, replyEncoder, this, &WebPage::computePagesForPrintingAndDrawToPDF);
        return;
    }
#endif
#if PLATFORM(GTK)
    if (decoder.messageName() == Messages::WebPage::GetCenterForZoomGesture::name()) {
        IPC::handleMessageSynchronous<Messages::WebPage::GetCenterForZoomGesture>(connection, decoder, replyEncoder, this, &WebPage::getCenterForZoomGesture);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::GetStringSelectionForPasteboard::name()) {
        IPC::handleMessageSynchronous<Messages::WebPage::GetStringSelectionForPasteboard>(connection, decoder, replyEncoder, this, &WebPage::getStringSelectionForPasteboard);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::GetDataSelectionForPasteboard::name()) {
        IPC::handleMessageSynchronous<Messages::WebPage::GetDataSelectionForPasteboard>(connection, decoder, replyEncoder, this, &WebPage::getDataSelectionForPasteboard);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::ReadSelectionFromPasteboard::name()) {
        IPC::handleMessageSynchronous<Messages::WebPage::ReadSelectionFromPasteboard>(connection, decoder, replyEncoder, this, &WebPage::readSelectionFromPasteboard);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::ShouldDelayWindowOrderingEvent::name()) {
        IPC::handleMessageSynchronous<Messages::WebPage::ShouldDelayWindowOrderingEvent>(connection, decoder, replyEncoder, this, &WebPage::shouldDelayWindowOrderingEvent);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPage::AcceptsFirstMouse::name()) {
        IPC::handleMessageSynchronous<Messages::WebPage::AcceptsFirstMouse>(connection, decoder, replyEncoder, this, &WebPage::acceptsFirstMouse);
        return;
    }
#endif
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    UNUSED_PARAM(replyEncoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit

