/*
 * Copyright (C) 2010-2019 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#include "WebPageProxy.h"

#include "ArgumentCoders.h"
#if PLATFORM(MAC)
#include "AttributedString.h"
#endif
#include "CallbackID.h"
#if ENABLE(CONTEXT_MENUS)
#include "ContextMenuContextData.h"
#endif
#if ENABLE(DATA_DETECTION)
#include "DataDetectionResult.h"
#endif
#include "DataReference.h"
#include "Decoder.h"
#include "DownloadID.h"
#include "EditingRange.h"
#include "EditorState.h"
#if PLATFORM(IOS_FAMILY)
#include "FocusedElementInformation.h"
#endif
#if PLATFORM(MAC)
#include "FontInfo.h"
#endif
#include "FormDataReference.h"
#include "FrameInfoData.h"
#include "HandleMessage.h"
#if PLATFORM(GTK) || PLATFORM(WPE)
#include "InputMethodState.h"
#endif
#if PLATFORM(IOS_FAMILY)
#include "InteractionInformationAtPosition.h"
#endif
#if HAVE(VISIBILITY_PROPAGATION_VIEW)
#include "LayerHostingContext.h"
#endif
#include "NavigationActionData.h"
#if ENABLE(PDFKIT_PLUGIN)
#include "PDFContextMenu.h"
#endif
#include "PlatformPopupMenuData.h"
#include "SandboxExtension.h"
#include "SessionState.h"
#include "ShareableBitmap.h"
#if USE(QUICK_LOOK)
#include "ShareableResource.h"
#endif
#include "SharedBufferDataReference.h"
#if PLATFORM(COCOA) && ENABLE(DRAG_SUPPORT) || PLATFORM(IOS_FAMILY)
#include "SharedMemory.h"
#endif
#if HAVE(TOUCH_BAR)
#include "TouchBarMenuData.h"
#endif
#if HAVE(TOUCH_BAR)
#include "TouchBarMenuItemData.h"
#endif
#include "URLSchemeTaskParameters.h"
#include "UndoOrRedo.h"
#include "UserData.h"
#if PLATFORM(GTK) || PLATFORM(WPE)
#include "UserMessage.h"
#endif
#if PLATFORM(IOS_FAMILY)
#include "WebAutocorrectionContext.h"
#endif
#include "WebBackForwardListCounts.h"
#include "WebCoreArgumentCoders.h"
#include "WebHitTestResultData.h"
#include "WebNavigationDataStore.h"
#include "WebPageCreationParameters.h"
#include "WebPageProxyIdentifier.h"
#include "WebPageProxyMessages.h"
#include "WebPopupItem.h"
#if PLATFORM(GTK) && ENABLE(DRAG_SUPPORT)
#include "WebSelectionData.h"
#endif
#if ENABLE(SPEECH_SYNTHESIS)
#include "WebSpeechSynthesisVoice.h"
#endif
#include "WebsitePoliciesData.h"
#include <JavaScriptCore/InspectorTarget.h>
#if PLATFORM(IOS_FAMILY)
#include <WebCore/ActivityState.h>
#endif
#if ENABLE(APPLICATION_MANIFEST)
#include <WebCore/ApplicationManifest.h>
#endif
#if PLATFORM(IOS_FAMILY)
#include <WebCore/AudioSession.h>
#endif
#include <WebCore/AutoplayEvent.h>
#include <WebCore/BackForwardItemIdentifier.h>
#if ENABLE(MEDIA_STREAM)
#include <WebCore/CaptureDevice.h>
#endif
#include <WebCore/CertificateInfo.h>
#include <WebCore/Color.h>
#if ENABLE(CONTENT_FILTERING)
#include <WebCore/ContentFilterUnblockHandler.h>
#endif
#include <WebCore/ContentRuleListResults.h>
#include <WebCore/Cursor.h>
#include <WebCore/DOMPasteAccess.h>
#if ENABLE(DATALIST_ELEMENT)
#include <WebCore/DataListSuggestionInformation.h>
#endif
#if ENABLE(DEVICE_ORIENTATION)
#include <WebCore/DeviceOrientationOrMotionPermissionState.h>
#endif
#include <WebCore/DiagnosticLoggingClient.h>
#if PLATFORM(COCOA)
#include <WebCore/DictionaryPopupInfo.h>
#endif
#if ENABLE(DRAG_SUPPORT)
#include <WebCore/DragActions.h>
#endif
#if PLATFORM(COCOA) && ENABLE(DRAG_SUPPORT)
#include <WebCore/DragItem.h>
#endif
#include <WebCore/FileChooser.h>
#if PLATFORM(IOS_FAMILY)
#include <WebCore/FloatPoint.h>
#endif
#if PLATFORM(IOS_FAMILY)
#include <WebCore/FloatQuad.h>
#endif
#include <WebCore/FloatRect.h>
#include <WebCore/FontAttributes.h>
#include <WebCore/FrameIdentifier.h>
#include <WebCore/FrameLoaderTypes.h>
#include <WebCore/GlobalWindowIdentifier.h>
#if PLATFORM(IOS_FAMILY)
#include <WebCore/InputMode.h>
#endif
#if PLATFORM(IOS_FAMILY)
#include <WebCore/InspectorOverlay.h>
#endif
#include <WebCore/IntPoint.h>
#include <WebCore/IntRect.h>
#include <WebCore/IntSize.h>
#include <WebCore/JSDOMExceptionHandling.h>
#include <WebCore/LayoutMilestone.h>
#include <WebCore/LinkIcon.h>
#if ENABLE(MEDIA_SESSION)
#include <WebCore/MediaSessionMetadata.h>
#endif
#if ENABLE(MEDIA_STREAM)
#include <WebCore/MediaStreamRequest.h>
#endif
#if ENABLE(WEB_AUTHN)
#include <WebCore/MockWebAuthenticationConfiguration.h>
#endif
#include <WebCore/NotificationDirection.h>
#include <WebCore/PageIdentifier.h>
#include <WebCore/RectEdges.h>
#include <WebCore/ResourceError.h>
#include <WebCore/ResourceRequest.h>
#include <WebCore/ResourceResponse.h>
#include <WebCore/SearchPopupMenu.h>
#include <WebCore/SecurityOriginData.h>
#if ENABLE(ATTACHMENT_ELEMENT)
#include <WebCore/SerializedAttachmentData.h>
#endif
#include <WebCore/ShareData.h>
#include <WebCore/TextCheckerClient.h>
#include <WebCore/TextChecking.h>
#include <WebCore/TextIndicator.h>
#include <WebCore/TextManipulationController.h>
#include <WebCore/ViewportArguments.h>
#include <WebCore/WindowFeatures.h>
#include <utility>
#if PLATFORM(COCOA)
#include <wtf/MachSendRight.h>
#endif
#include <wtf/MonotonicTime.h>
#include <wtf/OptionSet.h>
#include <wtf/Optional.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace Messages {

namespace WebPageProxy {

void CreateNewPage::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Optional<WebCore::PageIdentifier>& newPageID, const Optional<WebKit::WebPageCreationParameters>& newPageParameters)
{
    *encoder << newPageID;
    *encoder << newPageParameters;
    connection.sendSyncReply(WTFMove(encoder));
}

void RunJavaScriptAlert::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

void RunJavaScriptConfirm::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool result)
{
    *encoder << result;
    connection.sendSyncReply(WTFMove(encoder));
}

void RunJavaScriptPrompt::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const String& result)
{
    *encoder << result;
    connection.sendSyncReply(WTFMove(encoder));
}

#if ENABLE(WEBGL)

void WebGLPolicyForURL::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, uint32_t loadPolicy)
{
    *encoder << loadPolicy;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(WEBGL)

void ResolveWebGLPolicyForURL::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, uint32_t loadPolicy)
{
    *encoder << loadPolicy;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

void GetToolbarsAreVisible::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool toolbarsAreVisible)
{
    *encoder << toolbarsAreVisible;
    connection.sendSyncReply(WTFMove(encoder));
}

void GetMenuBarIsVisible::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool menuBarIsVisible)
{
    *encoder << menuBarIsVisible;
    connection.sendSyncReply(WTFMove(encoder));
}

void GetStatusBarIsVisible::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool statusBarIsVisible)
{
    *encoder << statusBarIsVisible;
    connection.sendSyncReply(WTFMove(encoder));
}

void GetWindowFrame::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebCore::FloatRect& windowFrame)
{
    *encoder << windowFrame;
    connection.sendSyncReply(WTFMove(encoder));
}

void ScreenToRootView::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebCore::IntPoint& windowPoint)
{
    *encoder << windowPoint;
    connection.sendSyncReply(WTFMove(encoder));
}

void RootViewToScreen::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebCore::IntRect& screenFrame)
{
    *encoder << screenFrame;
    connection.sendSyncReply(WTFMove(encoder));
}

void AccessibilityScreenToRootView::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebCore::IntPoint& windowPoint)
{
    *encoder << windowPoint;
    connection.sendSyncReply(WTFMove(encoder));
}

void RootViewToAccessibilityScreen::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebCore::IntRect& screenFrame)
{
    *encoder << screenFrame;
    connection.sendSyncReply(WTFMove(encoder));
}

void RunBeforeUnloadConfirmPanel::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool shouldClose)
{
    *encoder << shouldClose;
    connection.sendSyncReply(WTFMove(encoder));
}

void PrintFrame::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

void DecidePolicyForNavigationActionSync::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebCore::PolicyCheckIdentifier& policyCheckIdentifier, WebCore::PolicyAction policyAction, uint64_t newNavigationID, const WebKit::DownloadID& downloadID, const Optional<WebKit::WebsitePoliciesData>& websitePolicies)
{
    *encoder << policyCheckIdentifier;
    *encoder << policyAction;
    *encoder << newNavigationID;
    *encoder << downloadID;
    *encoder << websitePolicies;
    connection.sendSyncReply(WTFMove(encoder));
}

void HasInsecureContent::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, WebCore::HasInsecureContent hasInsecureContent)
{
    *encoder << hasInsecureContent;
    connection.sendSyncReply(WTFMove(encoder));
}

#if PLATFORM(IOS_FAMILY)

void InterpretKeyEvent::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool handled)
{
    *encoder << handled;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

void BackForwardGoToItem::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::SandboxExtension::Handle& sandboxExtensionHandle, const WebKit::WebBackForwardListCounts& counts)
{
    *encoder << sandboxExtensionHandle;
    *encoder << counts;
    connection.sendSyncReply(WTFMove(encoder));
}

void BackForwardItemAtIndex::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Optional<WebCore::BackForwardItemIdentifier>& itemID)
{
    *encoder << itemID;
    connection.sendSyncReply(WTFMove(encoder));
}

void BackForwardListCounts::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::WebBackForwardListCounts& counts)
{
    *encoder << counts;
    connection.sendSyncReply(WTFMove(encoder));
}

void CanUndoRedo::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool result)
{
    *encoder << result;
    connection.sendSyncReply(WTFMove(encoder));
}

void ExecuteUndoRedo::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

void RequestDOMPasteAccess::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, WebCore::DOMPasteAccessResponse response)
{
    *encoder << response;
    connection.sendSyncReply(WTFMove(encoder));
}

void ExceededDatabaseQuota::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, uint64_t newQuota)
{
    *encoder << newQuota;
    connection.sendSyncReply(WTFMove(encoder));
}

void ReachedApplicationCacheOriginQuota::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, uint64_t newQuota)
{
    *encoder << newQuota;
    connection.sendSyncReply(WTFMove(encoder));
}

#if ENABLE(MEDIA_STREAM)

void EnumerateMediaDevicesForFrame::callReply(IPC::Decoder& decoder, CompletionHandler<void(Vector<WebCore::CaptureDevice>&&, String&&)>&& completionHandler)
{
    Optional<Vector<WebCore::CaptureDevice>> devices;
    decoder >> devices;
    if (!devices) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    Optional<String> mediaDeviceIdentifierHashSalt;
    decoder >> mediaDeviceIdentifierHashSalt;
    if (!mediaDeviceIdentifierHashSalt) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*devices), WTFMove(*mediaDeviceIdentifierHashSalt));
}

void EnumerateMediaDevicesForFrame::cancelReply(CompletionHandler<void(Vector<WebCore::CaptureDevice>&&, String&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Vector<WebCore::CaptureDevice>>::create(), IPC::AsyncReplyError<String>::create());
}

void EnumerateMediaDevicesForFrame::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<WebCore::CaptureDevice>& devices, const String& mediaDeviceIdentifierHashSalt)
{
    *encoder << devices;
    *encoder << mediaDeviceIdentifierHashSalt;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if USE(UNIFIED_TEXT_CHECKING)

void CheckTextOfParagraph::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<WebCore::TextCheckingResult>& results)
{
    *encoder << results;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

void CheckSpellingOfString::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, int32_t misspellingLocation, int32_t misspellingLength)
{
    *encoder << misspellingLocation;
    *encoder << misspellingLength;
    connection.sendSyncReply(WTFMove(encoder));
}

void CheckGrammarOfString::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<WebCore::GrammarDetail>& results, int32_t badGrammarLocation, int32_t badGrammarLength)
{
    *encoder << results;
    *encoder << badGrammarLocation;
    *encoder << badGrammarLength;
    connection.sendSyncReply(WTFMove(encoder));
}

void SpellingUIIsShowing::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool isShowing)
{
    *encoder << isShowing;
    connection.sendSyncReply(WTFMove(encoder));
}

void GetGuessesForWord::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<String>& guesses)
{
    *encoder << guesses;
    connection.sendSyncReply(WTFMove(encoder));
}

#if PLATFORM(COCOA)

void ExecuteSavedCommandBySelector::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool handled)
{
    *encoder << handled;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void GetIsSpeaking::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool isSpeaking)
{
    *encoder << isSpeaking;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if USE(APPKIT)

void SubstitutionsPanelIsShowing::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool isShowing)
{
    *encoder << isShowing;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(MAC)

void DismissCorrectionPanelSoon::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const String& result)
{
    *encoder << result;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if USE(DICTATION_ALTERNATIVES)

void DictationAlternatives::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<String>& alternatives)
{
    *encoder << alternatives;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(X11)

void CreatePluginContainer::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, uint64_t windowID)
{
    *encoder << windowID;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

void LoadRecentSearches::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<WebCore::RecentSearch>& result)
{
    *encoder << result;
    connection.sendSyncReply(WTFMove(encoder));
}

#if ENABLE(PDFKIT_PLUGIN)

void ShowPDFContextMenu::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Optional<int32_t>& selectedItem)
{
    *encoder << selectedItem;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(NETSCAPE_PLUGIN_API)

void FindPlugin::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, uint64_t pluginProcessToken, const String& newMIMEType, uint32_t pluginLoadPolicy, const String& unavailabilityDescription, bool isUnsupported)
{
    *encoder << pluginProcessToken;
    *encoder << newMIMEType;
    *encoder << pluginLoadPolicy;
    *encoder << unavailabilityDescription;
    *encoder << isUnsupported;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(WEB_CRYPTO)

void WrapCryptoKey::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool succeeded, const Vector<uint8_t>& wrappedKey)
{
    *encoder << succeeded;
    *encoder << wrappedKey;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(WEB_CRYPTO)

void UnwrapCryptoKey::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool succeeded, const Vector<uint8_t>& key)
{
    *encoder << succeeded;
    *encoder << key;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if USE(QUICK_LOOK)

void RequestPasswordForQuickLookDocumentInMainFrame::callReply(IPC::Decoder& decoder, CompletionHandler<void(String&&)>&& completionHandler)
{
    Optional<String> password;
    decoder >> password;
    if (!password) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*password));
}

void RequestPasswordForQuickLookDocumentInMainFrame::cancelReply(CompletionHandler<void(String&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<String>::create());
}

void RequestPasswordForQuickLookDocumentInMainFrame::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const String& password)
{
    *encoder << password;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

void HandleSynchronousMessage::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::UserData& returnData)
{
    *encoder << returnData;
    connection.sendSyncReply(WTFMove(encoder));
}

void LoadSynchronousURLSchemeTask::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebCore::ResourceResponse& response, const WebCore::ResourceError& error, const Vector<char>& data)
{
    *encoder << response;
    *encoder << error;
    *encoder << data;
    connection.sendSyncReply(WTFMove(encoder));
}

#if ENABLE(DEVICE_ORIENTATION)

void ShouldAllowDeviceOrientationAndMotionAccess::callReply(IPC::Decoder& decoder, CompletionHandler<void(WebCore::DeviceOrientationOrMotionPermissionState&&)>&& completionHandler)
{
    Optional<WebCore::DeviceOrientationOrMotionPermissionState> permissionState;
    decoder >> permissionState;
    if (!permissionState) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*permissionState));
}

void ShouldAllowDeviceOrientationAndMotionAccess::cancelReply(CompletionHandler<void(WebCore::DeviceOrientationOrMotionPermissionState&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<WebCore::DeviceOrientationOrMotionPermissionState>::create());
}

void ShouldAllowDeviceOrientationAndMotionAccess::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, WebCore::DeviceOrientationOrMotionPermissionState permissionState)
{
    *encoder << permissionState;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(ATTACHMENT_ELEMENT)

void SerializedAttachmentDataForIdentifiers::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<WebCore::SerializedAttachmentData>& seralizedData)
{
    *encoder << seralizedData;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

void SignedPublicKeyAndChallengeString::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const String& result)
{
    *encoder << result;
    connection.sendSyncReply(WTFMove(encoder));
}

#if ENABLE(SPEECH_SYNTHESIS)

void SpeechSynthesisVoiceList::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<WebKit::WebSpeechSynthesisVoice>& voiceList)
{
    *encoder << voiceList;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(SPEECH_SYNTHESIS)

void SpeechSynthesisSpeak::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SpeechSynthesisSpeak::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SpeechSynthesisSpeak::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(SPEECH_SYNTHESIS)

void SpeechSynthesisSetFinishedCallback::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SpeechSynthesisSetFinishedCallback::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SpeechSynthesisSetFinishedCallback::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(SPEECH_SYNTHESIS)

void SpeechSynthesisPause::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SpeechSynthesisPause::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SpeechSynthesisPause::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(SPEECH_SYNTHESIS)

void SpeechSynthesisResume::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SpeechSynthesisResume::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SpeechSynthesisResume::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(GTK)

void ShowEmojiPicker::callReply(IPC::Decoder& decoder, CompletionHandler<void(String&&)>&& completionHandler)
{
    Optional<String> result;
    decoder >> result;
    if (!result) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*result));
}

void ShowEmojiPicker::cancelReply(CompletionHandler<void(String&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<String>::create());
}

void ShowEmojiPicker::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const String& result)
{
    *encoder << result;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(GTK) || PLATFORM(WPE)

void SendMessageToWebViewWithReply::callReply(IPC::Decoder& decoder, CompletionHandler<void(WebKit::UserMessage&&)>&& completionHandler)
{
    Optional<WebKit::UserMessage> replyMessage;
    decoder >> replyMessage;
    if (!replyMessage) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*replyMessage));
}

void SendMessageToWebViewWithReply::cancelReply(CompletionHandler<void(WebKit::UserMessage&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<WebKit::UserMessage>::create());
}

void SendMessageToWebViewWithReply::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::UserMessage& replyMessage)
{
    *encoder << replyMessage;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

} // namespace WebPageProxy

} // namespace Messages

namespace WebKit {

void WebPageProxy::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    auto protectedThis = makeRef(*this);
    if (decoder.messageName() == Messages::WebPageProxy::ShowPage::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ShowPage>(decoder, this, &WebPageProxy::showPage);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::ClosePage::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ClosePage>(decoder, this, &WebPageProxy::closePage);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::MouseDidMoveOverElement::name()) {
        IPC::handleMessage<Messages::WebPageProxy::MouseDidMoveOverElement>(decoder, this, &WebPageProxy::mouseDidMoveOverElement);
        return;
    }
#if ENABLE(NETSCAPE_PLUGIN_API)
    if (decoder.messageName() == Messages::WebPageProxy::UnavailablePluginButtonClicked::name()) {
        IPC::handleMessage<Messages::WebPageProxy::UnavailablePluginButtonClicked>(decoder, this, &WebPageProxy::unavailablePluginButtonClicked);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::DidChangeViewportProperties::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidChangeViewportProperties>(decoder, this, &WebPageProxy::didChangeViewportProperties);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidReceiveEvent::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidReceiveEvent>(decoder, this, &WebPageProxy::didReceiveEvent);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::SetCursor::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetCursor>(decoder, this, &WebPageProxy::setCursor);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::SetCursorHiddenUntilMouseMoves::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetCursorHiddenUntilMouseMoves>(decoder, this, &WebPageProxy::setCursorHiddenUntilMouseMoves);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::SetStatusText::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetStatusText>(decoder, this, &WebPageProxy::setStatusText);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::SetFocus::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetFocus>(decoder, this, &WebPageProxy::setFocus);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::TakeFocus::name()) {
        IPC::handleMessage<Messages::WebPageProxy::TakeFocus>(decoder, this, &WebPageProxy::takeFocus);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::FocusedFrameChanged::name()) {
        IPC::handleMessage<Messages::WebPageProxy::FocusedFrameChanged>(decoder, this, &WebPageProxy::focusedFrameChanged);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::FrameSetLargestFrameChanged::name()) {
        IPC::handleMessage<Messages::WebPageProxy::FrameSetLargestFrameChanged>(decoder, this, &WebPageProxy::frameSetLargestFrameChanged);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::SetRenderTreeSize::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetRenderTreeSize>(decoder, this, &WebPageProxy::setRenderTreeSize);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::SetToolbarsAreVisible::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetToolbarsAreVisible>(decoder, this, &WebPageProxy::setToolbarsAreVisible);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::SetMenuBarIsVisible::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetMenuBarIsVisible>(decoder, this, &WebPageProxy::setMenuBarIsVisible);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::SetStatusBarIsVisible::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetStatusBarIsVisible>(decoder, this, &WebPageProxy::setStatusBarIsVisible);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::SetIsResizable::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetIsResizable>(decoder, this, &WebPageProxy::setIsResizable);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::SetWindowFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetWindowFrame>(decoder, this, &WebPageProxy::setWindowFrame);
        return;
    }
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPageProxy::ShowValidationMessage::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ShowValidationMessage>(decoder, this, &WebPageProxy::showValidationMessage);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPageProxy::HideValidationMessage::name()) {
        IPC::handleMessage<Messages::WebPageProxy::HideValidationMessage>(decoder, this, &WebPageProxy::hideValidationMessage);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::PageDidScroll::name()) {
        IPC::handleMessage<Messages::WebPageProxy::PageDidScroll>(decoder, this, &WebPageProxy::pageDidScroll);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::RunOpenPanel::name()) {
        IPC::handleMessage<Messages::WebPageProxy::RunOpenPanel>(decoder, this, &WebPageProxy::runOpenPanel);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::ShowShareSheet::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ShowShareSheet>(decoder, this, &WebPageProxy::showShareSheet);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::RunModal::name()) {
        IPC::handleMessage<Messages::WebPageProxy::RunModal>(decoder, this, &WebPageProxy::runModal);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::NotifyScrollerThumbIsVisibleInRect::name()) {
        IPC::handleMessage<Messages::WebPageProxy::NotifyScrollerThumbIsVisibleInRect>(decoder, this, &WebPageProxy::notifyScrollerThumbIsVisibleInRect);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::RecommendedScrollbarStyleDidChange::name()) {
        IPC::handleMessage<Messages::WebPageProxy::RecommendedScrollbarStyleDidChange>(decoder, this, &WebPageProxy::recommendedScrollbarStyleDidChange);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidChangeScrollbarsForMainFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidChangeScrollbarsForMainFrame>(decoder, this, &WebPageProxy::didChangeScrollbarsForMainFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidChangeScrollOffsetPinningForMainFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidChangeScrollOffsetPinningForMainFrame>(decoder, this, &WebPageProxy::didChangeScrollOffsetPinningForMainFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidChangePageCount::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidChangePageCount>(decoder, this, &WebPageProxy::didChangePageCount);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::PageExtendedBackgroundColorDidChange::name()) {
        IPC::handleMessage<Messages::WebPageProxy::PageExtendedBackgroundColorDidChange>(decoder, this, &WebPageProxy::pageExtendedBackgroundColorDidChange);
        return;
    }
#if ENABLE(NETSCAPE_PLUGIN_API)
    if (decoder.messageName() == Messages::WebPageProxy::DidFailToInitializePlugin::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidFailToInitializePlugin>(decoder, this, &WebPageProxy::didFailToInitializePlugin);
        return;
    }
#endif
#if ENABLE(NETSCAPE_PLUGIN_API)
    if (decoder.messageName() == Messages::WebPageProxy::DidBlockInsecurePluginVersion::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidBlockInsecurePluginVersion>(decoder, this, &WebPageProxy::didBlockInsecurePluginVersion);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::SetCanShortCircuitHorizontalWheelEvents::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetCanShortCircuitHorizontalWheelEvents>(decoder, this, &WebPageProxy::setCanShortCircuitHorizontalWheelEvents);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidChangeContentSize::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidChangeContentSize>(decoder, this, &WebPageProxy::didChangeContentSize);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidChangeIntrinsicContentSize::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidChangeIntrinsicContentSize>(decoder, this, &WebPageProxy::didChangeIntrinsicContentSize);
        return;
    }
#if ENABLE(INPUT_TYPE_COLOR)
    if (decoder.messageName() == Messages::WebPageProxy::ShowColorPicker::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ShowColorPicker>(decoder, this, &WebPageProxy::showColorPicker);
        return;
    }
#endif
#if ENABLE(INPUT_TYPE_COLOR)
    if (decoder.messageName() == Messages::WebPageProxy::SetColorPickerColor::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetColorPickerColor>(decoder, this, &WebPageProxy::setColorPickerColor);
        return;
    }
#endif
#if ENABLE(INPUT_TYPE_COLOR)
    if (decoder.messageName() == Messages::WebPageProxy::EndColorPicker::name()) {
        IPC::handleMessage<Messages::WebPageProxy::EndColorPicker>(decoder, this, &WebPageProxy::endColorPicker);
        return;
    }
#endif
#if ENABLE(DATALIST_ELEMENT)
    if (decoder.messageName() == Messages::WebPageProxy::ShowDataListSuggestions::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ShowDataListSuggestions>(decoder, this, &WebPageProxy::showDataListSuggestions);
        return;
    }
#endif
#if ENABLE(DATALIST_ELEMENT)
    if (decoder.messageName() == Messages::WebPageProxy::HandleKeydownInDataList::name()) {
        IPC::handleMessage<Messages::WebPageProxy::HandleKeydownInDataList>(decoder, this, &WebPageProxy::handleKeydownInDataList);
        return;
    }
#endif
#if ENABLE(DATALIST_ELEMENT)
    if (decoder.messageName() == Messages::WebPageProxy::EndDataListSuggestions::name()) {
        IPC::handleMessage<Messages::WebPageProxy::EndDataListSuggestions>(decoder, this, &WebPageProxy::endDataListSuggestions);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::DecidePolicyForResponse::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DecidePolicyForResponse>(decoder, this, &WebPageProxy::decidePolicyForResponse);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DecidePolicyForNavigationActionAsync::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DecidePolicyForNavigationActionAsync>(decoder, this, &WebPageProxy::decidePolicyForNavigationActionAsync);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DecidePolicyForNewWindowAction::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DecidePolicyForNewWindowAction>(decoder, this, &WebPageProxy::decidePolicyForNewWindowAction);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::UnableToImplementPolicy::name()) {
        IPC::handleMessage<Messages::WebPageProxy::UnableToImplementPolicy>(decoder, this, &WebPageProxy::unableToImplementPolicy);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidChangeProgress::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidChangeProgress>(decoder, this, &WebPageProxy::didChangeProgress);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidFinishProgress::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidFinishProgress>(decoder, this, &WebPageProxy::didFinishProgress);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidStartProgress::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidStartProgress>(decoder, this, &WebPageProxy::didStartProgress);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::SetNetworkRequestsInProgress::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetNetworkRequestsInProgress>(decoder, this, &WebPageProxy::setNetworkRequestsInProgress);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidCreateMainFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidCreateMainFrame>(decoder, this, &WebPageProxy::didCreateMainFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidCreateSubframe::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidCreateSubframe>(decoder, this, &WebPageProxy::didCreateSubframe);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidCreateWindow::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidCreateWindow>(decoder, this, &WebPageProxy::didCreateWindow);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidStartProvisionalLoadForFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidStartProvisionalLoadForFrame>(decoder, this, &WebPageProxy::didStartProvisionalLoadForFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidReceiveServerRedirectForProvisionalLoadForFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidReceiveServerRedirectForProvisionalLoadForFrame>(decoder, this, &WebPageProxy::didReceiveServerRedirectForProvisionalLoadForFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::WillPerformClientRedirectForFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::WillPerformClientRedirectForFrame>(decoder, this, &WebPageProxy::willPerformClientRedirectForFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidCancelClientRedirectForFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidCancelClientRedirectForFrame>(decoder, this, &WebPageProxy::didCancelClientRedirectForFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidChangeProvisionalURLForFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidChangeProvisionalURLForFrame>(decoder, this, &WebPageProxy::didChangeProvisionalURLForFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidFailProvisionalLoadForFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidFailProvisionalLoadForFrame>(decoder, this, &WebPageProxy::didFailProvisionalLoadForFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidCommitLoadForFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidCommitLoadForFrame>(decoder, this, &WebPageProxy::didCommitLoadForFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidFailLoadForFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidFailLoadForFrame>(decoder, this, &WebPageProxy::didFailLoadForFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidFinishDocumentLoadForFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidFinishDocumentLoadForFrame>(decoder, this, &WebPageProxy::didFinishDocumentLoadForFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidFinishLoadForFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidFinishLoadForFrame>(decoder, this, &WebPageProxy::didFinishLoadForFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidFirstLayoutForFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidFirstLayoutForFrame>(decoder, this, &WebPageProxy::didFirstLayoutForFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidFirstVisuallyNonEmptyLayoutForFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidFirstVisuallyNonEmptyLayoutForFrame>(decoder, this, &WebPageProxy::didFirstVisuallyNonEmptyLayoutForFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidReachLayoutMilestone::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidReachLayoutMilestone>(decoder, this, &WebPageProxy::didReachLayoutMilestone);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidReceiveTitleForFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidReceiveTitleForFrame>(decoder, this, &WebPageProxy::didReceiveTitleForFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidDisplayInsecureContentForFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidDisplayInsecureContentForFrame>(decoder, this, &WebPageProxy::didDisplayInsecureContentForFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidRunInsecureContentForFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidRunInsecureContentForFrame>(decoder, this, &WebPageProxy::didRunInsecureContentForFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidDetectXSSForFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidDetectXSSForFrame>(decoder, this, &WebPageProxy::didDetectXSSForFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidSameDocumentNavigationForFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidSameDocumentNavigationForFrame>(decoder, this, &WebPageProxy::didSameDocumentNavigationForFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidChangeMainDocument::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidChangeMainDocument>(decoder, this, &WebPageProxy::didChangeMainDocument);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidExplicitOpenForFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidExplicitOpenForFrame>(decoder, this, &WebPageProxy::didExplicitOpenForFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidDestroyNavigation::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidDestroyNavigation>(decoder, this, &WebPageProxy::didDestroyNavigation);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::MainFramePluginHandlesPageScaleGestureDidChange::name()) {
        IPC::handleMessage<Messages::WebPageProxy::MainFramePluginHandlesPageScaleGestureDidChange>(decoder, this, &WebPageProxy::mainFramePluginHandlesPageScaleGestureDidChange);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::FrameDidBecomeFrameSet::name()) {
        IPC::handleMessage<Messages::WebPageProxy::FrameDidBecomeFrameSet>(decoder, this, &WebPageProxy::frameDidBecomeFrameSet);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidNavigateWithNavigationData::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidNavigateWithNavigationData>(decoder, this, &WebPageProxy::didNavigateWithNavigationData);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidPerformClientRedirect::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidPerformClientRedirect>(decoder, this, &WebPageProxy::didPerformClientRedirect);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidPerformServerRedirect::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidPerformServerRedirect>(decoder, this, &WebPageProxy::didPerformServerRedirect);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidUpdateHistoryTitle::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidUpdateHistoryTitle>(decoder, this, &WebPageProxy::didUpdateHistoryTitle);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidFinishLoadingDataForCustomContentProvider::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidFinishLoadingDataForCustomContentProvider>(decoder, this, &WebPageProxy::didFinishLoadingDataForCustomContentProvider);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::WillSubmitForm::name()) {
        IPC::handleMessage<Messages::WebPageProxy::WillSubmitForm>(decoder, this, &WebPageProxy::willSubmitForm);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::VoidCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::VoidCallback>(decoder, this, &WebPageProxy::voidCallback);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DataCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DataCallback>(decoder, this, &WebPageProxy::dataCallback);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::ImageCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ImageCallback>(decoder, this, &WebPageProxy::imageCallback);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::StringCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::StringCallback>(decoder, this, &WebPageProxy::stringCallback);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::BoolCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::BoolCallback>(decoder, this, &WebPageProxy::boolCallback);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::InvalidateStringCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::InvalidateStringCallback>(decoder, this, &WebPageProxy::invalidateStringCallback);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::ScriptValueCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ScriptValueCallback>(decoder, this, &WebPageProxy::scriptValueCallback);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::ComputedPagesCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ComputedPagesCallback>(decoder, this, &WebPageProxy::computedPagesCallback);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::ValidateCommandCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ValidateCommandCallback>(decoder, this, &WebPageProxy::validateCommandCallback);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::EditingRangeCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::EditingRangeCallback>(decoder, this, &WebPageProxy::editingRangeCallback);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::UnsignedCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::UnsignedCallback>(decoder, this, &WebPageProxy::unsignedCallback);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::RectForCharacterRangeCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::RectForCharacterRangeCallback>(decoder, this, &WebPageProxy::rectForCharacterRangeCallback);
        return;
    }
#if ENABLE(APPLICATION_MANIFEST)
    if (decoder.messageName() == Messages::WebPageProxy::ApplicationManifestCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ApplicationManifestCallback>(decoder, this, &WebPageProxy::applicationManifestCallback);
        return;
    }
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebPageProxy::AttributedStringForCharacterRangeCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::AttributedStringForCharacterRangeCallback>(decoder, this, &WebPageProxy::attributedStringForCharacterRangeCallback);
        return;
    }
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebPageProxy::FontAtSelectionCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::FontAtSelectionCallback>(decoder, this, &WebPageProxy::fontAtSelectionCallback);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::FontAttributesCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::FontAttributesCallback>(decoder, this, &WebPageProxy::fontAttributesCallback);
        return;
    }
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::GestureCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::GestureCallback>(decoder, this, &WebPageProxy::gestureCallback);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::TouchesCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::TouchesCallback>(decoder, this, &WebPageProxy::touchesCallback);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::SelectionContextCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SelectionContextCallback>(decoder, this, &WebPageProxy::selectionContextCallback);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::DidReceivePositionInformation::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidReceivePositionInformation>(decoder, this, &WebPageProxy::didReceivePositionInformation);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::SaveImageToLibrary::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SaveImageToLibrary>(decoder, this, &WebPageProxy::saveImageToLibrary);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::ShowPlaybackTargetPicker::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ShowPlaybackTargetPicker>(decoder, this, &WebPageProxy::showPlaybackTargetPicker);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::CommitPotentialTapFailed::name()) {
        IPC::handleMessage<Messages::WebPageProxy::CommitPotentialTapFailed>(decoder, this, &WebPageProxy::commitPotentialTapFailed);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::DidNotHandleTapAsClick::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidNotHandleTapAsClick>(decoder, this, &WebPageProxy::didNotHandleTapAsClick);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::DidCompleteSyntheticClick::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidCompleteSyntheticClick>(decoder, this, &WebPageProxy::didCompleteSyntheticClick);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::DisableDoubleTapGesturesDuringTapIfNecessary::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DisableDoubleTapGesturesDuringTapIfNecessary>(decoder, this, &WebPageProxy::disableDoubleTapGesturesDuringTapIfNecessary);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::HandleSmartMagnificationInformationForPotentialTap::name()) {
        IPC::handleMessage<Messages::WebPageProxy::HandleSmartMagnificationInformationForPotentialTap>(decoder, this, &WebPageProxy::handleSmartMagnificationInformationForPotentialTap);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::SelectionRectsCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SelectionRectsCallback>(decoder, this, &WebPageProxy::selectionRectsCallback);
        return;
    }
#endif
#if ENABLE(DATA_DETECTION)
    if (decoder.messageName() == Messages::WebPageProxy::SetDataDetectionResult::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetDataDetectionResult>(decoder, this, &WebPageProxy::setDataDetectionResult);
        return;
    }
#endif
#if PLATFORM(GTK)
    if (decoder.messageName() == Messages::WebPageProxy::PrintFinishedCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::PrintFinishedCallback>(decoder, this, &WebPageProxy::printFinishedCallback);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPageProxy::DrawToPDFCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DrawToPDFCallback>(decoder, this, &WebPageProxy::drawToPDFCallback);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPageProxy::MachSendRightCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::MachSendRightCallback>(decoder, this, &WebPageProxy::machSendRightCallback);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPageProxy::NowPlayingInfoCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::NowPlayingInfoCallback>(decoder, this, &WebPageProxy::nowPlayingInfoCallback);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::FindStringCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::FindStringCallback>(decoder, this, &WebPageProxy::findStringCallback);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::PageScaleFactorDidChange::name()) {
        IPC::handleMessage<Messages::WebPageProxy::PageScaleFactorDidChange>(decoder, this, &WebPageProxy::pageScaleFactorDidChange);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::PluginScaleFactorDidChange::name()) {
        IPC::handleMessage<Messages::WebPageProxy::PluginScaleFactorDidChange>(decoder, this, &WebPageProxy::pluginScaleFactorDidChange);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::PluginZoomFactorDidChange::name()) {
        IPC::handleMessage<Messages::WebPageProxy::PluginZoomFactorDidChange>(decoder, this, &WebPageProxy::pluginZoomFactorDidChange);
        return;
    }
#if USE(ATK)
    if (decoder.messageName() == Messages::WebPageProxy::BindAccessibilityTree::name()) {
        IPC::handleMessage<Messages::WebPageProxy::BindAccessibilityTree>(decoder, this, &WebPageProxy::bindAccessibilityTree);
        return;
    }
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    if (decoder.messageName() == Messages::WebPageProxy::SetInputMethodState::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetInputMethodState>(decoder, this, &WebPageProxy::setInputMethodState);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::BackForwardAddItem::name()) {
        IPC::handleMessage<Messages::WebPageProxy::BackForwardAddItem>(decoder, this, &WebPageProxy::backForwardAddItem);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::BackForwardClear::name()) {
        IPC::handleMessage<Messages::WebPageProxy::BackForwardClear>(decoder, this, &WebPageProxy::backForwardClear);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::WillGoToBackForwardListItem::name()) {
        IPC::handleMessage<Messages::WebPageProxy::WillGoToBackForwardListItem>(decoder, this, &WebPageProxy::willGoToBackForwardListItem);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::RegisterEditCommandForUndo::name()) {
        IPC::handleMessage<Messages::WebPageProxy::RegisterEditCommandForUndo>(decoder, this, &WebPageProxy::registerEditCommandForUndo);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::ClearAllEditCommands::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ClearAllEditCommands>(decoder, this, &WebPageProxy::clearAllEditCommands);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::RegisterInsertionUndoGrouping::name()) {
        IPC::handleMessage<Messages::WebPageProxy::RegisterInsertionUndoGrouping>(decoder, this, &WebPageProxy::registerInsertionUndoGrouping);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::LogDiagnosticMessage::name()) {
        IPC::handleMessage<Messages::WebPageProxy::LogDiagnosticMessage>(decoder, this, &WebPageProxy::logDiagnosticMessage);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::LogDiagnosticMessageWithResult::name()) {
        IPC::handleMessage<Messages::WebPageProxy::LogDiagnosticMessageWithResult>(decoder, this, &WebPageProxy::logDiagnosticMessageWithResult);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::LogDiagnosticMessageWithValue::name()) {
        IPC::handleMessage<Messages::WebPageProxy::LogDiagnosticMessageWithValue>(decoder, this, &WebPageProxy::logDiagnosticMessageWithValue);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::LogDiagnosticMessageWithEnhancedPrivacy::name()) {
        IPC::handleMessage<Messages::WebPageProxy::LogDiagnosticMessageWithEnhancedPrivacy>(decoder, this, &WebPageProxy::logDiagnosticMessageWithEnhancedPrivacy);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::LogDiagnosticMessageWithValueDictionary::name()) {
        IPC::handleMessage<Messages::WebPageProxy::LogDiagnosticMessageWithValueDictionary>(decoder, this, &WebPageProxy::logDiagnosticMessageWithValueDictionary);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::LogScrollingEvent::name()) {
        IPC::handleMessage<Messages::WebPageProxy::LogScrollingEvent>(decoder, this, &WebPageProxy::logScrollingEvent);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::EditorStateChanged::name()) {
        IPC::handleMessage<Messages::WebPageProxy::EditorStateChanged>(decoder, this, &WebPageProxy::editorStateChanged);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::CompositionWasCanceled::name()) {
        IPC::handleMessage<Messages::WebPageProxy::CompositionWasCanceled>(decoder, this, &WebPageProxy::compositionWasCanceled);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::SetHasHadSelectionChangesFromUserInteraction::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetHasHadSelectionChangesFromUserInteraction>(decoder, this, &WebPageProxy::setHasHadSelectionChangesFromUserInteraction);
        return;
    }
#if HAVE(TOUCH_BAR)
    if (decoder.messageName() == Messages::WebPageProxy::SetIsTouchBarUpdateSupressedForHiddenContentEditable::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetIsTouchBarUpdateSupressedForHiddenContentEditable>(decoder, this, &WebPageProxy::setIsTouchBarUpdateSupressedForHiddenContentEditable);
        return;
    }
#endif
#if HAVE(TOUCH_BAR)
    if (decoder.messageName() == Messages::WebPageProxy::SetIsNeverRichlyEditableForTouchBar::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetIsNeverRichlyEditableForTouchBar>(decoder, this, &WebPageProxy::setIsNeverRichlyEditableForTouchBar);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::DidCountStringMatches::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidCountStringMatches>(decoder, this, &WebPageProxy::didCountStringMatches);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::SetTextIndicator::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetTextIndicator>(decoder, this, &WebPageProxy::setTextIndicator);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::ClearTextIndicator::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ClearTextIndicator>(decoder, this, &WebPageProxy::clearTextIndicator);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidFindString::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidFindString>(decoder, this, &WebPageProxy::didFindString);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidFailToFindString::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidFailToFindString>(decoder, this, &WebPageProxy::didFailToFindString);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidFindStringMatches::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidFindStringMatches>(decoder, this, &WebPageProxy::didFindStringMatches);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidGetImageForFindMatch::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidGetImageForFindMatch>(decoder, this, &WebPageProxy::didGetImageForFindMatch);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::ShowPopupMenu::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ShowPopupMenu>(decoder, this, &WebPageProxy::showPopupMenu);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::HidePopupMenu::name()) {
        IPC::handleMessage<Messages::WebPageProxy::HidePopupMenu>(decoder, this, &WebPageProxy::hidePopupMenu);
        return;
    }
#if ENABLE(CONTEXT_MENUS)
    if (decoder.messageName() == Messages::WebPageProxy::ShowContextMenu::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ShowContextMenu>(decoder, this, &WebPageProxy::showContextMenu);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::RequestGeolocationPermissionForFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::RequestGeolocationPermissionForFrame>(decoder, this, &WebPageProxy::requestGeolocationPermissionForFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::RevokeGeolocationAuthorizationToken::name()) {
        IPC::handleMessage<Messages::WebPageProxy::RevokeGeolocationAuthorizationToken>(decoder, this, &WebPageProxy::revokeGeolocationAuthorizationToken);
        return;
    }
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::WebPageProxy::RequestUserMediaPermissionForFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::RequestUserMediaPermissionForFrame>(decoder, this, &WebPageProxy::requestUserMediaPermissionForFrame);
        return;
    }
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::WebPageProxy::EnumerateMediaDevicesForFrame::name()) {
        IPC::handleMessageAsync<Messages::WebPageProxy::EnumerateMediaDevicesForFrame>(connection, decoder, this, &WebPageProxy::enumerateMediaDevicesForFrame);
        return;
    }
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::WebPageProxy::BeginMonitoringCaptureDevices::name()) {
        IPC::handleMessage<Messages::WebPageProxy::BeginMonitoringCaptureDevices>(decoder, this, &WebPageProxy::beginMonitoringCaptureDevices);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::RequestNotificationPermission::name()) {
        IPC::handleMessage<Messages::WebPageProxy::RequestNotificationPermission>(decoder, this, &WebPageProxy::requestNotificationPermission);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::ShowNotification::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ShowNotification>(decoder, this, &WebPageProxy::showNotification);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::CancelNotification::name()) {
        IPC::handleMessage<Messages::WebPageProxy::CancelNotification>(decoder, this, &WebPageProxy::cancelNotification);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::ClearNotifications::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ClearNotifications>(decoder, this, &WebPageProxy::clearNotifications);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidDestroyNotification::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidDestroyNotification>(decoder, this, &WebPageProxy::didDestroyNotification);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::UpdateSpellingUIWithMisspelledWord::name()) {
        IPC::handleMessage<Messages::WebPageProxy::UpdateSpellingUIWithMisspelledWord>(decoder, this, &WebPageProxy::updateSpellingUIWithMisspelledWord);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::UpdateSpellingUIWithGrammarString::name()) {
        IPC::handleMessage<Messages::WebPageProxy::UpdateSpellingUIWithGrammarString>(decoder, this, &WebPageProxy::updateSpellingUIWithGrammarString);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::LearnWord::name()) {
        IPC::handleMessage<Messages::WebPageProxy::LearnWord>(decoder, this, &WebPageProxy::learnWord);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::IgnoreWord::name()) {
        IPC::handleMessage<Messages::WebPageProxy::IgnoreWord>(decoder, this, &WebPageProxy::ignoreWord);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::RequestCheckingOfString::name()) {
        IPC::handleMessage<Messages::WebPageProxy::RequestCheckingOfString>(decoder, this, &WebPageProxy::requestCheckingOfString);
        return;
    }
#if ENABLE(DRAG_SUPPORT)
    if (decoder.messageName() == Messages::WebPageProxy::DidPerformDragControllerAction::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidPerformDragControllerAction>(decoder, this, &WebPageProxy::didPerformDragControllerAction);
        return;
    }
#endif
#if ENABLE(DRAG_SUPPORT)
    if (decoder.messageName() == Messages::WebPageProxy::DidEndDragging::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidEndDragging>(decoder, this, &WebPageProxy::didEndDragging);
        return;
    }
#endif
#if PLATFORM(COCOA) && ENABLE(DRAG_SUPPORT)
    if (decoder.messageName() == Messages::WebPageProxy::StartDrag::name()) {
        IPC::handleMessage<Messages::WebPageProxy::StartDrag>(decoder, this, &WebPageProxy::startDrag);
        return;
    }
#endif
#if PLATFORM(COCOA) && ENABLE(DRAG_SUPPORT)
    if (decoder.messageName() == Messages::WebPageProxy::SetPromisedDataForImage::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetPromisedDataForImage>(decoder, this, &WebPageProxy::setPromisedDataForImage);
        return;
    }
#endif
#if PLATFORM(GTK) && ENABLE(DRAG_SUPPORT)
    if (decoder.messageName() == Messages::WebPageProxy::StartDrag::name()) {
        IPC::handleMessage<Messages::WebPageProxy::StartDrag>(decoder, this, &WebPageProxy::startDrag);
        return;
    }
#endif
#if ENABLE(DRAG_SUPPORT)
    if (decoder.messageName() == Messages::WebPageProxy::DidPerformDragOperation::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidPerformDragOperation>(decoder, this, &WebPageProxy::didPerformDragOperation);
        return;
    }
#endif
#if ENABLE(DATA_INTERACTION)
    if (decoder.messageName() == Messages::WebPageProxy::DidHandleDragStartRequest::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidHandleDragStartRequest>(decoder, this, &WebPageProxy::didHandleDragStartRequest);
        return;
    }
#endif
#if ENABLE(DATA_INTERACTION)
    if (decoder.messageName() == Messages::WebPageProxy::DidHandleAdditionalDragItemsRequest::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidHandleAdditionalDragItemsRequest>(decoder, this, &WebPageProxy::didHandleAdditionalDragItemsRequest);
        return;
    }
#endif
#if ENABLE(DATA_INTERACTION)
    if (decoder.messageName() == Messages::WebPageProxy::WillReceiveEditDragSnapshot::name()) {
        IPC::handleMessage<Messages::WebPageProxy::WillReceiveEditDragSnapshot>(decoder, this, &WebPageProxy::willReceiveEditDragSnapshot);
        return;
    }
#endif
#if ENABLE(DATA_INTERACTION)
    if (decoder.messageName() == Messages::WebPageProxy::DidReceiveEditDragSnapshot::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidReceiveEditDragSnapshot>(decoder, this, &WebPageProxy::didReceiveEditDragSnapshot);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPageProxy::DidPerformDictionaryLookup::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidPerformDictionaryLookup>(decoder, this, &WebPageProxy::didPerformDictionaryLookup);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPageProxy::RegisterWebProcessAccessibilityToken::name()) {
        IPC::handleMessage<Messages::WebPageProxy::RegisterWebProcessAccessibilityToken>(decoder, this, &WebPageProxy::registerWebProcessAccessibilityToken);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPageProxy::PluginFocusOrWindowFocusChanged::name()) {
        IPC::handleMessage<Messages::WebPageProxy::PluginFocusOrWindowFocusChanged>(decoder, this, &WebPageProxy::pluginFocusOrWindowFocusChanged);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPageProxy::SetPluginComplexTextInputState::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetPluginComplexTextInputState>(decoder, this, &WebPageProxy::setPluginComplexTextInputState);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPageProxy::Speak::name()) {
        IPC::handleMessage<Messages::WebPageProxy::Speak>(decoder, this, &WebPageProxy::speak);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPageProxy::StopSpeaking::name()) {
        IPC::handleMessage<Messages::WebPageProxy::StopSpeaking>(decoder, this, &WebPageProxy::stopSpeaking);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPageProxy::MakeFirstResponder::name()) {
        IPC::handleMessage<Messages::WebPageProxy::MakeFirstResponder>(decoder, this, &WebPageProxy::makeFirstResponder);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPageProxy::AssistiveTechnologyMakeFirstResponder::name()) {
        IPC::handleMessage<Messages::WebPageProxy::AssistiveTechnologyMakeFirstResponder>(decoder, this, &WebPageProxy::assistiveTechnologyMakeFirstResponder);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPageProxy::SearchWithSpotlight::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SearchWithSpotlight>(decoder, this, &WebPageProxy::searchWithSpotlight);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPageProxy::SearchTheWeb::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SearchTheWeb>(decoder, this, &WebPageProxy::searchTheWeb);
        return;
    }
#endif
#if HAVE(TOUCH_BAR)
    if (decoder.messageName() == Messages::WebPageProxy::TouchBarMenuDataChanged::name()) {
        IPC::handleMessage<Messages::WebPageProxy::TouchBarMenuDataChanged>(decoder, this, &WebPageProxy::touchBarMenuDataChanged);
        return;
    }
#endif
#if HAVE(TOUCH_BAR)
    if (decoder.messageName() == Messages::WebPageProxy::TouchBarMenuItemDataAdded::name()) {
        IPC::handleMessage<Messages::WebPageProxy::TouchBarMenuItemDataAdded>(decoder, this, &WebPageProxy::touchBarMenuItemDataAdded);
        return;
    }
#endif
#if HAVE(TOUCH_BAR)
    if (decoder.messageName() == Messages::WebPageProxy::TouchBarMenuItemDataRemoved::name()) {
        IPC::handleMessage<Messages::WebPageProxy::TouchBarMenuItemDataRemoved>(decoder, this, &WebPageProxy::touchBarMenuItemDataRemoved);
        return;
    }
#endif
#if USE(AUTOMATIC_TEXT_REPLACEMENT)
    if (decoder.messageName() == Messages::WebPageProxy::toggleSmartInsertDelete::name()) {
        IPC::handleMessage<Messages::WebPageProxy::toggleSmartInsertDelete>(decoder, this, &WebPageProxy::toggleSmartInsertDelete);
        return;
    }
#endif
#if USE(AUTOMATIC_TEXT_REPLACEMENT)
    if (decoder.messageName() == Messages::WebPageProxy::toggleAutomaticQuoteSubstitution::name()) {
        IPC::handleMessage<Messages::WebPageProxy::toggleAutomaticQuoteSubstitution>(decoder, this, &WebPageProxy::toggleAutomaticQuoteSubstitution);
        return;
    }
#endif
#if USE(AUTOMATIC_TEXT_REPLACEMENT)
    if (decoder.messageName() == Messages::WebPageProxy::toggleAutomaticLinkDetection::name()) {
        IPC::handleMessage<Messages::WebPageProxy::toggleAutomaticLinkDetection>(decoder, this, &WebPageProxy::toggleAutomaticLinkDetection);
        return;
    }
#endif
#if USE(AUTOMATIC_TEXT_REPLACEMENT)
    if (decoder.messageName() == Messages::WebPageProxy::toggleAutomaticDashSubstitution::name()) {
        IPC::handleMessage<Messages::WebPageProxy::toggleAutomaticDashSubstitution>(decoder, this, &WebPageProxy::toggleAutomaticDashSubstitution);
        return;
    }
#endif
#if USE(AUTOMATIC_TEXT_REPLACEMENT)
    if (decoder.messageName() == Messages::WebPageProxy::toggleAutomaticTextReplacement::name()) {
        IPC::handleMessage<Messages::WebPageProxy::toggleAutomaticTextReplacement>(decoder, this, &WebPageProxy::toggleAutomaticTextReplacement);
        return;
    }
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebPageProxy::ShowCorrectionPanel::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ShowCorrectionPanel>(decoder, this, &WebPageProxy::showCorrectionPanel);
        return;
    }
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebPageProxy::DismissCorrectionPanel::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DismissCorrectionPanel>(decoder, this, &WebPageProxy::dismissCorrectionPanel);
        return;
    }
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebPageProxy::RecordAutocorrectionResponse::name()) {
        IPC::handleMessage<Messages::WebPageProxy::RecordAutocorrectionResponse>(decoder, this, &WebPageProxy::recordAutocorrectionResponse);
        return;
    }
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebPageProxy::SetEditableElementIsFocused::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetEditableElementIsFocused>(decoder, this, &WebPageProxy::setEditableElementIsFocused);
        return;
    }
#endif
#if USE(DICTATION_ALTERNATIVES)
    if (decoder.messageName() == Messages::WebPageProxy::ShowDictationAlternativeUI::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ShowDictationAlternativeUI>(decoder, this, &WebPageProxy::showDictationAlternativeUI);
        return;
    }
#endif
#if USE(DICTATION_ALTERNATIVES)
    if (decoder.messageName() == Messages::WebPageProxy::RemoveDictationAlternatives::name()) {
        IPC::handleMessage<Messages::WebPageProxy::RemoveDictationAlternatives>(decoder, this, &WebPageProxy::removeDictationAlternatives);
        return;
    }
#endif
#if PLATFORM(X11)
    if (decoder.messageName() == Messages::WebPageProxy::WindowedPluginGeometryDidChange::name()) {
        IPC::handleMessage<Messages::WebPageProxy::WindowedPluginGeometryDidChange>(decoder, this, &WebPageProxy::windowedPluginGeometryDidChange);
        return;
    }
#endif
#if PLATFORM(X11)
    if (decoder.messageName() == Messages::WebPageProxy::WindowedPluginVisibilityDidChange::name()) {
        IPC::handleMessage<Messages::WebPageProxy::WindowedPluginVisibilityDidChange>(decoder, this, &WebPageProxy::windowedPluginVisibilityDidChange);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::CouldNotRestorePageState::name()) {
        IPC::handleMessage<Messages::WebPageProxy::CouldNotRestorePageState>(decoder, this, &WebPageProxy::couldNotRestorePageState);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::RestorePageState::name()) {
        IPC::handleMessage<Messages::WebPageProxy::RestorePageState>(decoder, this, &WebPageProxy::restorePageState);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::RestorePageCenterAndScale::name()) {
        IPC::handleMessage<Messages::WebPageProxy::RestorePageCenterAndScale>(decoder, this, &WebPageProxy::restorePageCenterAndScale);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::DidGetTapHighlightGeometries::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidGetTapHighlightGeometries>(decoder, this, &WebPageProxy::didGetTapHighlightGeometries);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::ElementDidFocus::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ElementDidFocus>(decoder, this, &WebPageProxy::elementDidFocus);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::ElementDidBlur::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ElementDidBlur>(decoder, this, &WebPageProxy::elementDidBlur);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::UpdateInputContextAfterBlurringAndRefocusingElement::name()) {
        IPC::handleMessage<Messages::WebPageProxy::UpdateInputContextAfterBlurringAndRefocusingElement>(decoder, this, &WebPageProxy::updateInputContextAfterBlurringAndRefocusingElement);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::FocusedElementDidChangeInputMode::name()) {
        IPC::handleMessage<Messages::WebPageProxy::FocusedElementDidChangeInputMode>(decoder, this, &WebPageProxy::focusedElementDidChangeInputMode);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::ScrollingNodeScrollWillStartScroll::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ScrollingNodeScrollWillStartScroll>(decoder, this, &WebPageProxy::scrollingNodeScrollWillStartScroll);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::ScrollingNodeScrollDidEndScroll::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ScrollingNodeScrollDidEndScroll>(decoder, this, &WebPageProxy::scrollingNodeScrollDidEndScroll);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::ShowInspectorHighlight::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ShowInspectorHighlight>(decoder, this, &WebPageProxy::showInspectorHighlight);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::HideInspectorHighlight::name()) {
        IPC::handleMessage<Messages::WebPageProxy::HideInspectorHighlight>(decoder, this, &WebPageProxy::hideInspectorHighlight);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::FocusedElementInformationCallback::name()) {
        IPC::handleMessage<Messages::WebPageProxy::FocusedElementInformationCallback>(decoder, this, &WebPageProxy::focusedElementInformationCallback);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::ShowInspectorIndication::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ShowInspectorIndication>(decoder, this, &WebPageProxy::showInspectorIndication);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::HideInspectorIndication::name()) {
        IPC::handleMessage<Messages::WebPageProxy::HideInspectorIndication>(decoder, this, &WebPageProxy::hideInspectorIndication);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::EnableInspectorNodeSearch::name()) {
        IPC::handleMessage<Messages::WebPageProxy::EnableInspectorNodeSearch>(decoder, this, &WebPageProxy::enableInspectorNodeSearch);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::DisableInspectorNodeSearch::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DisableInspectorNodeSearch>(decoder, this, &WebPageProxy::disableInspectorNodeSearch);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::UpdateStringForFind::name()) {
        IPC::handleMessage<Messages::WebPageProxy::UpdateStringForFind>(decoder, this, &WebPageProxy::updateStringForFind);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::HandleAutocorrectionContext::name()) {
        IPC::handleMessage<Messages::WebPageProxy::HandleAutocorrectionContext>(decoder, this, &WebPageProxy::handleAutocorrectionContext);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::DidChangeInspectorFrontendCount::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidChangeInspectorFrontendCount>(decoder, this, &WebPageProxy::didChangeInspectorFrontendCount);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::CreateInspectorTarget::name()) {
        IPC::handleMessage<Messages::WebPageProxy::CreateInspectorTarget>(decoder, this, &WebPageProxy::createInspectorTarget);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DestroyInspectorTarget::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DestroyInspectorTarget>(decoder, this, &WebPageProxy::destroyInspectorTarget);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::SendMessageToInspectorFrontend::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SendMessageToInspectorFrontend>(decoder, this, &WebPageProxy::sendMessageToInspectorFrontend);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::SaveRecentSearches::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SaveRecentSearches>(decoder, this, &WebPageProxy::saveRecentSearches);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::SavePDFToFileInDownloadsFolder::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SavePDFToFileInDownloadsFolder>(decoder, this, &WebPageProxy::savePDFToFileInDownloadsFolder);
        return;
    }
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPageProxy::SavePDFToTemporaryFolderAndOpenWithNativeApplication::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SavePDFToTemporaryFolderAndOpenWithNativeApplication>(decoder, this, &WebPageProxy::savePDFToTemporaryFolderAndOpenWithNativeApplication);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPageProxy::OpenPDFFromTemporaryFolderWithNativeApplication::name()) {
        IPC::handleMessage<Messages::WebPageProxy::OpenPDFFromTemporaryFolderWithNativeApplication>(decoder, this, &WebPageProxy::openPDFFromTemporaryFolderWithNativeApplication);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::DidUpdateActivityState::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidUpdateActivityState>(decoder, this, &WebPageProxy::didUpdateActivityState);
        return;
    }
#if (ENABLE(TELEPHONE_NUMBER_DETECTION) && PLATFORM(MAC))
    if (decoder.messageName() == Messages::WebPageProxy::ShowTelephoneNumberMenu::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ShowTelephoneNumberMenu>(decoder, this, &WebPageProxy::showTelephoneNumberMenu);
        return;
    }
#endif
#if USE(QUICK_LOOK)
    if (decoder.messageName() == Messages::WebPageProxy::DidStartLoadForQuickLookDocumentInMainFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidStartLoadForQuickLookDocumentInMainFrame>(decoder, this, &WebPageProxy::didStartLoadForQuickLookDocumentInMainFrame);
        return;
    }
#endif
#if USE(QUICK_LOOK)
    if (decoder.messageName() == Messages::WebPageProxy::DidFinishLoadForQuickLookDocumentInMainFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidFinishLoadForQuickLookDocumentInMainFrame>(decoder, this, &WebPageProxy::didFinishLoadForQuickLookDocumentInMainFrame);
        return;
    }
#endif
#if USE(QUICK_LOOK)
    if (decoder.messageName() == Messages::WebPageProxy::RequestPasswordForQuickLookDocumentInMainFrame::name()) {
        IPC::handleMessageAsync<Messages::WebPageProxy::RequestPasswordForQuickLookDocumentInMainFrame>(connection, decoder, this, &WebPageProxy::requestPasswordForQuickLookDocumentInMainFrame);
        return;
    }
#endif
#if ENABLE(CONTENT_FILTERING)
    if (decoder.messageName() == Messages::WebPageProxy::ContentFilterDidBlockLoadForFrame::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ContentFilterDidBlockLoadForFrame>(decoder, this, &WebPageProxy::contentFilterDidBlockLoadForFrame);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::IsPlayingMediaDidChange::name()) {
        IPC::handleMessage<Messages::WebPageProxy::IsPlayingMediaDidChange>(decoder, this, &WebPageProxy::isPlayingMediaDidChange);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::HandleAutoplayEvent::name()) {
        IPC::handleMessage<Messages::WebPageProxy::HandleAutoplayEvent>(decoder, this, &WebPageProxy::handleAutoplayEvent);
        return;
    }
#if ENABLE(MEDIA_SESSION)
    if (decoder.messageName() == Messages::WebPageProxy::HasMediaSessionWithActiveMediaElementsDidChange::name()) {
        IPC::handleMessage<Messages::WebPageProxy::HasMediaSessionWithActiveMediaElementsDidChange>(decoder, this, &WebPageProxy::hasMediaSessionWithActiveMediaElementsDidChange);
        return;
    }
#endif
#if ENABLE(MEDIA_SESSION)
    if (decoder.messageName() == Messages::WebPageProxy::MediaSessionMetadataDidChange::name()) {
        IPC::handleMessage<Messages::WebPageProxy::MediaSessionMetadataDidChange>(decoder, this, &WebPageProxy::mediaSessionMetadataDidChange);
        return;
    }
#endif
#if ENABLE(MEDIA_SESSION)
    if (decoder.messageName() == Messages::WebPageProxy::FocusedContentMediaElementDidChange::name()) {
        IPC::handleMessage<Messages::WebPageProxy::FocusedContentMediaElementDidChange>(decoder, this, &WebPageProxy::focusedContentMediaElementDidChange);
        return;
    }
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebPageProxy::DidPerformImmediateActionHitTest::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidPerformImmediateActionHitTest>(decoder, this, &WebPageProxy::didPerformImmediateActionHitTest);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::HandleMessage::name()) {
        IPC::handleMessage<Messages::WebPageProxy::HandleMessage>(connection, decoder, this, &WebPageProxy::handleMessage);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::HandleAutoFillButtonClick::name()) {
        IPC::handleMessage<Messages::WebPageProxy::HandleAutoFillButtonClick>(decoder, this, &WebPageProxy::handleAutoFillButtonClick);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidResignInputElementStrongPasswordAppearance::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidResignInputElementStrongPasswordAppearance>(decoder, this, &WebPageProxy::didResignInputElementStrongPasswordAppearance);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::ContentRuleListNotification::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ContentRuleListNotification>(decoder, this, &WebPageProxy::contentRuleListNotification);
        return;
    }
#if ENABLE(WIRELESS_PLAYBACK_TARGET) && !PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::AddPlaybackTargetPickerClient::name()) {
        IPC::handleMessage<Messages::WebPageProxy::AddPlaybackTargetPickerClient>(decoder, this, &WebPageProxy::addPlaybackTargetPickerClient);
        return;
    }
#endif
#if ENABLE(WIRELESS_PLAYBACK_TARGET) && !PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::RemovePlaybackTargetPickerClient::name()) {
        IPC::handleMessage<Messages::WebPageProxy::RemovePlaybackTargetPickerClient>(decoder, this, &WebPageProxy::removePlaybackTargetPickerClient);
        return;
    }
#endif
#if ENABLE(WIRELESS_PLAYBACK_TARGET) && !PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::ShowPlaybackTargetPicker::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ShowPlaybackTargetPicker>(decoder, this, &WebPageProxy::showPlaybackTargetPicker);
        return;
    }
#endif
#if ENABLE(WIRELESS_PLAYBACK_TARGET) && !PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::PlaybackTargetPickerClientStateDidChange::name()) {
        IPC::handleMessage<Messages::WebPageProxy::PlaybackTargetPickerClientStateDidChange>(decoder, this, &WebPageProxy::playbackTargetPickerClientStateDidChange);
        return;
    }
#endif
#if ENABLE(WIRELESS_PLAYBACK_TARGET) && !PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::SetMockMediaPlaybackTargetPickerEnabled::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetMockMediaPlaybackTargetPickerEnabled>(decoder, this, &WebPageProxy::setMockMediaPlaybackTargetPickerEnabled);
        return;
    }
#endif
#if ENABLE(WIRELESS_PLAYBACK_TARGET) && !PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::SetMockMediaPlaybackTargetPickerState::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetMockMediaPlaybackTargetPickerState>(decoder, this, &WebPageProxy::setMockMediaPlaybackTargetPickerState);
        return;
    }
#endif
#if ENABLE(WIRELESS_PLAYBACK_TARGET) && !PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::MockMediaPlaybackTargetPickerDismissPopup::name()) {
        IPC::handleMessage<Messages::WebPageProxy::MockMediaPlaybackTargetPickerDismissPopup>(decoder, this, &WebPageProxy::mockMediaPlaybackTargetPickerDismissPopup);
        return;
    }
#endif
#if ENABLE(POINTER_LOCK)
    if (decoder.messageName() == Messages::WebPageProxy::RequestPointerLock::name()) {
        IPC::handleMessage<Messages::WebPageProxy::RequestPointerLock>(decoder, this, &WebPageProxy::requestPointerLock);
        return;
    }
#endif
#if ENABLE(POINTER_LOCK)
    if (decoder.messageName() == Messages::WebPageProxy::RequestPointerUnlock::name()) {
        IPC::handleMessage<Messages::WebPageProxy::RequestPointerUnlock>(decoder, this, &WebPageProxy::requestPointerUnlock);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::DidFailToSuspendAfterProcessSwap::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidFailToSuspendAfterProcessSwap>(decoder, this, &WebPageProxy::didFailToSuspendAfterProcessSwap);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DidSuspendAfterProcessSwap::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidSuspendAfterProcessSwap>(decoder, this, &WebPageProxy::didSuspendAfterProcessSwap);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::ImageOrMediaDocumentSizeChanged::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ImageOrMediaDocumentSizeChanged>(decoder, this, &WebPageProxy::imageOrMediaDocumentSizeChanged);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::UseFixedLayoutDidChange::name()) {
        IPC::handleMessage<Messages::WebPageProxy::UseFixedLayoutDidChange>(decoder, this, &WebPageProxy::useFixedLayoutDidChange);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::FixedLayoutSizeDidChange::name()) {
        IPC::handleMessage<Messages::WebPageProxy::FixedLayoutSizeDidChange>(decoder, this, &WebPageProxy::fixedLayoutSizeDidChange);
        return;
    }
#if ENABLE(VIDEO) && USE(GSTREAMER)
    if (decoder.messageName() == Messages::WebPageProxy::RequestInstallMissingMediaPlugins::name()) {
        IPC::handleMessage<Messages::WebPageProxy::RequestInstallMissingMediaPlugins>(decoder, this, &WebPageProxy::requestInstallMissingMediaPlugins);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::DidRestoreScrollPosition::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidRestoreScrollPosition>(decoder, this, &WebPageProxy::didRestoreScrollPosition);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::GetLoadDecisionForIcon::name()) {
        IPC::handleMessage<Messages::WebPageProxy::GetLoadDecisionForIcon>(decoder, this, &WebPageProxy::getLoadDecisionForIcon);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::FinishedLoadingIcon::name()) {
        IPC::handleMessage<Messages::WebPageProxy::FinishedLoadingIcon>(decoder, this, &WebPageProxy::finishedLoadingIcon);
        return;
    }
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebPageProxy::DidHandleAcceptedCandidate::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidHandleAcceptedCandidate>(decoder, this, &WebPageProxy::didHandleAcceptedCandidate);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::SetIsUsingHighPerformanceWebGL::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetIsUsingHighPerformanceWebGL>(decoder, this, &WebPageProxy::setIsUsingHighPerformanceWebGL);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::StartURLSchemeTask::name()) {
        IPC::handleMessage<Messages::WebPageProxy::StartURLSchemeTask>(decoder, this, &WebPageProxy::startURLSchemeTask);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::StopURLSchemeTask::name()) {
        IPC::handleMessage<Messages::WebPageProxy::StopURLSchemeTask>(decoder, this, &WebPageProxy::stopURLSchemeTask);
        return;
    }
#if ENABLE(DEVICE_ORIENTATION)
    if (decoder.messageName() == Messages::WebPageProxy::ShouldAllowDeviceOrientationAndMotionAccess::name()) {
        IPC::handleMessageAsync<Messages::WebPageProxy::ShouldAllowDeviceOrientationAndMotionAccess>(connection, decoder, this, &WebPageProxy::shouldAllowDeviceOrientationAndMotionAccess);
        return;
    }
#endif
#if ENABLE(ATTACHMENT_ELEMENT)
    if (decoder.messageName() == Messages::WebPageProxy::RegisterAttachmentIdentifierFromData::name()) {
        IPC::handleMessage<Messages::WebPageProxy::RegisterAttachmentIdentifierFromData>(decoder, this, &WebPageProxy::registerAttachmentIdentifierFromData);
        return;
    }
#endif
#if ENABLE(ATTACHMENT_ELEMENT)
    if (decoder.messageName() == Messages::WebPageProxy::RegisterAttachmentIdentifierFromFilePath::name()) {
        IPC::handleMessage<Messages::WebPageProxy::RegisterAttachmentIdentifierFromFilePath>(decoder, this, &WebPageProxy::registerAttachmentIdentifierFromFilePath);
        return;
    }
#endif
#if ENABLE(ATTACHMENT_ELEMENT)
    if (decoder.messageName() == Messages::WebPageProxy::RegisterAttachmentIdentifier::name()) {
        IPC::handleMessage<Messages::WebPageProxy::RegisterAttachmentIdentifier>(decoder, this, &WebPageProxy::registerAttachmentIdentifier);
        return;
    }
#endif
#if ENABLE(ATTACHMENT_ELEMENT)
    if (decoder.messageName() == Messages::WebPageProxy::registerAttachmentsFromSerializedData::name()) {
        IPC::handleMessage<Messages::WebPageProxy::registerAttachmentsFromSerializedData>(decoder, this, &WebPageProxy::registerAttachmentsFromSerializedData);
        return;
    }
#endif
#if ENABLE(ATTACHMENT_ELEMENT)
    if (decoder.messageName() == Messages::WebPageProxy::CloneAttachmentData::name()) {
        IPC::handleMessage<Messages::WebPageProxy::CloneAttachmentData>(decoder, this, &WebPageProxy::cloneAttachmentData);
        return;
    }
#endif
#if ENABLE(ATTACHMENT_ELEMENT)
    if (decoder.messageName() == Messages::WebPageProxy::DidInsertAttachmentWithIdentifier::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidInsertAttachmentWithIdentifier>(decoder, this, &WebPageProxy::didInsertAttachmentWithIdentifier);
        return;
    }
#endif
#if ENABLE(ATTACHMENT_ELEMENT)
    if (decoder.messageName() == Messages::WebPageProxy::DidRemoveAttachmentWithIdentifier::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidRemoveAttachmentWithIdentifier>(decoder, this, &WebPageProxy::didRemoveAttachmentWithIdentifier);
        return;
    }
#endif
#if ENABLE(SPEECH_SYNTHESIS)
    if (decoder.messageName() == Messages::WebPageProxy::SpeechSynthesisSpeak::name()) {
        IPC::handleMessageAsync<Messages::WebPageProxy::SpeechSynthesisSpeak>(connection, decoder, this, &WebPageProxy::speechSynthesisSpeak);
        return;
    }
#endif
#if ENABLE(SPEECH_SYNTHESIS)
    if (decoder.messageName() == Messages::WebPageProxy::SpeechSynthesisSetFinishedCallback::name()) {
        IPC::handleMessageAsync<Messages::WebPageProxy::SpeechSynthesisSetFinishedCallback>(connection, decoder, this, &WebPageProxy::speechSynthesisSetFinishedCallback);
        return;
    }
#endif
#if ENABLE(SPEECH_SYNTHESIS)
    if (decoder.messageName() == Messages::WebPageProxy::SpeechSynthesisCancel::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SpeechSynthesisCancel>(decoder, this, &WebPageProxy::speechSynthesisCancel);
        return;
    }
#endif
#if ENABLE(SPEECH_SYNTHESIS)
    if (decoder.messageName() == Messages::WebPageProxy::SpeechSynthesisPause::name()) {
        IPC::handleMessageAsync<Messages::WebPageProxy::SpeechSynthesisPause>(connection, decoder, this, &WebPageProxy::speechSynthesisPause);
        return;
    }
#endif
#if ENABLE(SPEECH_SYNTHESIS)
    if (decoder.messageName() == Messages::WebPageProxy::SpeechSynthesisResume::name()) {
        IPC::handleMessageAsync<Messages::WebPageProxy::SpeechSynthesisResume>(connection, decoder, this, &WebPageProxy::speechSynthesisResume);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::ConfigureLoggingChannel::name()) {
        IPC::handleMessage<Messages::WebPageProxy::ConfigureLoggingChannel>(decoder, this, &WebPageProxy::configureLoggingChannel);
        return;
    }
#if PLATFORM(GTK)
    if (decoder.messageName() == Messages::WebPageProxy::ShowEmojiPicker::name()) {
        IPC::handleMessageAsync<Messages::WebPageProxy::ShowEmojiPicker>(connection, decoder, this, &WebPageProxy::showEmojiPicker);
        return;
    }
#endif
#if HAVE(VISIBILITY_PROPAGATION_VIEW)
    if (decoder.messageName() == Messages::WebPageProxy::DidCreateContextForVisibilityPropagation::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidCreateContextForVisibilityPropagation>(decoder, this, &WebPageProxy::didCreateContextForVisibilityPropagation);
        return;
    }
#endif
#if ENABLE(WEB_AUTHN)
    if (decoder.messageName() == Messages::WebPageProxy::SetMockWebAuthenticationConfiguration::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SetMockWebAuthenticationConfiguration>(decoder, this, &WebPageProxy::setMockWebAuthenticationConfiguration);
        return;
    }
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    if (decoder.messageName() == Messages::WebPageProxy::SendMessageToWebView::name()) {
        IPC::handleMessage<Messages::WebPageProxy::SendMessageToWebView>(decoder, this, &WebPageProxy::sendMessageToWebView);
        return;
    }
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    if (decoder.messageName() == Messages::WebPageProxy::SendMessageToWebViewWithReply::name()) {
        IPC::handleMessageAsync<Messages::WebPageProxy::SendMessageToWebViewWithReply>(connection, decoder, this, &WebPageProxy::sendMessageToWebViewWithReply);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::DidFindTextManipulationItem::name()) {
        IPC::handleMessage<Messages::WebPageProxy::DidFindTextManipulationItem>(decoder, this, &WebPageProxy::didFindTextManipulationItem);
        return;
    }
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED();
}

void WebPageProxy::didReceiveSyncMessage(IPC::Connection& connection, IPC::Decoder& decoder, std::unique_ptr<IPC::Encoder>& replyEncoder)
{
    auto protectedThis = makeRef(*this);
    if (decoder.messageName() == Messages::WebPageProxy::CreateNewPage::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::CreateNewPage>(connection, decoder, replyEncoder, this, &WebPageProxy::createNewPage);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::RunJavaScriptAlert::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::RunJavaScriptAlert>(connection, decoder, replyEncoder, this, &WebPageProxy::runJavaScriptAlert);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::RunJavaScriptConfirm::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::RunJavaScriptConfirm>(connection, decoder, replyEncoder, this, &WebPageProxy::runJavaScriptConfirm);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::RunJavaScriptPrompt::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::RunJavaScriptPrompt>(connection, decoder, replyEncoder, this, &WebPageProxy::runJavaScriptPrompt);
        return;
    }
#if ENABLE(WEBGL)
    if (decoder.messageName() == Messages::WebPageProxy::WebGLPolicyForURL::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::WebGLPolicyForURL>(connection, decoder, replyEncoder, this, &WebPageProxy::webGLPolicyForURL);
        return;
    }
#endif
#if ENABLE(WEBGL)
    if (decoder.messageName() == Messages::WebPageProxy::ResolveWebGLPolicyForURL::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::ResolveWebGLPolicyForURL>(connection, decoder, replyEncoder, this, &WebPageProxy::resolveWebGLPolicyForURL);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::GetToolbarsAreVisible::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::GetToolbarsAreVisible>(connection, decoder, replyEncoder, this, &WebPageProxy::getToolbarsAreVisible);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::GetMenuBarIsVisible::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::GetMenuBarIsVisible>(connection, decoder, replyEncoder, this, &WebPageProxy::getMenuBarIsVisible);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::GetStatusBarIsVisible::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::GetStatusBarIsVisible>(connection, decoder, replyEncoder, this, &WebPageProxy::getStatusBarIsVisible);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::GetWindowFrame::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::GetWindowFrame>(connection, decoder, replyEncoder, this, &WebPageProxy::getWindowFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::ScreenToRootView::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::ScreenToRootView>(connection, decoder, replyEncoder, this, &WebPageProxy::screenToRootView);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::RootViewToScreen::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::RootViewToScreen>(connection, decoder, replyEncoder, this, &WebPageProxy::rootViewToScreen);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::AccessibilityScreenToRootView::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::AccessibilityScreenToRootView>(connection, decoder, replyEncoder, this, &WebPageProxy::accessibilityScreenToRootView);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::RootViewToAccessibilityScreen::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::RootViewToAccessibilityScreen>(connection, decoder, replyEncoder, this, &WebPageProxy::rootViewToAccessibilityScreen);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::RunBeforeUnloadConfirmPanel::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::RunBeforeUnloadConfirmPanel>(connection, decoder, replyEncoder, this, &WebPageProxy::runBeforeUnloadConfirmPanel);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::PrintFrame::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::PrintFrame>(connection, decoder, replyEncoder, this, &WebPageProxy::printFrame);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::DecidePolicyForNavigationActionSync::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::DecidePolicyForNavigationActionSync>(connection, decoder, replyEncoder, this, &WebPageProxy::decidePolicyForNavigationActionSync);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::HasInsecureContent::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::HasInsecureContent>(connection, decoder, replyEncoder, this, &WebPageProxy::hasInsecureContent);
        return;
    }
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPageProxy::InterpretKeyEvent::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::InterpretKeyEvent>(connection, decoder, replyEncoder, this, &WebPageProxy::interpretKeyEvent);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::BackForwardGoToItem::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::BackForwardGoToItem>(connection, decoder, replyEncoder, this, &WebPageProxy::backForwardGoToItem);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::BackForwardItemAtIndex::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::BackForwardItemAtIndex>(connection, decoder, replyEncoder, this, &WebPageProxy::backForwardItemAtIndex);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::BackForwardListCounts::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::BackForwardListCounts>(connection, decoder, replyEncoder, this, &WebPageProxy::backForwardListCounts);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::CanUndoRedo::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::CanUndoRedo>(connection, decoder, replyEncoder, this, &WebPageProxy::canUndoRedo);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::ExecuteUndoRedo::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::ExecuteUndoRedo>(connection, decoder, replyEncoder, this, &WebPageProxy::executeUndoRedo);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::RequestDOMPasteAccess::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::RequestDOMPasteAccess>(connection, decoder, replyEncoder, this, &WebPageProxy::requestDOMPasteAccess);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::ExceededDatabaseQuota::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::ExceededDatabaseQuota>(connection, decoder, replyEncoder, this, &WebPageProxy::exceededDatabaseQuota);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::ReachedApplicationCacheOriginQuota::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::ReachedApplicationCacheOriginQuota>(connection, decoder, replyEncoder, this, &WebPageProxy::reachedApplicationCacheOriginQuota);
        return;
    }
#if USE(UNIFIED_TEXT_CHECKING)
    if (decoder.messageName() == Messages::WebPageProxy::CheckTextOfParagraph::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::CheckTextOfParagraph>(connection, decoder, replyEncoder, this, &WebPageProxy::checkTextOfParagraph);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::CheckSpellingOfString::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::CheckSpellingOfString>(connection, decoder, replyEncoder, this, &WebPageProxy::checkSpellingOfString);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::CheckGrammarOfString::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::CheckGrammarOfString>(connection, decoder, replyEncoder, this, &WebPageProxy::checkGrammarOfString);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::SpellingUIIsShowing::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::SpellingUIIsShowing>(connection, decoder, replyEncoder, this, &WebPageProxy::spellingUIIsShowing);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::GetGuessesForWord::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::GetGuessesForWord>(connection, decoder, replyEncoder, this, &WebPageProxy::getGuessesForWord);
        return;
    }
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPageProxy::ExecuteSavedCommandBySelector::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::ExecuteSavedCommandBySelector>(connection, decoder, replyEncoder, this, &WebPageProxy::executeSavedCommandBySelector);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPageProxy::GetIsSpeaking::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::GetIsSpeaking>(connection, decoder, replyEncoder, this, &WebPageProxy::getIsSpeaking);
        return;
    }
#endif
#if USE(APPKIT)
    if (decoder.messageName() == Messages::WebPageProxy::SubstitutionsPanelIsShowing::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::SubstitutionsPanelIsShowing>(connection, decoder, replyEncoder, this, &WebPageProxy::substitutionsPanelIsShowing);
        return;
    }
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebPageProxy::DismissCorrectionPanelSoon::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::DismissCorrectionPanelSoon>(connection, decoder, replyEncoder, this, &WebPageProxy::dismissCorrectionPanelSoon);
        return;
    }
#endif
#if USE(DICTATION_ALTERNATIVES)
    if (decoder.messageName() == Messages::WebPageProxy::DictationAlternatives::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::DictationAlternatives>(connection, decoder, replyEncoder, this, &WebPageProxy::dictationAlternatives);
        return;
    }
#endif
#if PLATFORM(X11)
    if (decoder.messageName() == Messages::WebPageProxy::CreatePluginContainer::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::CreatePluginContainer>(connection, decoder, replyEncoder, this, &WebPageProxy::createPluginContainer);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::LoadRecentSearches::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::LoadRecentSearches>(connection, decoder, replyEncoder, this, &WebPageProxy::loadRecentSearches);
        return;
    }
#if ENABLE(PDFKIT_PLUGIN)
    if (decoder.messageName() == Messages::WebPageProxy::ShowPDFContextMenu::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::ShowPDFContextMenu>(connection, decoder, replyEncoder, this, &WebPageProxy::showPDFContextMenu);
        return;
    }
#endif
#if ENABLE(NETSCAPE_PLUGIN_API)
    if (decoder.messageName() == Messages::WebPageProxy::FindPlugin::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::FindPlugin>(connection, decoder, replyEncoder, this, &WebPageProxy::findPlugin);
        return;
    }
#endif
#if ENABLE(WEB_CRYPTO)
    if (decoder.messageName() == Messages::WebPageProxy::WrapCryptoKey::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::WrapCryptoKey>(connection, decoder, replyEncoder, this, &WebPageProxy::wrapCryptoKey);
        return;
    }
#endif
#if ENABLE(WEB_CRYPTO)
    if (decoder.messageName() == Messages::WebPageProxy::UnwrapCryptoKey::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::UnwrapCryptoKey>(connection, decoder, replyEncoder, this, &WebPageProxy::unwrapCryptoKey);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::HandleSynchronousMessage::name()) {
        IPC::handleMessageSynchronousWantsConnection<Messages::WebPageProxy::HandleSynchronousMessage>(connection, decoder, replyEncoder, this, &WebPageProxy::handleSynchronousMessage);
        return;
    }
    if (decoder.messageName() == Messages::WebPageProxy::LoadSynchronousURLSchemeTask::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::LoadSynchronousURLSchemeTask>(connection, decoder, replyEncoder, this, &WebPageProxy::loadSynchronousURLSchemeTask);
        return;
    }
#if ENABLE(ATTACHMENT_ELEMENT)
    if (decoder.messageName() == Messages::WebPageProxy::SerializedAttachmentDataForIdentifiers::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::SerializedAttachmentDataForIdentifiers>(connection, decoder, replyEncoder, this, &WebPageProxy::serializedAttachmentDataForIdentifiers);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebPageProxy::SignedPublicKeyAndChallengeString::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::SignedPublicKeyAndChallengeString>(connection, decoder, replyEncoder, this, &WebPageProxy::signedPublicKeyAndChallengeString);
        return;
    }
#if ENABLE(SPEECH_SYNTHESIS)
    if (decoder.messageName() == Messages::WebPageProxy::SpeechSynthesisVoiceList::name()) {
        IPC::handleMessageSynchronous<Messages::WebPageProxy::SpeechSynthesisVoiceList>(connection, decoder, replyEncoder, this, &WebPageProxy::speechSynthesisVoiceList);
        return;
    }
#endif
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    UNUSED_PARAM(replyEncoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit

