/*
 * Copyright (C) 2010-2019 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#include "WebPasteboardProxy.h"

#include "ArgumentCoders.h"
#include "Decoder.h"
#include "HandleMessage.h"
#if PLATFORM(COCOA)
#include "SandboxExtension.h"
#endif
#include "SharedMemory.h"
#include "WebCoreArgumentCoders.h"
#include "WebPasteboardProxyMessages.h"
#if PLATFORM(GTK)
#include "WebSelectionData.h"
#endif
#if PLATFORM(COCOA)
#include <WebCore/Color.h>
#endif
#include <WebCore/Pasteboard.h>
#include <WebCore/PasteboardItemInfo.h>
#include <wtf/Optional.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace Messages {

namespace WebPasteboardProxy {

void WriteCustomData::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, int64_t changeCount)
{
    *encoder << changeCount;
    connection.sendSyncReply(WTFMove(encoder));
}

void TypesSafeForDOMToReadAndWrite::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<String>& types)
{
    *encoder << types;
    connection.sendSyncReply(WTFMove(encoder));
}

void AllPasteboardItemInfo::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Optional<Vector<WebCore::PasteboardItemInfo>>& allInfo)
{
    *encoder << allInfo;
    connection.sendSyncReply(WTFMove(encoder));
}

void InformationForItemAtIndex::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Optional<WebCore::PasteboardItemInfo>& info)
{
    *encoder << info;
    connection.sendSyncReply(WTFMove(encoder));
}

void GetPasteboardItemsCount::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, uint64_t itemsCount)
{
    *encoder << itemsCount;
    connection.sendSyncReply(WTFMove(encoder));
}

void ReadStringFromPasteboard::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const String& string)
{
    *encoder << string;
    connection.sendSyncReply(WTFMove(encoder));
}

void ReadURLFromPasteboard::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const String& url, const String& title)
{
    *encoder << url;
    *encoder << title;
    connection.sendSyncReply(WTFMove(encoder));
}

void ReadBufferFromPasteboard::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::SharedMemory::Handle& handle, uint64_t size)
{
    *encoder << handle;
    *encoder << size;
    connection.sendSyncReply(WTFMove(encoder));
}

#if PLATFORM(COCOA)

void GetNumberOfFiles::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, uint64_t numberOfFiles)
{
    *encoder << numberOfFiles;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void GetPasteboardTypes::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<String>& types)
{
    *encoder << types;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void GetPasteboardPathnamesForType::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<String>& pathnames, const WebKit::SandboxExtension::HandleArray& sandboxExtensions)
{
    *encoder << pathnames;
    *encoder << sandboxExtensions;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void GetPasteboardStringForType::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const String& string)
{
    *encoder << string;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void GetPasteboardStringsForType::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<String>& strings)
{
    *encoder << strings;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void GetPasteboardBufferForType::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::SharedMemory::Handle& handle, uint64_t size)
{
    *encoder << handle;
    *encoder << size;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void PasteboardCopy::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, int64_t changeCount)
{
    *encoder << changeCount;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void GetPasteboardChangeCount::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, int64_t changeCount)
{
    *encoder << changeCount;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void GetPasteboardUniqueName::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const String& pasteboardName)
{
    *encoder << pasteboardName;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void GetPasteboardColor::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebCore::Color& color)
{
    *encoder << color;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void GetPasteboardURL::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const String& urlString)
{
    *encoder << urlString;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void AddPasteboardTypes::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, int64_t changeCount)
{
    *encoder << changeCount;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void SetPasteboardTypes::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, int64_t changeCount)
{
    *encoder << changeCount;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void SetPasteboardURL::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, int64_t changeCount)
{
    *encoder << changeCount;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void SetPasteboardColor::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, int64_t changeCount)
{
    *encoder << changeCount;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void SetPasteboardStringForType::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, int64_t changeCount)
{
    *encoder << changeCount;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(COCOA)

void SetPasteboardBufferForType::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, int64_t changeCount)
{
    *encoder << changeCount;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(GTK)

void ReadFromClipboard::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::WebSelectionData& pasteboardContent)
{
    *encoder << pasteboardContent;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if USE(LIBWPE)

void GetPasteboardTypes::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<String>& types)
{
    *encoder << types;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

} // namespace WebPasteboardProxy

} // namespace Messages

namespace WebKit {

void WebPasteboardProxy::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPasteboardProxy::WriteURLToPasteboard::name()) {
        IPC::handleMessage<Messages::WebPasteboardProxy::WriteURLToPasteboard>(decoder, this, &WebPasteboardProxy::writeURLToPasteboard);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPasteboardProxy::WriteWebContentToPasteboard::name()) {
        IPC::handleMessage<Messages::WebPasteboardProxy::WriteWebContentToPasteboard>(decoder, this, &WebPasteboardProxy::writeWebContentToPasteboard);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPasteboardProxy::WriteImageToPasteboard::name()) {
        IPC::handleMessage<Messages::WebPasteboardProxy::WriteImageToPasteboard>(decoder, this, &WebPasteboardProxy::writeImageToPasteboard);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPasteboardProxy::WriteStringToPasteboard::name()) {
        IPC::handleMessage<Messages::WebPasteboardProxy::WriteStringToPasteboard>(decoder, this, &WebPasteboardProxy::writeStringToPasteboard);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebPasteboardProxy::UpdateSupportedTypeIdentifiers::name()) {
        IPC::handleMessage<Messages::WebPasteboardProxy::UpdateSupportedTypeIdentifiers>(decoder, this, &WebPasteboardProxy::updateSupportedTypeIdentifiers);
        return;
    }
#endif
#if PLATFORM(GTK)
    if (decoder.messageName() == Messages::WebPasteboardProxy::WriteToClipboard::name()) {
        IPC::handleMessage<Messages::WebPasteboardProxy::WriteToClipboard>(decoder, this, &WebPasteboardProxy::writeToClipboard);
        return;
    }
#endif
#if USE(LIBWPE)
    if (decoder.messageName() == Messages::WebPasteboardProxy::WriteWebContentToPasteboard::name()) {
        IPC::handleMessage<Messages::WebPasteboardProxy::WriteWebContentToPasteboard>(decoder, this, &WebPasteboardProxy::writeWebContentToPasteboard);
        return;
    }
#endif
#if USE(LIBWPE)
    if (decoder.messageName() == Messages::WebPasteboardProxy::WriteStringToPasteboard::name()) {
        IPC::handleMessage<Messages::WebPasteboardProxy::WriteStringToPasteboard>(decoder, this, &WebPasteboardProxy::writeStringToPasteboard);
        return;
    }
#endif
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED();
}

void WebPasteboardProxy::didReceiveSyncMessage(IPC::Connection& connection, IPC::Decoder& decoder, std::unique_ptr<IPC::Encoder>& replyEncoder)
{
    if (decoder.messageName() == Messages::WebPasteboardProxy::WriteCustomData::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::WriteCustomData>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::writeCustomData);
        return;
    }
    if (decoder.messageName() == Messages::WebPasteboardProxy::TypesSafeForDOMToReadAndWrite::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::TypesSafeForDOMToReadAndWrite>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::typesSafeForDOMToReadAndWrite);
        return;
    }
    if (decoder.messageName() == Messages::WebPasteboardProxy::AllPasteboardItemInfo::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::AllPasteboardItemInfo>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::allPasteboardItemInfo);
        return;
    }
    if (decoder.messageName() == Messages::WebPasteboardProxy::InformationForItemAtIndex::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::InformationForItemAtIndex>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::informationForItemAtIndex);
        return;
    }
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetPasteboardItemsCount::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::GetPasteboardItemsCount>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::getPasteboardItemsCount);
        return;
    }
    if (decoder.messageName() == Messages::WebPasteboardProxy::ReadStringFromPasteboard::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::ReadStringFromPasteboard>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::readStringFromPasteboard);
        return;
    }
    if (decoder.messageName() == Messages::WebPasteboardProxy::ReadURLFromPasteboard::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::ReadURLFromPasteboard>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::readURLFromPasteboard);
        return;
    }
    if (decoder.messageName() == Messages::WebPasteboardProxy::ReadBufferFromPasteboard::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::ReadBufferFromPasteboard>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::readBufferFromPasteboard);
        return;
    }
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetNumberOfFiles::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::GetNumberOfFiles>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::getNumberOfFiles);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetPasteboardTypes::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::GetPasteboardTypes>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::getPasteboardTypes);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetPasteboardPathnamesForType::name()) {
        IPC::handleMessageSynchronousWantsConnection<Messages::WebPasteboardProxy::GetPasteboardPathnamesForType>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::getPasteboardPathnamesForType);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetPasteboardStringForType::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::GetPasteboardStringForType>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::getPasteboardStringForType);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetPasteboardStringsForType::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::GetPasteboardStringsForType>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::getPasteboardStringsForType);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetPasteboardBufferForType::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::GetPasteboardBufferForType>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::getPasteboardBufferForType);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::PasteboardCopy::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::PasteboardCopy>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::pasteboardCopy);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetPasteboardChangeCount::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::GetPasteboardChangeCount>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::getPasteboardChangeCount);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetPasteboardUniqueName::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::GetPasteboardUniqueName>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::getPasteboardUniqueName);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetPasteboardColor::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::GetPasteboardColor>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::getPasteboardColor);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetPasteboardURL::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::GetPasteboardURL>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::getPasteboardURL);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::AddPasteboardTypes::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::AddPasteboardTypes>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::addPasteboardTypes);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::SetPasteboardTypes::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::SetPasteboardTypes>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::setPasteboardTypes);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::SetPasteboardURL::name()) {
        IPC::handleMessageSynchronousWantsConnection<Messages::WebPasteboardProxy::SetPasteboardURL>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::setPasteboardURL);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::SetPasteboardColor::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::SetPasteboardColor>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::setPasteboardColor);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::SetPasteboardStringForType::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::SetPasteboardStringForType>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::setPasteboardStringForType);
        return;
    }
#endif
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebPasteboardProxy::SetPasteboardBufferForType::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::SetPasteboardBufferForType>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::setPasteboardBufferForType);
        return;
    }
#endif
#if PLATFORM(GTK)
    if (decoder.messageName() == Messages::WebPasteboardProxy::ReadFromClipboard::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::ReadFromClipboard>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::readFromClipboard);
        return;
    }
#endif
#if USE(LIBWPE)
    if (decoder.messageName() == Messages::WebPasteboardProxy::GetPasteboardTypes::name()) {
        IPC::handleMessageSynchronous<Messages::WebPasteboardProxy::GetPasteboardTypes>(connection, decoder, replyEncoder, this, &WebPasteboardProxy::getPasteboardTypes);
        return;
    }
#endif
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    UNUSED_PARAM(replyEncoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit

