/*
 * Copyright (C) 2010-2019 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ArgumentCoders.h"
#include "Connection.h"
#include "SandboxExtension.h"
#include "SharedMemory.h"
#include "WebPasteboardProxyMessagesReplies.h"
#include <WebCore/Pasteboard.h>
#include <WebCore/PasteboardItemInfo.h>
#include <wtf/Forward.h>
#include <wtf/Optional.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace WebCore {
class Color;
struct PasteboardImage;
struct PasteboardURL;
struct PasteboardWebContent;
}

namespace WebKit {
struct WebSelectionData;
}

namespace Messages {
namespace WebPasteboardProxy {

static inline IPC::StringReference messageReceiverName()
{
    return IPC::StringReference("WebPasteboardProxy");
}

#if PLATFORM(IOS_FAMILY)
class WriteURLToPasteboard {
public:
    typedef std::tuple<const WebCore::PasteboardURL&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("WriteURLToPasteboard"); }
    static const bool isSync = false;

    WriteURLToPasteboard(const WebCore::PasteboardURL& url, const String& pasteboardName)
        : m_arguments(url, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class WriteWebContentToPasteboard {
public:
    typedef std::tuple<const WebCore::PasteboardWebContent&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("WriteWebContentToPasteboard"); }
    static const bool isSync = false;

    WriteWebContentToPasteboard(const WebCore::PasteboardWebContent& content, const String& pasteboardName)
        : m_arguments(content, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class WriteImageToPasteboard {
public:
    typedef std::tuple<const WebCore::PasteboardImage&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("WriteImageToPasteboard"); }
    static const bool isSync = false;

    WriteImageToPasteboard(const WebCore::PasteboardImage& pasteboardImage, const String& pasteboardName)
        : m_arguments(pasteboardImage, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class WriteStringToPasteboard {
public:
    typedef std::tuple<const String&, const String&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("WriteStringToPasteboard"); }
    static const bool isSync = false;

    WriteStringToPasteboard(const String& pasteboardType, const String& text, const String& pasteboardName)
        : m_arguments(pasteboardType, text, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class UpdateSupportedTypeIdentifiers {
public:
    typedef std::tuple<const Vector<String>&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("UpdateSupportedTypeIdentifiers"); }
    static const bool isSync = false;

    UpdateSupportedTypeIdentifiers(const Vector<String>& identifiers, const String& pasteboardName)
        : m_arguments(identifiers, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class WriteCustomData {
public:
    typedef std::tuple<const Vector<WebCore::PasteboardCustomData>&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("WriteCustomData"); }
    static const bool isSync = true;

    using DelayedReply = WriteCustomDataDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, int64_t changeCount);
    using Reply = std::tuple<int64_t&>;
    using ReplyArguments = std::tuple<int64_t>;
    WriteCustomData(const Vector<WebCore::PasteboardCustomData>& data, const String& pasteboardName)
        : m_arguments(data, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class TypesSafeForDOMToReadAndWrite {
public:
    typedef std::tuple<const String&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("TypesSafeForDOMToReadAndWrite"); }
    static const bool isSync = true;

    using DelayedReply = TypesSafeForDOMToReadAndWriteDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<String>& types);
    using Reply = std::tuple<Vector<String>&>;
    using ReplyArguments = std::tuple<Vector<String>>;
    TypesSafeForDOMToReadAndWrite(const String& pasteboardName, const String& origin)
        : m_arguments(pasteboardName, origin)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class AllPasteboardItemInfo {
public:
    typedef std::tuple<const String&, int64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("AllPasteboardItemInfo"); }
    static const bool isSync = true;

    using DelayedReply = AllPasteboardItemInfoDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Optional<Vector<WebCore::PasteboardItemInfo>>& allInfo);
    using Reply = std::tuple<Optional<Vector<WebCore::PasteboardItemInfo>>&>;
    using ReplyArguments = std::tuple<Optional<Vector<WebCore::PasteboardItemInfo>>>;
    AllPasteboardItemInfo(const String& pasteboardName, int64_t changeCount)
        : m_arguments(pasteboardName, changeCount)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class InformationForItemAtIndex {
public:
    typedef std::tuple<uint64_t, const String&, int64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("InformationForItemAtIndex"); }
    static const bool isSync = true;

    using DelayedReply = InformationForItemAtIndexDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Optional<WebCore::PasteboardItemInfo>& info);
    using Reply = std::tuple<Optional<WebCore::PasteboardItemInfo>&>;
    using ReplyArguments = std::tuple<Optional<WebCore::PasteboardItemInfo>>;
    InformationForItemAtIndex(uint64_t index, const String& pasteboardName, int64_t changeCount)
        : m_arguments(index, pasteboardName, changeCount)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetPasteboardItemsCount {
public:
    typedef std::tuple<const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetPasteboardItemsCount"); }
    static const bool isSync = true;

    using DelayedReply = GetPasteboardItemsCountDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, uint64_t itemsCount);
    using Reply = std::tuple<uint64_t&>;
    using ReplyArguments = std::tuple<uint64_t>;
    explicit GetPasteboardItemsCount(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ReadStringFromPasteboard {
public:
    typedef std::tuple<uint64_t, const String&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("ReadStringFromPasteboard"); }
    static const bool isSync = true;

    using DelayedReply = ReadStringFromPasteboardDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& string);
    using Reply = std::tuple<String&>;
    using ReplyArguments = std::tuple<String>;
    ReadStringFromPasteboard(uint64_t index, const String& pasteboardType, const String& pasteboardName)
        : m_arguments(index, pasteboardType, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ReadURLFromPasteboard {
public:
    typedef std::tuple<uint64_t, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("ReadURLFromPasteboard"); }
    static const bool isSync = true;

    using DelayedReply = ReadURLFromPasteboardDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& url, const String& title);
    using Reply = std::tuple<String&, String&>;
    using ReplyArguments = std::tuple<String, String>;
    ReadURLFromPasteboard(uint64_t index, const String& pasteboardName)
        : m_arguments(index, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ReadBufferFromPasteboard {
public:
    typedef std::tuple<uint64_t, const String&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("ReadBufferFromPasteboard"); }
    static const bool isSync = true;

    using DelayedReply = ReadBufferFromPasteboardDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebKit::SharedMemory::Handle& handle, uint64_t size);
    using Reply = std::tuple<WebKit::SharedMemory::Handle&, uint64_t&>;
    using ReplyArguments = std::tuple<WebKit::SharedMemory::Handle, uint64_t>;
    ReadBufferFromPasteboard(uint64_t index, const String& pasteboardType, const String& pasteboardName)
        : m_arguments(index, pasteboardType, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if PLATFORM(COCOA)
class GetNumberOfFiles {
public:
    typedef std::tuple<const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetNumberOfFiles"); }
    static const bool isSync = true;

    using DelayedReply = GetNumberOfFilesDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, uint64_t numberOfFiles);
    using Reply = std::tuple<uint64_t&>;
    using ReplyArguments = std::tuple<uint64_t>;
    explicit GetNumberOfFiles(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardTypes {
public:
    typedef std::tuple<const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetPasteboardTypes"); }
    static const bool isSync = true;

    using DelayedReply = GetPasteboardTypesDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<String>& types);
    using Reply = std::tuple<Vector<String>&>;
    using ReplyArguments = std::tuple<Vector<String>>;
    explicit GetPasteboardTypes(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardPathnamesForType {
public:
    typedef std::tuple<const String&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetPasteboardPathnamesForType"); }
    static const bool isSync = true;

    using DelayedReply = GetPasteboardPathnamesForTypeDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<String>& pathnames, const WebKit::SandboxExtension::HandleArray& sandboxExtensions);
    using Reply = std::tuple<Vector<String>&, WebKit::SandboxExtension::HandleArray&>;
    using ReplyArguments = std::tuple<Vector<String>, WebKit::SandboxExtension::HandleArray>;
    GetPasteboardPathnamesForType(const String& pasteboardName, const String& pasteboardType)
        : m_arguments(pasteboardName, pasteboardType)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardStringForType {
public:
    typedef std::tuple<const String&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetPasteboardStringForType"); }
    static const bool isSync = true;

    using DelayedReply = GetPasteboardStringForTypeDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& string);
    using Reply = std::tuple<String&>;
    using ReplyArguments = std::tuple<String>;
    GetPasteboardStringForType(const String& pasteboardName, const String& pasteboardType)
        : m_arguments(pasteboardName, pasteboardType)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardStringsForType {
public:
    typedef std::tuple<const String&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetPasteboardStringsForType"); }
    static const bool isSync = true;

    using DelayedReply = GetPasteboardStringsForTypeDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<String>& strings);
    using Reply = std::tuple<Vector<String>&>;
    using ReplyArguments = std::tuple<Vector<String>>;
    GetPasteboardStringsForType(const String& pasteboardName, const String& pasteboardType)
        : m_arguments(pasteboardName, pasteboardType)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardBufferForType {
public:
    typedef std::tuple<const String&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetPasteboardBufferForType"); }
    static const bool isSync = true;

    using DelayedReply = GetPasteboardBufferForTypeDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebKit::SharedMemory::Handle& handle, uint64_t size);
    using Reply = std::tuple<WebKit::SharedMemory::Handle&, uint64_t&>;
    using ReplyArguments = std::tuple<WebKit::SharedMemory::Handle, uint64_t>;
    GetPasteboardBufferForType(const String& pasteboardName, const String& pasteboardType)
        : m_arguments(pasteboardName, pasteboardType)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class PasteboardCopy {
public:
    typedef std::tuple<const String&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("PasteboardCopy"); }
    static const bool isSync = true;

    using DelayedReply = PasteboardCopyDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, int64_t changeCount);
    using Reply = std::tuple<int64_t&>;
    using ReplyArguments = std::tuple<int64_t>;
    PasteboardCopy(const String& fromPasteboard, const String& toPasteboard)
        : m_arguments(fromPasteboard, toPasteboard)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardChangeCount {
public:
    typedef std::tuple<const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetPasteboardChangeCount"); }
    static const bool isSync = true;

    using DelayedReply = GetPasteboardChangeCountDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, int64_t changeCount);
    using Reply = std::tuple<int64_t&>;
    using ReplyArguments = std::tuple<int64_t>;
    explicit GetPasteboardChangeCount(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardUniqueName {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetPasteboardUniqueName"); }
    static const bool isSync = true;

    using DelayedReply = GetPasteboardUniqueNameDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& pasteboardName);
    using Reply = std::tuple<String&>;
    using ReplyArguments = std::tuple<String>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardColor {
public:
    typedef std::tuple<const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetPasteboardColor"); }
    static const bool isSync = true;

    using DelayedReply = GetPasteboardColorDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebCore::Color& color);
    using Reply = std::tuple<WebCore::Color&>;
    using ReplyArguments = std::tuple<WebCore::Color>;
    explicit GetPasteboardColor(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardURL {
public:
    typedef std::tuple<const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetPasteboardURL"); }
    static const bool isSync = true;

    using DelayedReply = GetPasteboardURLDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& urlString);
    using Reply = std::tuple<String&>;
    using ReplyArguments = std::tuple<String>;
    explicit GetPasteboardURL(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class AddPasteboardTypes {
public:
    typedef std::tuple<const String&, const Vector<String>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("AddPasteboardTypes"); }
    static const bool isSync = true;

    using DelayedReply = AddPasteboardTypesDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, int64_t changeCount);
    using Reply = std::tuple<int64_t&>;
    using ReplyArguments = std::tuple<int64_t>;
    AddPasteboardTypes(const String& pasteboardName, const Vector<String>& pasteboardTypes)
        : m_arguments(pasteboardName, pasteboardTypes)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SetPasteboardTypes {
public:
    typedef std::tuple<const String&, const Vector<String>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SetPasteboardTypes"); }
    static const bool isSync = true;

    using DelayedReply = SetPasteboardTypesDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, int64_t changeCount);
    using Reply = std::tuple<int64_t&>;
    using ReplyArguments = std::tuple<int64_t>;
    SetPasteboardTypes(const String& pasteboardName, const Vector<String>& pasteboardTypes)
        : m_arguments(pasteboardName, pasteboardTypes)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SetPasteboardURL {
public:
    typedef std::tuple<const WebCore::PasteboardURL&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SetPasteboardURL"); }
    static const bool isSync = true;

    using DelayedReply = SetPasteboardURLDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, int64_t changeCount);
    using Reply = std::tuple<int64_t&>;
    using ReplyArguments = std::tuple<int64_t>;
    SetPasteboardURL(const WebCore::PasteboardURL& pasteboardURL, const String& pasteboardName)
        : m_arguments(pasteboardURL, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SetPasteboardColor {
public:
    typedef std::tuple<const String&, const WebCore::Color&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SetPasteboardColor"); }
    static const bool isSync = true;

    using DelayedReply = SetPasteboardColorDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, int64_t changeCount);
    using Reply = std::tuple<int64_t&>;
    using ReplyArguments = std::tuple<int64_t>;
    SetPasteboardColor(const String& pasteboardName, const WebCore::Color& color)
        : m_arguments(pasteboardName, color)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SetPasteboardStringForType {
public:
    typedef std::tuple<const String&, const String&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SetPasteboardStringForType"); }
    static const bool isSync = true;

    using DelayedReply = SetPasteboardStringForTypeDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, int64_t changeCount);
    using Reply = std::tuple<int64_t&>;
    using ReplyArguments = std::tuple<int64_t>;
    SetPasteboardStringForType(const String& pasteboardName, const String& pasteboardType, const String& string)
        : m_arguments(pasteboardName, pasteboardType, string)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SetPasteboardBufferForType {
public:
    typedef std::tuple<const String&, const String&, const WebKit::SharedMemory::Handle&, uint64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SetPasteboardBufferForType"); }
    static const bool isSync = true;

    using DelayedReply = SetPasteboardBufferForTypeDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, int64_t changeCount);
    using Reply = std::tuple<int64_t&>;
    using ReplyArguments = std::tuple<int64_t>;
    SetPasteboardBufferForType(const String& pasteboardName, const String& pasteboardType, const WebKit::SharedMemory::Handle& handle, uint64_t size)
        : m_arguments(pasteboardName, pasteboardType, handle, size)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(GTK)
class WriteToClipboard {
public:
    typedef std::tuple<const String&, const WebKit::WebSelectionData&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("WriteToClipboard"); }
    static const bool isSync = false;

    WriteToClipboard(const String& pasteboardName, const WebKit::WebSelectionData& pasteboardContent)
        : m_arguments(pasteboardName, pasteboardContent)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(GTK)
class ReadFromClipboard {
public:
    typedef std::tuple<const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("ReadFromClipboard"); }
    static const bool isSync = true;

    using DelayedReply = ReadFromClipboardDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebKit::WebSelectionData& pasteboardContent);
    using Reply = std::tuple<WebKit::WebSelectionData&>;
    using ReplyArguments = std::tuple<WebKit::WebSelectionData>;
    explicit ReadFromClipboard(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if USE(LIBWPE)
class GetPasteboardTypes {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetPasteboardTypes"); }
    static const bool isSync = true;

    using DelayedReply = GetPasteboardTypesDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<String>& types);
    using Reply = std::tuple<Vector<String>&>;
    using ReplyArguments = std::tuple<Vector<String>>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if USE(LIBWPE)
class WriteWebContentToPasteboard {
public:
    typedef std::tuple<const WebCore::PasteboardWebContent&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("WriteWebContentToPasteboard"); }
    static const bool isSync = false;

    explicit WriteWebContentToPasteboard(const WebCore::PasteboardWebContent& content)
        : m_arguments(content)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if USE(LIBWPE)
class WriteStringToPasteboard {
public:
    typedef std::tuple<const String&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("WriteStringToPasteboard"); }
    static const bool isSync = false;

    WriteStringToPasteboard(const String& pasteboardType, const String& text)
        : m_arguments(pasteboardType, text)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

} // namespace WebPasteboardProxy
} // namespace Messages
