/*
 * Copyright (C) 2010-2019 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#include "WebProcess.h"

#include "ArgumentCoders.h"
#include "CacheModel.h"
#include "DataReference.h"
#include "Decoder.h"
#if ENABLE(GAMEPAD)
#include "GamepadData.h"
#endif
#include "HandleMessage.h"
#if (ENABLE(MEDIA_STREAM) && ENABLE(SANDBOX_EXTENSIONS))
#include "MediaDeviceSandboxExtensions.h"
#endif
#include "SandboxExtension.h"
#if ENABLE(SERVICE_WORKER)
#include "ServiceWorkerInitializationData.h"
#endif
#include "TextCheckerState.h"
#include "UserData.h"
#if PLATFORM(GTK) || PLATFORM(WPE)
#include "UserMessage.h"
#endif
#include "WebCoreArgumentCoders.h"
#include "WebPageCreationParameters.h"
#if ENABLE(SERVICE_WORKER)
#include "WebPageProxyIdentifier.h"
#endif
#if ENABLE(SERVICE_WORKER)
#include "WebPreferencesStore.h"
#endif
#include "WebProcessCreationParameters.h"
#include "WebProcessDataStoreParameters.h"
#include "WebProcessMessages.h"
#include "WebsiteData.h"
#include "WebsiteDataType.h"
#include <WebCore/BackForwardItemIdentifier.h>
#if ENABLE(MEDIA_STREAM)
#include <WebCore/MockMediaDevice.h>
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
#include <WebCore/NetworkStorageSession.h>
#endif
#include <WebCore/PageIdentifier.h>
#include <WebCore/PrewarmInformation.h>
#if ENABLE(RESOURCE_LOAD_STATISTICS) || ENABLE(SERVICE_WORKER)
#include <WebCore/RegistrableDomain.h>
#endif
#if PLATFORM(MAC)
#include <WebCore/ScreenProperties.h>
#endif
#include <WebCore/SecurityOriginData.h>
#include <wtf/HashMap.h>
#include <wtf/OptionSet.h>
#include <wtf/Vector.h>
#include <wtf/WallTime.h>
#include <wtf/text/WTFString.h>

namespace Messages {

namespace WebProcess {

void FetchWebsiteData::callReply(IPC::Decoder& decoder, CompletionHandler<void(WebKit::WebsiteData&&)>&& completionHandler)
{
    Optional<WebKit::WebsiteData> websiteData;
    decoder >> websiteData;
    if (!websiteData) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*websiteData));
}

void FetchWebsiteData::cancelReply(CompletionHandler<void(WebKit::WebsiteData&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<WebKit::WebsiteData>::create());
}

void FetchWebsiteData::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::WebsiteData& websiteData)
{
    *encoder << websiteData;
    connection.sendSyncReply(WTFMove(encoder));
}

void DeleteWebsiteData::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void DeleteWebsiteData::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void DeleteWebsiteData::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

void DeleteWebsiteDataForOrigins::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void DeleteWebsiteDataForOrigins::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void DeleteWebsiteDataForOrigins::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

void PrepareToSuspend::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void PrepareToSuspend::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void PrepareToSuspend::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#if ENABLE(SERVICE_WORKER)

void EstablishWorkerContextConnectionToNetworkProcess::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void EstablishWorkerContextConnectionToNetworkProcess::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void EstablishWorkerContextConnectionToNetworkProcess::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

void GetActivePagesOriginsForTesting::callReply(IPC::Decoder& decoder, CompletionHandler<void(Vector<String>&&)>&& completionHandler)
{
    Optional<Vector<String>> activeOrigins;
    decoder >> activeOrigins;
    if (!activeOrigins) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*activeOrigins));
}

void GetActivePagesOriginsForTesting::cancelReply(CompletionHandler<void(Vector<String>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Vector<String>>::create());
}

void GetActivePagesOriginsForTesting::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<String>& activeOrigins)
{
    *encoder << activeOrigins;
    connection.sendSyncReply(WTFMove(encoder));
}

void IsJITEnabled::callReply(IPC::Decoder& decoder, CompletionHandler<void(bool&&)>&& completionHandler)
{
    Optional<bool> enabled;
    decoder >> enabled;
    if (!enabled) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*enabled));
}

void IsJITEnabled::cancelReply(CompletionHandler<void(bool&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<bool>::create());
}

void IsJITEnabled::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool enabled)
{
    *encoder << enabled;
    connection.sendSyncReply(WTFMove(encoder));
}

void ClearCachedPage::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void ClearCachedPage::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void ClearCachedPage::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SeedResourceLoadStatisticsForTesting::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SeedResourceLoadStatisticsForTesting::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SeedResourceLoadStatisticsForTesting::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)

void SetShouldBlockThirdPartyCookiesForTesting::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetShouldBlockThirdPartyCookiesForTesting::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SetShouldBlockThirdPartyCookiesForTesting::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

} // namespace WebProcess

} // namespace Messages

namespace WebKit {

void WebProcess::didReceiveWebProcessMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    if (decoder.messageName() == Messages::WebProcess::InitializeWebProcess::name()) {
        IPC::handleMessage<Messages::WebProcess::InitializeWebProcess>(decoder, this, &WebProcess::initializeWebProcess);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::SetWebsiteDataStoreParameters::name()) {
        IPC::handleMessage<Messages::WebProcess::SetWebsiteDataStoreParameters>(decoder, this, &WebProcess::setWebsiteDataStoreParameters);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::CreateWebPage::name()) {
        IPC::handleMessage<Messages::WebProcess::CreateWebPage>(decoder, this, &WebProcess::createWebPage);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::PrewarmGlobally::name()) {
        IPC::handleMessage<Messages::WebProcess::PrewarmGlobally>(decoder, this, &WebProcess::prewarmGlobally);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::PrewarmWithDomainInformation::name()) {
        IPC::handleMessage<Messages::WebProcess::PrewarmWithDomainInformation>(decoder, this, &WebProcess::prewarmWithDomainInformation);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::SetCacheModel::name()) {
        IPC::handleMessage<Messages::WebProcess::SetCacheModel>(decoder, this, &WebProcess::setCacheModel);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::RegisterURLSchemeAsEmptyDocument::name()) {
        IPC::handleMessage<Messages::WebProcess::RegisterURLSchemeAsEmptyDocument>(decoder, this, &WebProcess::registerURLSchemeAsEmptyDocument);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::RegisterURLSchemeAsSecure::name()) {
        IPC::handleMessage<Messages::WebProcess::RegisterURLSchemeAsSecure>(decoder, this, &WebProcess::registerURLSchemeAsSecure);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::RegisterURLSchemeAsBypassingContentSecurityPolicy::name()) {
        IPC::handleMessage<Messages::WebProcess::RegisterURLSchemeAsBypassingContentSecurityPolicy>(decoder, this, &WebProcess::registerURLSchemeAsBypassingContentSecurityPolicy);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::SetDomainRelaxationForbiddenForURLScheme::name()) {
        IPC::handleMessage<Messages::WebProcess::SetDomainRelaxationForbiddenForURLScheme>(decoder, this, &WebProcess::setDomainRelaxationForbiddenForURLScheme);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::RegisterURLSchemeAsLocal::name()) {
        IPC::handleMessage<Messages::WebProcess::RegisterURLSchemeAsLocal>(decoder, this, &WebProcess::registerURLSchemeAsLocal);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::RegisterURLSchemeAsNoAccess::name()) {
        IPC::handleMessage<Messages::WebProcess::RegisterURLSchemeAsNoAccess>(decoder, this, &WebProcess::registerURLSchemeAsNoAccess);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::RegisterURLSchemeAsDisplayIsolated::name()) {
        IPC::handleMessage<Messages::WebProcess::RegisterURLSchemeAsDisplayIsolated>(decoder, this, &WebProcess::registerURLSchemeAsDisplayIsolated);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::RegisterURLSchemeAsCORSEnabled::name()) {
        IPC::handleMessage<Messages::WebProcess::RegisterURLSchemeAsCORSEnabled>(decoder, this, &WebProcess::registerURLSchemeAsCORSEnabled);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::RegisterURLSchemeAsCachePartitioned::name()) {
        IPC::handleMessage<Messages::WebProcess::RegisterURLSchemeAsCachePartitioned>(decoder, this, &WebProcess::registerURLSchemeAsCachePartitioned);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::RegisterURLSchemeAsCanDisplayOnlyIfCanRequest::name()) {
        IPC::handleMessage<Messages::WebProcess::RegisterURLSchemeAsCanDisplayOnlyIfCanRequest>(decoder, this, &WebProcess::registerURLSchemeAsCanDisplayOnlyIfCanRequest);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::SetDefaultRequestTimeoutInterval::name()) {
        IPC::handleMessage<Messages::WebProcess::SetDefaultRequestTimeoutInterval>(decoder, this, &WebProcess::setDefaultRequestTimeoutInterval);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::SetAlwaysUsesComplexTextCodePath::name()) {
        IPC::handleMessage<Messages::WebProcess::SetAlwaysUsesComplexTextCodePath>(decoder, this, &WebProcess::setAlwaysUsesComplexTextCodePath);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::SetShouldUseFontSmoothing::name()) {
        IPC::handleMessage<Messages::WebProcess::SetShouldUseFontSmoothing>(decoder, this, &WebProcess::setShouldUseFontSmoothing);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::SetResourceLoadStatisticsEnabled::name()) {
        IPC::handleMessage<Messages::WebProcess::SetResourceLoadStatisticsEnabled>(decoder, this, &WebProcess::setResourceLoadStatisticsEnabled);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::ClearResourceLoadStatistics::name()) {
        IPC::handleMessage<Messages::WebProcess::ClearResourceLoadStatistics>(decoder, this, &WebProcess::clearResourceLoadStatistics);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::UserPreferredLanguagesChanged::name()) {
        IPC::handleMessage<Messages::WebProcess::UserPreferredLanguagesChanged>(decoder, this, &WebProcess::userPreferredLanguagesChanged);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::FullKeyboardAccessModeChanged::name()) {
        IPC::handleMessage<Messages::WebProcess::FullKeyboardAccessModeChanged>(decoder, this, &WebProcess::fullKeyboardAccessModeChanged);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::DidAddPlugInAutoStartOriginHash::name()) {
        IPC::handleMessage<Messages::WebProcess::DidAddPlugInAutoStartOriginHash>(decoder, this, &WebProcess::didAddPlugInAutoStartOriginHash);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::ResetPlugInAutoStartOriginHashes::name()) {
        IPC::handleMessage<Messages::WebProcess::ResetPlugInAutoStartOriginHashes>(decoder, this, &WebProcess::resetPlugInAutoStartOriginHashes);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::SetPluginLoadClientPolicy::name()) {
        IPC::handleMessage<Messages::WebProcess::SetPluginLoadClientPolicy>(decoder, this, &WebProcess::setPluginLoadClientPolicy);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::ResetPluginLoadClientPolicies::name()) {
        IPC::handleMessage<Messages::WebProcess::ResetPluginLoadClientPolicies>(decoder, this, &WebProcess::resetPluginLoadClientPolicies);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::ClearPluginClientPolicies::name()) {
        IPC::handleMessage<Messages::WebProcess::ClearPluginClientPolicies>(decoder, this, &WebProcess::clearPluginClientPolicies);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::RefreshPlugins::name()) {
        IPC::handleMessage<Messages::WebProcess::RefreshPlugins>(decoder, this, &WebProcess::refreshPlugins);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::StartMemorySampler::name()) {
        IPC::handleMessage<Messages::WebProcess::StartMemorySampler>(decoder, this, &WebProcess::startMemorySampler);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::StopMemorySampler::name()) {
        IPC::handleMessage<Messages::WebProcess::StopMemorySampler>(decoder, this, &WebProcess::stopMemorySampler);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::SetTextCheckerState::name()) {
        IPC::handleMessage<Messages::WebProcess::SetTextCheckerState>(decoder, this, &WebProcess::setTextCheckerState);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::SetEnhancedAccessibility::name()) {
        IPC::handleMessage<Messages::WebProcess::SetEnhancedAccessibility>(decoder, this, &WebProcess::setEnhancedAccessibility);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::GetWebCoreStatistics::name()) {
        IPC::handleMessage<Messages::WebProcess::GetWebCoreStatistics>(decoder, this, &WebProcess::getWebCoreStatistics);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::GarbageCollectJavaScriptObjects::name()) {
        IPC::handleMessage<Messages::WebProcess::GarbageCollectJavaScriptObjects>(decoder, this, &WebProcess::garbageCollectJavaScriptObjects);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::SetJavaScriptGarbageCollectorTimerEnabled::name()) {
        IPC::handleMessage<Messages::WebProcess::SetJavaScriptGarbageCollectorTimerEnabled>(decoder, this, &WebProcess::setJavaScriptGarbageCollectorTimerEnabled);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::SetInjectedBundleParameter::name()) {
        IPC::handleMessage<Messages::WebProcess::SetInjectedBundleParameter>(decoder, this, &WebProcess::setInjectedBundleParameter);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::SetInjectedBundleParameters::name()) {
        IPC::handleMessage<Messages::WebProcess::SetInjectedBundleParameters>(decoder, this, &WebProcess::setInjectedBundleParameters);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::HandleInjectedBundleMessage::name()) {
        IPC::handleMessage<Messages::WebProcess::HandleInjectedBundleMessage>(decoder, this, &WebProcess::handleInjectedBundleMessage);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::FetchWebsiteData::name()) {
        IPC::handleMessageAsync<Messages::WebProcess::FetchWebsiteData>(connection, decoder, this, &WebProcess::fetchWebsiteData);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::DeleteWebsiteData::name()) {
        IPC::handleMessageAsync<Messages::WebProcess::DeleteWebsiteData>(connection, decoder, this, &WebProcess::deleteWebsiteData);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::DeleteWebsiteDataForOrigins::name()) {
        IPC::handleMessageAsync<Messages::WebProcess::DeleteWebsiteDataForOrigins>(connection, decoder, this, &WebProcess::deleteWebsiteDataForOrigins);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::SetHiddenPageDOMTimerThrottlingIncreaseLimit::name()) {
        IPC::handleMessage<Messages::WebProcess::SetHiddenPageDOMTimerThrottlingIncreaseLimit>(decoder, this, &WebProcess::setHiddenPageDOMTimerThrottlingIncreaseLimit);
        return;
    }
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebProcess::SetQOS::name()) {
        IPC::handleMessage<Messages::WebProcess::SetQOS>(decoder, this, &WebProcess::setQOS);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebProcess::SetMemoryCacheDisabled::name()) {
        IPC::handleMessage<Messages::WebProcess::SetMemoryCacheDisabled>(decoder, this, &WebProcess::setMemoryCacheDisabled);
        return;
    }
#if ENABLE(SERVICE_CONTROLS)
    if (decoder.messageName() == Messages::WebProcess::SetEnabledServices::name()) {
        IPC::handleMessage<Messages::WebProcess::SetEnabledServices>(decoder, this, &WebProcess::setEnabledServices);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebProcess::EnsureAutomationSessionProxy::name()) {
        IPC::handleMessage<Messages::WebProcess::EnsureAutomationSessionProxy>(decoder, this, &WebProcess::ensureAutomationSessionProxy);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::DestroyAutomationSessionProxy::name()) {
        IPC::handleMessage<Messages::WebProcess::DestroyAutomationSessionProxy>(decoder, this, &WebProcess::destroyAutomationSessionProxy);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::PrepareToSuspend::name()) {
        IPC::handleMessageAsync<Messages::WebProcess::PrepareToSuspend>(connection, decoder, this, &WebProcess::prepareToSuspend);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::ProcessDidResume::name()) {
        IPC::handleMessage<Messages::WebProcess::ProcessDidResume>(decoder, this, &WebProcess::processDidResume);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::MainThreadPing::name()) {
        IPC::handleMessage<Messages::WebProcess::MainThreadPing>(decoder, this, &WebProcess::mainThreadPing);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::BackgroundResponsivenessPing::name()) {
        IPC::handleMessage<Messages::WebProcess::BackgroundResponsivenessPing>(decoder, this, &WebProcess::backgroundResponsivenessPing);
        return;
    }
#if ENABLE(GAMEPAD)
    if (decoder.messageName() == Messages::WebProcess::SetInitialGamepads::name()) {
        IPC::handleMessage<Messages::WebProcess::SetInitialGamepads>(decoder, this, &WebProcess::setInitialGamepads);
        return;
    }
#endif
#if ENABLE(GAMEPAD)
    if (decoder.messageName() == Messages::WebProcess::GamepadConnected::name()) {
        IPC::handleMessage<Messages::WebProcess::GamepadConnected>(decoder, this, &WebProcess::gamepadConnected);
        return;
    }
#endif
#if ENABLE(GAMEPAD)
    if (decoder.messageName() == Messages::WebProcess::GamepadDisconnected::name()) {
        IPC::handleMessage<Messages::WebProcess::GamepadDisconnected>(decoder, this, &WebProcess::gamepadDisconnected);
        return;
    }
#endif
#if ENABLE(SERVICE_WORKER)
    if (decoder.messageName() == Messages::WebProcess::EstablishWorkerContextConnectionToNetworkProcess::name()) {
        IPC::handleMessageAsync<Messages::WebProcess::EstablishWorkerContextConnectionToNetworkProcess>(connection, decoder, this, &WebProcess::establishWorkerContextConnectionToNetworkProcess);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebProcess::SetHasSuspendedPageProxy::name()) {
        IPC::handleMessage<Messages::WebProcess::SetHasSuspendedPageProxy>(decoder, this, &WebProcess::setHasSuspendedPageProxy);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::SetIsInProcessCache::name()) {
        IPC::handleMessage<Messages::WebProcess::SetIsInProcessCache>(decoder, this, &WebProcess::setIsInProcessCache);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::MarkIsNoLongerPrewarmed::name()) {
        IPC::handleMessage<Messages::WebProcess::MarkIsNoLongerPrewarmed>(decoder, this, &WebProcess::markIsNoLongerPrewarmed);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::UpdateActivePages::name()) {
        IPC::handleMessage<Messages::WebProcess::UpdateActivePages>(decoder, this, &WebProcess::updateActivePages);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::GetActivePagesOriginsForTesting::name()) {
        IPC::handleMessageAsync<Messages::WebProcess::GetActivePagesOriginsForTesting>(connection, decoder, this, &WebProcess::getActivePagesOriginsForTesting);
        return;
    }
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebProcess::SetScreenProperties::name()) {
        IPC::handleMessage<Messages::WebProcess::SetScreenProperties>(decoder, this, &WebProcess::setScreenProperties);
        return;
    }
#endif
#if (PLATFORM(MAC) && ENABLE(WEBPROCESS_WINDOWSERVER_BLOCKING))
    if (decoder.messageName() == Messages::WebProcess::ScrollerStylePreferenceChanged::name()) {
        IPC::handleMessage<Messages::WebProcess::ScrollerStylePreferenceChanged>(decoder, this, &WebProcess::scrollerStylePreferenceChanged);
        return;
    }
#endif
#if (PLATFORM(MAC) && ENABLE(WEBPROCESS_WINDOWSERVER_BLOCKING))
    if (decoder.messageName() == Messages::WebProcess::DisplayConfigurationChanged::name()) {
        IPC::handleMessage<Messages::WebProcess::DisplayConfigurationChanged>(decoder, this, &WebProcess::displayConfigurationChanged);
        return;
    }
#endif
#if (PLATFORM(MAC) && ENABLE(WEBPROCESS_WINDOWSERVER_BLOCKING))
    if (decoder.messageName() == Messages::WebProcess::DisplayWasRefreshed::name()) {
        IPC::handleMessage<Messages::WebProcess::DisplayWasRefreshed>(decoder, this, &WebProcess::displayWasRefreshed);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY) && !PLATFORM(MACCATALYST)
    if (decoder.messageName() == Messages::WebProcess::BacklightLevelDidChange::name()) {
        IPC::handleMessage<Messages::WebProcess::BacklightLevelDidChange>(decoder, this, &WebProcess::backlightLevelDidChange);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebProcess::IsJITEnabled::name()) {
        IPC::handleMessageAsync<Messages::WebProcess::IsJITEnabled>(connection, decoder, this, &WebProcess::isJITEnabled);
        return;
    }
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebProcess::SetMediaMIMETypes::name()) {
        IPC::handleMessage<Messages::WebProcess::SetMediaMIMETypes>(decoder, this, &WebProcess::setMediaMIMETypes);
        return;
    }
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::WebProcess::AddMockMediaDevice::name()) {
        IPC::handleMessage<Messages::WebProcess::AddMockMediaDevice>(decoder, this, &WebProcess::addMockMediaDevice);
        return;
    }
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::WebProcess::ClearMockMediaDevices::name()) {
        IPC::handleMessage<Messages::WebProcess::ClearMockMediaDevices>(decoder, this, &WebProcess::clearMockMediaDevices);
        return;
    }
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::WebProcess::RemoveMockMediaDevice::name()) {
        IPC::handleMessage<Messages::WebProcess::RemoveMockMediaDevice>(decoder, this, &WebProcess::removeMockMediaDevice);
        return;
    }
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::WebProcess::ResetMockMediaDevices::name()) {
        IPC::handleMessage<Messages::WebProcess::ResetMockMediaDevices>(decoder, this, &WebProcess::resetMockMediaDevices);
        return;
    }
#endif
#if (ENABLE(MEDIA_STREAM) && ENABLE(SANDBOX_EXTENSIONS))
    if (decoder.messageName() == Messages::WebProcess::GrantUserMediaDeviceSandboxExtensions::name()) {
        IPC::handleMessage<Messages::WebProcess::GrantUserMediaDeviceSandboxExtensions>(decoder, this, &WebProcess::grantUserMediaDeviceSandboxExtensions);
        return;
    }
#endif
#if (ENABLE(MEDIA_STREAM) && ENABLE(SANDBOX_EXTENSIONS))
    if (decoder.messageName() == Messages::WebProcess::RevokeUserMediaDeviceSandboxExtensions::name()) {
        IPC::handleMessage<Messages::WebProcess::RevokeUserMediaDeviceSandboxExtensions>(decoder, this, &WebProcess::revokeUserMediaDeviceSandboxExtensions);
        return;
    }
#endif
    if (decoder.messageName() == Messages::WebProcess::ClearCurrentModifierStateForTesting::name()) {
        IPC::handleMessage<Messages::WebProcess::ClearCurrentModifierStateForTesting>(decoder, this, &WebProcess::clearCurrentModifierStateForTesting);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::SetBackForwardCacheCapacity::name()) {
        IPC::handleMessage<Messages::WebProcess::SetBackForwardCacheCapacity>(decoder, this, &WebProcess::setBackForwardCacheCapacity);
        return;
    }
    if (decoder.messageName() == Messages::WebProcess::ClearCachedPage::name()) {
        IPC::handleMessageAsync<Messages::WebProcess::ClearCachedPage>(connection, decoder, this, &WebProcess::clearCachedPage);
        return;
    }
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::WebProcess::UnblockAccessibilityServer::name()) {
        IPC::handleMessage<Messages::WebProcess::UnblockAccessibilityServer>(decoder, this, &WebProcess::unblockAccessibilityServer);
        return;
    }
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    if (decoder.messageName() == Messages::WebProcess::SendMessageToWebExtension::name()) {
        IPC::handleMessage<Messages::WebProcess::SendMessageToWebExtension>(decoder, this, &WebProcess::sendMessageToWebExtension);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::WebProcess::SeedResourceLoadStatisticsForTesting::name()) {
        IPC::handleMessageAsync<Messages::WebProcess::SeedResourceLoadStatisticsForTesting>(connection, decoder, this, &WebProcess::seedResourceLoadStatisticsForTesting);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::WebProcess::SetShouldBlockThirdPartyCookiesForTesting::name()) {
        IPC::handleMessageAsync<Messages::WebProcess::SetShouldBlockThirdPartyCookiesForTesting>(connection, decoder, this, &WebProcess::setShouldBlockThirdPartyCookiesForTesting);
        return;
    }
#endif
#if PLATFORM(IOS)
    if (decoder.messageName() == Messages::WebProcess::GrantAccessToAssetServices::name()) {
        IPC::handleMessage<Messages::WebProcess::GrantAccessToAssetServices>(decoder, this, &WebProcess::grantAccessToAssetServices);
        return;
    }
#endif
#if PLATFORM(IOS)
    if (decoder.messageName() == Messages::WebProcess::RevokeAccessToAssetServices::name()) {
        IPC::handleMessage<Messages::WebProcess::RevokeAccessToAssetServices>(decoder, this, &WebProcess::revokeAccessToAssetServices);
        return;
    }
#endif
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit

