/*
 * Copyright (C) 2010-2019 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ArgumentCoders.h"
#include "Connection.h"
#include "WebProcessProxyMessagesReplies.h"
#include <WebCore/FrameIdentifier.h>
#include <WebCore/PluginData.h>
#include <wtf/Forward.h>
#include <wtf/Optional.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace IPC {
class Attachment;
}

namespace WebCore {
class RegistrableDomain;
struct PrewarmInformation;
}

namespace WebKit {
struct BackForwardListItemState;
struct GPUProcessConnectionInfo;
struct NetworkProcessConnectionInfo;
struct UserMessage;
}

namespace Messages {
namespace WebProcessProxy {

static inline IPC::StringReference messageReceiverName()
{
    return IPC::StringReference("WebProcessProxy");
}

class UpdateBackForwardItem {
public:
    typedef std::tuple<const WebKit::BackForwardListItemState&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("UpdateBackForwardItem"); }
    static const bool isSync = false;

    explicit UpdateBackForwardItem(const WebKit::BackForwardListItemState& backForwardListItemState)
        : m_arguments(backForwardListItemState)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidDestroyFrame {
public:
    typedef std::tuple<const WebCore::FrameIdentifier&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("DidDestroyFrame"); }
    static const bool isSync = false;

    explicit DidDestroyFrame(const WebCore::FrameIdentifier& frameID)
        : m_arguments(frameID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidDestroyUserGestureToken {
public:
    typedef std::tuple<uint64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("DidDestroyUserGestureToken"); }
    static const bool isSync = false;

    explicit DidDestroyUserGestureToken(uint64_t userGestureTokenID)
        : m_arguments(userGestureTokenID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ShouldTerminate {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("ShouldTerminate"); }
    static const bool isSync = true;

    using DelayedReply = ShouldTerminateDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool shouldTerminate);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class EnableSuddenTermination {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("EnableSuddenTermination"); }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DisableSuddenTermination {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("DisableSuddenTermination"); }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if ENABLE(NETSCAPE_PLUGIN_API)
class GetPlugins {
public:
    typedef std::tuple<bool> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetPlugins"); }
    static const bool isSync = true;

    using DelayedReply = GetPluginsDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<WebCore::PluginInfo>& plugins, const Vector<WebCore::PluginInfo>& applicationPlugins, const Optional<Vector<WebCore::SupportedPluginIdentifier>>& supportedPluginIdentifiers);
    using Reply = std::tuple<Vector<WebCore::PluginInfo>&, Vector<WebCore::PluginInfo>&, Optional<Vector<WebCore::SupportedPluginIdentifier>>&>;
    using ReplyArguments = std::tuple<Vector<WebCore::PluginInfo>, Vector<WebCore::PluginInfo>, Optional<Vector<WebCore::SupportedPluginIdentifier>>>;
    explicit GetPlugins(bool refresh)
        : m_arguments(refresh)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(NETSCAPE_PLUGIN_API)
class GetPluginProcessConnection {
public:
    typedef std::tuple<uint64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetPluginProcessConnection"); }
    static const bool isSync = true;

    using DelayedReply = GetPluginProcessConnectionDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const IPC::Attachment& connectionHandle, bool supportsAsynchronousInitialization);
    using Reply = std::tuple<IPC::Attachment&, bool&>;
    using ReplyArguments = std::tuple<IPC::Attachment, bool>;
    explicit GetPluginProcessConnection(uint64_t pluginProcessToken)
        : m_arguments(pluginProcessToken)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class GetNetworkProcessConnection {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetNetworkProcessConnection"); }
    static const bool isSync = true;

    using DelayedReply = GetNetworkProcessConnectionDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebKit::NetworkProcessConnectionInfo& connectionInfo);
    using Reply = std::tuple<WebKit::NetworkProcessConnectionInfo&>;
    using ReplyArguments = std::tuple<WebKit::NetworkProcessConnectionInfo>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if ENABLE(GPU_PROCESS)
class GetGPUProcessConnection {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetGPUProcessConnection"); }
    static const bool isSync = true;

    using DelayedReply = GetGPUProcessConnectionDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebKit::GPUProcessConnectionInfo& connectionInfo);
    using Reply = std::tuple<WebKit::GPUProcessConnectionInfo&>;
    using ReplyArguments = std::tuple<WebKit::GPUProcessConnectionInfo>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class SetIsHoldingLockedFiles {
public:
    typedef std::tuple<bool> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SetIsHoldingLockedFiles"); }
    static const bool isSync = false;

    explicit SetIsHoldingLockedFiles(bool isHoldingLockedFiles)
        : m_arguments(isHoldingLockedFiles)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidExceedActiveMemoryLimit {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("DidExceedActiveMemoryLimit"); }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidExceedInactiveMemoryLimit {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("DidExceedInactiveMemoryLimit"); }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidExceedCPULimit {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("DidExceedCPULimit"); }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class StopResponsivenessTimer {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("StopResponsivenessTimer"); }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidReceiveMainThreadPing {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("DidReceiveMainThreadPing"); }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidReceiveBackgroundResponsivenessPing {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("DidReceiveBackgroundResponsivenessPing"); }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class MemoryPressureStatusChanged {
public:
    typedef std::tuple<bool> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("MemoryPressureStatusChanged"); }
    static const bool isSync = false;

    explicit MemoryPressureStatusChanged(bool isUnderMemoryPressure)
        : m_arguments(isUnderMemoryPressure)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidExceedInactiveMemoryLimitWhileActive {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("DidExceedInactiveMemoryLimitWhileActive"); }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidCollectPrewarmInformation {
public:
    typedef std::tuple<const WebCore::RegistrableDomain&, const WebCore::PrewarmInformation&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("DidCollectPrewarmInformation"); }
    static const bool isSync = false;

    DidCollectPrewarmInformation(const WebCore::RegistrableDomain& domain, const WebCore::PrewarmInformation& prewarmInformation)
        : m_arguments(domain, prewarmInformation)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if PLATFORM(COCOA)
class CacheMediaMIMETypes {
public:
    typedef std::tuple<const Vector<String>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("CacheMediaMIMETypes"); }
    static const bool isSync = false;

    explicit CacheMediaMIMETypes(const Vector<String>& types)
        : m_arguments(types)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(MAC)
class RequestHighPerformanceGPU {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RequestHighPerformanceGPU"); }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(MAC)
class ReleaseHighPerformanceGPU {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("ReleaseHighPerformanceGPU"); }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(MAC) && ENABLE(WEBPROCESS_WINDOWSERVER_BLOCKING)
class StartDisplayLink {
public:
    typedef std::tuple<const unsigned&, uint32_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("StartDisplayLink"); }
    static const bool isSync = false;

    StartDisplayLink(const unsigned& observerID, uint32_t displayID)
        : m_arguments(observerID, displayID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(MAC) && ENABLE(WEBPROCESS_WINDOWSERVER_BLOCKING)
class StopDisplayLink {
public:
    typedef std::tuple<const unsigned&, uint32_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("StopDisplayLink"); }
    static const bool isSync = false;

    StopDisplayLink(const unsigned& observerID, uint32_t displayID)
        : m_arguments(observerID, displayID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class ProcessWasResumed {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("ProcessWasResumed"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "ProcessWasResumedReply" }; }
    using AsyncReply = ProcessWasResumedAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class AddPlugInAutoStartOriginHash {
public:
    typedef std::tuple<const String&, uint32_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("AddPlugInAutoStartOriginHash"); }
    static const bool isSync = false;

    AddPlugInAutoStartOriginHash(const String& pageOrigin, uint32_t hash)
        : m_arguments(pageOrigin, hash)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class PlugInDidReceiveUserInteraction {
public:
    typedef std::tuple<uint32_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("PlugInDidReceiveUserInteraction"); }
    static const bool isSync = false;

    explicit PlugInDidReceiveUserInteraction(uint32_t hash)
        : m_arguments(hash)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if PLATFORM(GTK) || PLATFORM(WPE)
class SendMessageToWebContext {
public:
    typedef std::tuple<const WebKit::UserMessage&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SendMessageToWebContext"); }
    static const bool isSync = false;

    explicit SendMessageToWebContext(const WebKit::UserMessage& userMessage)
        : m_arguments(userMessage)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(GTK) || PLATFORM(WPE)
class SendMessageToWebContextWithReply {
public:
    typedef std::tuple<const WebKit::UserMessage&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SendMessageToWebContextWithReply"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebKit::UserMessage&&)>&&);
    static void cancelReply(CompletionHandler<void(WebKit::UserMessage&&)>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "SendMessageToWebContextWithReplyReply" }; }
    using AsyncReply = SendMessageToWebContextWithReplyAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebKit::UserMessage& replyMessage);
    using Reply = std::tuple<WebKit::UserMessage&>;
    using ReplyArguments = std::tuple<WebKit::UserMessage>;
    explicit SendMessageToWebContextWithReply(const WebKit::UserMessage& userMessage)
        : m_arguments(userMessage)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

} // namespace WebProcessProxy
} // namespace Messages
