/*
 * Copyright (C) 2010-2019 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#if ENABLE(SERVICE_WORKER)

#include "WebSWClientConnection.h"

#include "ArgumentCoders.h"
#include "Decoder.h"
#include "HandleMessage.h"
#include "SharedMemory.h"
#include "WebCoreArgumentCoders.h"
#include "WebSWClientConnectionMessages.h"
#include <WebCore/DocumentIdentifier.h>
#include <WebCore/ExceptionData.h>
#include <WebCore/FetchOptions.h>
#include <WebCore/MessageWithMessagePorts.h>
#include <WebCore/ServiceWorkerData.h>
#include <WebCore/ServiceWorkerIdentifier.h>
#include <WebCore/ServiceWorkerRegistrationData.h>
#include <WebCore/ServiceWorkerRegistrationKey.h>
#include <WebCore/ServiceWorkerTypes.h>
#include <WebCore/ServiceWorkerUpdateViaCache.h>
#include <wtf/HashSet.h>
#include <wtf/Optional.h>
#include <wtf/Vector.h>
#include <wtf/WallTime.h>
#include <wtf/text/WTFString.h>

namespace Messages {

namespace WebSWClientConnection {

void SetDocumentIsControlled::callReply(IPC::Decoder& decoder, CompletionHandler<void(bool&&)>&& completionHandler)
{
    Optional<bool> isSuccess;
    decoder >> isSuccess;
    if (!isSuccess) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*isSuccess));
}

void SetDocumentIsControlled::cancelReply(CompletionHandler<void(bool&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<bool>::create());
}

void SetDocumentIsControlled::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool isSuccess)
{
    *encoder << isSuccess;
    connection.sendSyncReply(WTFMove(encoder));
}

} // namespace WebSWClientConnection

} // namespace Messages

namespace WebKit {

void WebSWClientConnection::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    auto protectedThis = makeRef(*this);
    if (decoder.messageName() == Messages::WebSWClientConnection::JobRejectedInServer::name()) {
        IPC::handleMessage<Messages::WebSWClientConnection::JobRejectedInServer>(decoder, this, &WebSWClientConnection::jobRejectedInServer);
        return;
    }
    if (decoder.messageName() == Messages::WebSWClientConnection::RegistrationJobResolvedInServer::name()) {
        IPC::handleMessage<Messages::WebSWClientConnection::RegistrationJobResolvedInServer>(decoder, this, &WebSWClientConnection::registrationJobResolvedInServer);
        return;
    }
    if (decoder.messageName() == Messages::WebSWClientConnection::UnregistrationJobResolvedInServer::name()) {
        IPC::handleMessage<Messages::WebSWClientConnection::UnregistrationJobResolvedInServer>(decoder, this, &WebSWClientConnection::unregistrationJobResolvedInServer);
        return;
    }
    if (decoder.messageName() == Messages::WebSWClientConnection::StartScriptFetchForServer::name()) {
        IPC::handleMessage<Messages::WebSWClientConnection::StartScriptFetchForServer>(decoder, this, &WebSWClientConnection::startScriptFetchForServer);
        return;
    }
    if (decoder.messageName() == Messages::WebSWClientConnection::UpdateRegistrationState::name()) {
        IPC::handleMessage<Messages::WebSWClientConnection::UpdateRegistrationState>(decoder, this, &WebSWClientConnection::updateRegistrationState);
        return;
    }
    if (decoder.messageName() == Messages::WebSWClientConnection::UpdateWorkerState::name()) {
        IPC::handleMessage<Messages::WebSWClientConnection::UpdateWorkerState>(decoder, this, &WebSWClientConnection::updateWorkerState);
        return;
    }
    if (decoder.messageName() == Messages::WebSWClientConnection::FireUpdateFoundEvent::name()) {
        IPC::handleMessage<Messages::WebSWClientConnection::FireUpdateFoundEvent>(decoder, this, &WebSWClientConnection::fireUpdateFoundEvent);
        return;
    }
    if (decoder.messageName() == Messages::WebSWClientConnection::SetRegistrationLastUpdateTime::name()) {
        IPC::handleMessage<Messages::WebSWClientConnection::SetRegistrationLastUpdateTime>(decoder, this, &WebSWClientConnection::setRegistrationLastUpdateTime);
        return;
    }
    if (decoder.messageName() == Messages::WebSWClientConnection::SetRegistrationUpdateViaCache::name()) {
        IPC::handleMessage<Messages::WebSWClientConnection::SetRegistrationUpdateViaCache>(decoder, this, &WebSWClientConnection::setRegistrationUpdateViaCache);
        return;
    }
    if (decoder.messageName() == Messages::WebSWClientConnection::NotifyClientsOfControllerChange::name()) {
        IPC::handleMessage<Messages::WebSWClientConnection::NotifyClientsOfControllerChange>(decoder, this, &WebSWClientConnection::notifyClientsOfControllerChange);
        return;
    }
    if (decoder.messageName() == Messages::WebSWClientConnection::SetSWOriginTableIsImported::name()) {
        IPC::handleMessage<Messages::WebSWClientConnection::SetSWOriginTableIsImported>(decoder, this, &WebSWClientConnection::setSWOriginTableIsImported);
        return;
    }
    if (decoder.messageName() == Messages::WebSWClientConnection::SetSWOriginTableSharedMemory::name()) {
        IPC::handleMessage<Messages::WebSWClientConnection::SetSWOriginTableSharedMemory>(decoder, this, &WebSWClientConnection::setSWOriginTableSharedMemory);
        return;
    }
    if (decoder.messageName() == Messages::WebSWClientConnection::PostMessageToServiceWorkerClient::name()) {
        IPC::handleMessage<Messages::WebSWClientConnection::PostMessageToServiceWorkerClient>(decoder, this, &WebSWClientConnection::postMessageToServiceWorkerClient);
        return;
    }
    if (decoder.messageName() == Messages::WebSWClientConnection::DidMatchRegistration::name()) {
        IPC::handleMessage<Messages::WebSWClientConnection::DidMatchRegistration>(decoder, this, &WebSWClientConnection::didMatchRegistration);
        return;
    }
    if (decoder.messageName() == Messages::WebSWClientConnection::DidGetRegistrations::name()) {
        IPC::handleMessage<Messages::WebSWClientConnection::DidGetRegistrations>(decoder, this, &WebSWClientConnection::didGetRegistrations);
        return;
    }
    if (decoder.messageName() == Messages::WebSWClientConnection::RegistrationReady::name()) {
        IPC::handleMessage<Messages::WebSWClientConnection::RegistrationReady>(decoder, this, &WebSWClientConnection::registrationReady);
        return;
    }
    if (decoder.messageName() == Messages::WebSWClientConnection::SetDocumentIsControlled::name()) {
        IPC::handleMessageAsync<Messages::WebSWClientConnection::SetDocumentIsControlled>(connection, decoder, this, &WebSWClientConnection::setDocumentIsControlled);
        return;
    }
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit


#endif // ENABLE(SERVICE_WORKER)
