/*
 * Copyright (C) 2010-2019 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if ENABLE(SERVICE_WORKER)

#include "ArgumentCoders.h"
#include "Connection.h"
#include "SharedMemory.h"
#include "WebSWClientConnectionMessagesReplies.h"
#include <WebCore/DocumentIdentifier.h>
#include <WebCore/FetchOptions.h>
#include <WebCore/ServiceWorkerData.h>
#include <WebCore/ServiceWorkerIdentifier.h>
#include <WebCore/ServiceWorkerRegistrationData.h>
#include <WebCore/ServiceWorkerTypes.h>
#include <wtf/Forward.h>
#include <wtf/HashSet.h>
#include <wtf/Optional.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace WebCore {
class ServiceWorkerRegistrationKey;
enum class ShouldNotifyWhenResolved : bool;
enum class ServiceWorkerRegistrationState : uint8_t;
enum class ServiceWorkerState : uint8_t;
enum class ServiceWorkerUpdateViaCache : uint8_t;
struct ExceptionData;
struct MessageWithMessagePorts;
struct ServiceWorkerData;
struct ServiceWorkerRegistrationData;
}

namespace Messages {
namespace WebSWClientConnection {

static inline IPC::StringReference messageReceiverName()
{
    return IPC::StringReference("WebSWClientConnection");
}

class JobRejectedInServer {
public:
    typedef std::tuple<const WebCore::ServiceWorkerJobIdentifier&, const WebCore::ExceptionData&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("JobRejectedInServer"); }
    static const bool isSync = false;

    JobRejectedInServer(const WebCore::ServiceWorkerJobIdentifier& jobDataIdentifier, const WebCore::ExceptionData& exception)
        : m_arguments(jobDataIdentifier, exception)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RegistrationJobResolvedInServer {
public:
    typedef std::tuple<const WebCore::ServiceWorkerJobIdentifier&, const WebCore::ServiceWorkerRegistrationData&, WebCore::ShouldNotifyWhenResolved> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RegistrationJobResolvedInServer"); }
    static const bool isSync = false;

    RegistrationJobResolvedInServer(const WebCore::ServiceWorkerJobIdentifier& jobDataIdentifier, const WebCore::ServiceWorkerRegistrationData& registration, WebCore::ShouldNotifyWhenResolved shouldNotifyWhenResolved)
        : m_arguments(jobDataIdentifier, registration, shouldNotifyWhenResolved)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class UnregistrationJobResolvedInServer {
public:
    typedef std::tuple<const WebCore::ServiceWorkerJobIdentifier&, bool> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("UnregistrationJobResolvedInServer"); }
    static const bool isSync = false;

    UnregistrationJobResolvedInServer(const WebCore::ServiceWorkerJobIdentifier& jobDataIdentifier, bool unregistrationResult)
        : m_arguments(jobDataIdentifier, unregistrationResult)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class StartScriptFetchForServer {
public:
    typedef std::tuple<const WebCore::ServiceWorkerJobIdentifier&, const WebCore::ServiceWorkerRegistrationKey&, const WebCore::FetchOptions::Cache&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("StartScriptFetchForServer"); }
    static const bool isSync = false;

    StartScriptFetchForServer(const WebCore::ServiceWorkerJobIdentifier& jobDataIdentifier, const WebCore::ServiceWorkerRegistrationKey& registrationKey, const WebCore::FetchOptions::Cache& cachePolicy)
        : m_arguments(jobDataIdentifier, registrationKey, cachePolicy)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class UpdateRegistrationState {
public:
    typedef std::tuple<const WebCore::ServiceWorkerRegistrationIdentifier&, WebCore::ServiceWorkerRegistrationState, const Optional<WebCore::ServiceWorkerData>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("UpdateRegistrationState"); }
    static const bool isSync = false;

    UpdateRegistrationState(const WebCore::ServiceWorkerRegistrationIdentifier& identifier, WebCore::ServiceWorkerRegistrationState state, const Optional<WebCore::ServiceWorkerData>& serviceWorkerIdentifier)
        : m_arguments(identifier, state, serviceWorkerIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class UpdateWorkerState {
public:
    typedef std::tuple<const WebCore::ServiceWorkerIdentifier&, WebCore::ServiceWorkerState> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("UpdateWorkerState"); }
    static const bool isSync = false;

    UpdateWorkerState(const WebCore::ServiceWorkerIdentifier& serviceWorkerIdentifier, WebCore::ServiceWorkerState state)
        : m_arguments(serviceWorkerIdentifier, state)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class FireUpdateFoundEvent {
public:
    typedef std::tuple<const WebCore::ServiceWorkerRegistrationIdentifier&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("FireUpdateFoundEvent"); }
    static const bool isSync = false;

    explicit FireUpdateFoundEvent(const WebCore::ServiceWorkerRegistrationIdentifier& identifier)
        : m_arguments(identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetRegistrationLastUpdateTime {
public:
    typedef std::tuple<const WebCore::ServiceWorkerRegistrationIdentifier&, const WallTime&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SetRegistrationLastUpdateTime"); }
    static const bool isSync = false;

    SetRegistrationLastUpdateTime(const WebCore::ServiceWorkerRegistrationIdentifier& identifier, const WallTime& lastUpdateTime)
        : m_arguments(identifier, lastUpdateTime)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetRegistrationUpdateViaCache {
public:
    typedef std::tuple<const WebCore::ServiceWorkerRegistrationIdentifier&, WebCore::ServiceWorkerUpdateViaCache> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SetRegistrationUpdateViaCache"); }
    static const bool isSync = false;

    SetRegistrationUpdateViaCache(const WebCore::ServiceWorkerRegistrationIdentifier& identifier, WebCore::ServiceWorkerUpdateViaCache updateViaCache)
        : m_arguments(identifier, updateViaCache)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class NotifyClientsOfControllerChange {
public:
    typedef std::tuple<const HashSet<WebCore::DocumentIdentifier>&, const WebCore::ServiceWorkerData&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("NotifyClientsOfControllerChange"); }
    static const bool isSync = false;

    NotifyClientsOfControllerChange(const HashSet<WebCore::DocumentIdentifier>& contextIdentifiers, const WebCore::ServiceWorkerData& newController)
        : m_arguments(contextIdentifiers, newController)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetSWOriginTableIsImported {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SetSWOriginTableIsImported"); }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetSWOriginTableSharedMemory {
public:
    typedef std::tuple<const WebKit::SharedMemory::Handle&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SetSWOriginTableSharedMemory"); }
    static const bool isSync = false;

    explicit SetSWOriginTableSharedMemory(const WebKit::SharedMemory::Handle& handle)
        : m_arguments(handle)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class PostMessageToServiceWorkerClient {
public:
    typedef std::tuple<const WebCore::DocumentIdentifier&, const WebCore::MessageWithMessagePorts&, const WebCore::ServiceWorkerData&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("PostMessageToServiceWorkerClient"); }
    static const bool isSync = false;

    PostMessageToServiceWorkerClient(const WebCore::DocumentIdentifier& destinationContextIdentifier, const WebCore::MessageWithMessagePorts& message, const WebCore::ServiceWorkerData& source, const String& sourceOrigin)
        : m_arguments(destinationContextIdentifier, message, source, sourceOrigin)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidMatchRegistration {
public:
    typedef std::tuple<uint64_t, const Optional<WebCore::ServiceWorkerRegistrationData>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("DidMatchRegistration"); }
    static const bool isSync = false;

    DidMatchRegistration(uint64_t matchRequestIdentifier, const Optional<WebCore::ServiceWorkerRegistrationData>& data)
        : m_arguments(matchRequestIdentifier, data)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidGetRegistrations {
public:
    typedef std::tuple<uint64_t, const Vector<WebCore::ServiceWorkerRegistrationData>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("DidGetRegistrations"); }
    static const bool isSync = false;

    DidGetRegistrations(uint64_t matchRequestIdentifier, const Vector<WebCore::ServiceWorkerRegistrationData>& registrations)
        : m_arguments(matchRequestIdentifier, registrations)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RegistrationReady {
public:
    typedef std::tuple<uint64_t, const WebCore::ServiceWorkerRegistrationData&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RegistrationReady"); }
    static const bool isSync = false;

    RegistrationReady(uint64_t registrationReadyRequestIdentifier, const WebCore::ServiceWorkerRegistrationData& data)
        : m_arguments(registrationReadyRequestIdentifier, data)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetDocumentIsControlled {
public:
    typedef std::tuple<const WebCore::DocumentIdentifier&, const WebCore::ServiceWorkerRegistrationData&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SetDocumentIsControlled"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&)>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "SetDocumentIsControlledReply" }; }
    using AsyncReply = SetDocumentIsControlledAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool isSuccess);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    SetDocumentIsControlled(const WebCore::DocumentIdentifier& temporaryDocumentIdentifier, const WebCore::ServiceWorkerRegistrationData& data)
        : m_arguments(temporaryDocumentIdentifier, data)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

} // namespace WebSWClientConnection
} // namespace Messages

#endif // ENABLE(SERVICE_WORKER)
