/*
 * Copyright (C) 2010-2019 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#if ENABLE(SERVICE_WORKER)

#include "WebSWContextManagerConnection.h"

#include "ArgumentCoders.h"
#include "Decoder.h"
#include "FormDataReference.h"
#include "HandleMessage.h"
#include "WebCoreArgumentCoders.h"
#include "WebPreferencesStore.h"
#include "WebSWContextManagerConnectionMessages.h"
#include <WebCore/FetchIdentifier.h>
#include <WebCore/FetchOptions.h>
#include <WebCore/MessageWithMessagePorts.h>
#include <WebCore/ResourceRequest.h>
#include <WebCore/ServiceWorkerClientData.h>
#include <WebCore/ServiceWorkerContextData.h>
#include <WebCore/ServiceWorkerData.h>
#include <WebCore/ServiceWorkerIdentifier.h>
#include <WebCore/ServiceWorkerTypes.h>
#include <wtf/Optional.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace Messages {

namespace WebSWContextManagerConnection {

void SyncTerminateWorker::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

} // namespace WebSWContextManagerConnection

} // namespace Messages

namespace WebKit {

void WebSWContextManagerConnection::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    if (decoder.messageName() == Messages::WebSWContextManagerConnection::InstallServiceWorker::name()) {
        IPC::handleMessage<Messages::WebSWContextManagerConnection::InstallServiceWorker>(decoder, this, &WebSWContextManagerConnection::installServiceWorker);
        return;
    }
    if (decoder.messageName() == Messages::WebSWContextManagerConnection::StartFetch::name()) {
        IPC::handleMessage<Messages::WebSWContextManagerConnection::StartFetch>(decoder, this, &WebSWContextManagerConnection::startFetch);
        return;
    }
    if (decoder.messageName() == Messages::WebSWContextManagerConnection::CancelFetch::name()) {
        IPC::handleMessage<Messages::WebSWContextManagerConnection::CancelFetch>(decoder, this, &WebSWContextManagerConnection::cancelFetch);
        return;
    }
    if (decoder.messageName() == Messages::WebSWContextManagerConnection::ContinueDidReceiveFetchResponse::name()) {
        IPC::handleMessage<Messages::WebSWContextManagerConnection::ContinueDidReceiveFetchResponse>(decoder, this, &WebSWContextManagerConnection::continueDidReceiveFetchResponse);
        return;
    }
    if (decoder.messageName() == Messages::WebSWContextManagerConnection::PostMessageToServiceWorker::name()) {
        IPC::handleMessage<Messages::WebSWContextManagerConnection::PostMessageToServiceWorker>(decoder, this, &WebSWContextManagerConnection::postMessageToServiceWorker);
        return;
    }
    if (decoder.messageName() == Messages::WebSWContextManagerConnection::FireInstallEvent::name()) {
        IPC::handleMessage<Messages::WebSWContextManagerConnection::FireInstallEvent>(decoder, this, &WebSWContextManagerConnection::fireInstallEvent);
        return;
    }
    if (decoder.messageName() == Messages::WebSWContextManagerConnection::FireActivateEvent::name()) {
        IPC::handleMessage<Messages::WebSWContextManagerConnection::FireActivateEvent>(decoder, this, &WebSWContextManagerConnection::fireActivateEvent);
        return;
    }
    if (decoder.messageName() == Messages::WebSWContextManagerConnection::TerminateWorker::name()) {
        IPC::handleMessage<Messages::WebSWContextManagerConnection::TerminateWorker>(decoder, this, &WebSWContextManagerConnection::terminateWorker);
        return;
    }
    if (decoder.messageName() == Messages::WebSWContextManagerConnection::FindClientByIdentifierCompleted::name()) {
        IPC::handleMessage<Messages::WebSWContextManagerConnection::FindClientByIdentifierCompleted>(decoder, this, &WebSWContextManagerConnection::findClientByIdentifierCompleted);
        return;
    }
    if (decoder.messageName() == Messages::WebSWContextManagerConnection::MatchAllCompleted::name()) {
        IPC::handleMessage<Messages::WebSWContextManagerConnection::MatchAllCompleted>(decoder, this, &WebSWContextManagerConnection::matchAllCompleted);
        return;
    }
    if (decoder.messageName() == Messages::WebSWContextManagerConnection::ClaimCompleted::name()) {
        IPC::handleMessage<Messages::WebSWContextManagerConnection::ClaimCompleted>(decoder, this, &WebSWContextManagerConnection::claimCompleted);
        return;
    }
    if (decoder.messageName() == Messages::WebSWContextManagerConnection::SetUserAgent::name()) {
        IPC::handleMessage<Messages::WebSWContextManagerConnection::SetUserAgent>(decoder, this, &WebSWContextManagerConnection::setUserAgent);
        return;
    }
    if (decoder.messageName() == Messages::WebSWContextManagerConnection::UpdatePreferencesStore::name()) {
        IPC::handleMessage<Messages::WebSWContextManagerConnection::UpdatePreferencesStore>(decoder, this, &WebSWContextManagerConnection::updatePreferencesStore);
        return;
    }
    if (decoder.messageName() == Messages::WebSWContextManagerConnection::Close::name()) {
        IPC::handleMessage<Messages::WebSWContextManagerConnection::Close>(decoder, this, &WebSWContextManagerConnection::close);
        return;
    }
    if (decoder.messageName() == Messages::WebSWContextManagerConnection::SetThrottleState::name()) {
        IPC::handleMessage<Messages::WebSWContextManagerConnection::SetThrottleState>(decoder, this, &WebSWContextManagerConnection::setThrottleState);
        return;
    }
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED();
}

void WebSWContextManagerConnection::didReceiveSyncMessage(IPC::Connection& connection, IPC::Decoder& decoder, std::unique_ptr<IPC::Encoder>& replyEncoder)
{
    if (decoder.messageName() == Messages::WebSWContextManagerConnection::SyncTerminateWorker::name()) {
        IPC::handleMessageSynchronous<Messages::WebSWContextManagerConnection::SyncTerminateWorker>(connection, decoder, replyEncoder, this, &WebSWContextManagerConnection::syncTerminateWorker);
        return;
    }
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    UNUSED_PARAM(replyEncoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit


#endif // ENABLE(SERVICE_WORKER)
