/*
 * Copyright (C) 2010-2019 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if ENABLE(SERVICE_WORKER)

#include "ArgumentCoders.h"
#include "Connection.h"
#include "WebSWServerConnectionMessagesReplies.h"
#include <WebCore/ServiceWorkerClientIdentifier.h>
#include <WebCore/ServiceWorkerIdentifier.h>
#include <WebCore/ServiceWorkerTypes.h>
#include <wtf/Forward.h>
#include <wtf/Optional.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/text/WTFString.h>

namespace WebCore {
class ServiceWorkerRegistrationKey;
struct MessageWithMessagePorts;
struct SecurityOriginData;
struct ServiceWorkerClientData;
struct ServiceWorkerClientIdentifier;
struct ServiceWorkerFetchResult;
struct ServiceWorkerJobData;
}

namespace Messages {
namespace WebSWServerConnection {

static inline IPC::StringReference messageReceiverName()
{
    return IPC::StringReference("WebSWServerConnection");
}

class ScheduleJobInServer {
public:
    typedef std::tuple<const WebCore::ServiceWorkerJobData&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("ScheduleJobInServer"); }
    static const bool isSync = false;

    explicit ScheduleJobInServer(const WebCore::ServiceWorkerJobData& jobData)
        : m_arguments(jobData)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class FinishFetchingScriptInServer {
public:
    typedef std::tuple<const WebCore::ServiceWorkerFetchResult&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("FinishFetchingScriptInServer"); }
    static const bool isSync = false;

    explicit FinishFetchingScriptInServer(const WebCore::ServiceWorkerFetchResult& result)
        : m_arguments(result)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class AddServiceWorkerRegistrationInServer {
public:
    typedef std::tuple<const WebCore::ServiceWorkerRegistrationIdentifier&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("AddServiceWorkerRegistrationInServer"); }
    static const bool isSync = false;

    explicit AddServiceWorkerRegistrationInServer(const WebCore::ServiceWorkerRegistrationIdentifier& identifier)
        : m_arguments(identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RemoveServiceWorkerRegistrationInServer {
public:
    typedef std::tuple<const WebCore::ServiceWorkerRegistrationIdentifier&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RemoveServiceWorkerRegistrationInServer"); }
    static const bool isSync = false;

    explicit RemoveServiceWorkerRegistrationInServer(const WebCore::ServiceWorkerRegistrationIdentifier& identifier)
        : m_arguments(identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class PostMessageToServiceWorker {
public:
    typedef std::tuple<const WebCore::ServiceWorkerIdentifier&, const WebCore::MessageWithMessagePorts&, const WebCore::ServiceWorkerOrClientIdentifier&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("PostMessageToServiceWorker"); }
    static const bool isSync = false;

    PostMessageToServiceWorker(const WebCore::ServiceWorkerIdentifier& destination, const WebCore::MessageWithMessagePorts& message, const WebCore::ServiceWorkerOrClientIdentifier& source)
        : m_arguments(destination, message, source)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidResolveRegistrationPromise {
public:
    typedef std::tuple<const WebCore::ServiceWorkerRegistrationKey&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("DidResolveRegistrationPromise"); }
    static const bool isSync = false;

    explicit DidResolveRegistrationPromise(const WebCore::ServiceWorkerRegistrationKey& key)
        : m_arguments(key)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class MatchRegistration {
public:
    typedef std::tuple<uint64_t, const WebCore::SecurityOriginData&, const URL&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("MatchRegistration"); }
    static const bool isSync = false;

    MatchRegistration(uint64_t serviceRegistrationMatchRequestIdentifier, const WebCore::SecurityOriginData& topOrigin, const URL& clientURL)
        : m_arguments(serviceRegistrationMatchRequestIdentifier, topOrigin, clientURL)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class WhenRegistrationReady {
public:
    typedef std::tuple<uint64_t, const WebCore::SecurityOriginData&, const URL&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("WhenRegistrationReady"); }
    static const bool isSync = false;

    WhenRegistrationReady(uint64_t serviceRegistrationMatchRequestIdentifier, const WebCore::SecurityOriginData& topOrigin, const URL& clientURL)
        : m_arguments(serviceRegistrationMatchRequestIdentifier, topOrigin, clientURL)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetRegistrations {
public:
    typedef std::tuple<uint64_t, const WebCore::SecurityOriginData&, const URL&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetRegistrations"); }
    static const bool isSync = false;

    GetRegistrations(uint64_t serviceRegistrationMatchRequestIdentifier, const WebCore::SecurityOriginData& topOrigin, const URL& clientURL)
        : m_arguments(serviceRegistrationMatchRequestIdentifier, topOrigin, clientURL)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RegisterServiceWorkerClient {
public:
    typedef std::tuple<const WebCore::SecurityOriginData&, const WebCore::ServiceWorkerClientData&, const Optional<WebCore::ServiceWorkerRegistrationIdentifier>&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RegisterServiceWorkerClient"); }
    static const bool isSync = false;

    RegisterServiceWorkerClient(const WebCore::SecurityOriginData& topOrigin, const WebCore::ServiceWorkerClientData& data, const Optional<WebCore::ServiceWorkerRegistrationIdentifier>& controllingServiceWorkerRegistrationIdentifier, const String& userAgent)
        : m_arguments(topOrigin, data, controllingServiceWorkerRegistrationIdentifier, userAgent)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class UnregisterServiceWorkerClient {
public:
    typedef std::tuple<const WebCore::ServiceWorkerClientIdentifier&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("UnregisterServiceWorkerClient"); }
    static const bool isSync = false;

    explicit UnregisterServiceWorkerClient(const WebCore::ServiceWorkerClientIdentifier& identifier)
        : m_arguments(identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SyncTerminateWorkerFromClient {
public:
    typedef std::tuple<const WebCore::ServiceWorkerIdentifier&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SyncTerminateWorkerFromClient"); }
    static const bool isSync = true;

    using DelayedReply = SyncTerminateWorkerFromClientDelayedReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    explicit SyncTerminateWorkerFromClient(const WebCore::ServiceWorkerIdentifier& workerIdentifier)
        : m_arguments(workerIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetThrottleState {
public:
    typedef std::tuple<bool> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SetThrottleState"); }
    static const bool isSync = false;

    explicit SetThrottleState(bool isThrottleable)
        : m_arguments(isThrottleable)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class StoreRegistrationsOnDisk {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("StoreRegistrationsOnDisk"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "StoreRegistrationsOnDiskReply" }; }
    using AsyncReply = StoreRegistrationsOnDiskAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class IsServiceWorkerRunning {
public:
    typedef std::tuple<const WebCore::ServiceWorkerIdentifier&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("IsServiceWorkerRunning"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&)>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "IsServiceWorkerRunningReply" }; }
    using AsyncReply = IsServiceWorkerRunningAsyncReply;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool isRunning);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit IsServiceWorkerRunning(const WebCore::ServiceWorkerIdentifier& workerIdentifier)
        : m_arguments(workerIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

} // namespace WebSWServerConnection
} // namespace Messages

#endif // ENABLE(SERVICE_WORKER)
