/*
 * Copyright (C) 2010-2019 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#if ENABLE(SERVICE_WORKER)

#include "WebSWServerToContextConnection.h"

#include "ArgumentCoders.h"
#include "Decoder.h"
#include "HandleMessage.h"
#include "WebCoreArgumentCoders.h"
#include "WebSWServerToContextConnectionMessages.h"
#include <WebCore/MessageWithMessagePorts.h>
#include <WebCore/ServiceWorkerClientIdentifier.h>
#include <WebCore/ServiceWorkerClientQueryOptions.h>
#include <WebCore/ServiceWorkerIdentifier.h>
#include <WebCore/ServiceWorkerJobDataIdentifier.h>
#include <wtf/Optional.h>
#include <wtf/text/WTFString.h>

namespace Messages {

namespace WebSWServerToContextConnection {

void SkipWaiting::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SkipWaiting::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void SkipWaiting::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

} // namespace WebSWServerToContextConnection

} // namespace Messages

namespace WebKit {

void WebSWServerToContextConnection::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    if (decoder.messageName() == Messages::WebSWServerToContextConnection::ScriptContextFailedToStart::name()) {
        IPC::handleMessage<Messages::WebSWServerToContextConnection::ScriptContextFailedToStart>(decoder, this, &WebSWServerToContextConnection::scriptContextFailedToStart);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerToContextConnection::ScriptContextStarted::name()) {
        IPC::handleMessage<Messages::WebSWServerToContextConnection::ScriptContextStarted>(decoder, this, &WebSWServerToContextConnection::scriptContextStarted);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerToContextConnection::DidFinishInstall::name()) {
        IPC::handleMessage<Messages::WebSWServerToContextConnection::DidFinishInstall>(decoder, this, &WebSWServerToContextConnection::didFinishInstall);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerToContextConnection::DidFinishActivation::name()) {
        IPC::handleMessage<Messages::WebSWServerToContextConnection::DidFinishActivation>(decoder, this, &WebSWServerToContextConnection::didFinishActivation);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerToContextConnection::SetServiceWorkerHasPendingEvents::name()) {
        IPC::handleMessage<Messages::WebSWServerToContextConnection::SetServiceWorkerHasPendingEvents>(decoder, this, &WebSWServerToContextConnection::setServiceWorkerHasPendingEvents);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerToContextConnection::SkipWaiting::name()) {
        IPC::handleMessageAsync<Messages::WebSWServerToContextConnection::SkipWaiting>(connection, decoder, this, &WebSWServerToContextConnection::skipWaiting);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerToContextConnection::WorkerTerminated::name()) {
        IPC::handleMessage<Messages::WebSWServerToContextConnection::WorkerTerminated>(decoder, this, &WebSWServerToContextConnection::workerTerminated);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerToContextConnection::FindClientByIdentifier::name()) {
        IPC::handleMessage<Messages::WebSWServerToContextConnection::FindClientByIdentifier>(decoder, this, &WebSWServerToContextConnection::findClientByIdentifier);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerToContextConnection::MatchAll::name()) {
        IPC::handleMessage<Messages::WebSWServerToContextConnection::MatchAll>(decoder, this, &WebSWServerToContextConnection::matchAll);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerToContextConnection::Claim::name()) {
        IPC::handleMessage<Messages::WebSWServerToContextConnection::Claim>(decoder, this, &WebSWServerToContextConnection::claim);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerToContextConnection::SetScriptResource::name()) {
        IPC::handleMessage<Messages::WebSWServerToContextConnection::SetScriptResource>(decoder, this, &WebSWServerToContextConnection::setScriptResource);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerToContextConnection::PostMessageToServiceWorkerClient::name()) {
        IPC::handleMessage<Messages::WebSWServerToContextConnection::PostMessageToServiceWorkerClient>(decoder, this, &WebSWServerToContextConnection::postMessageToServiceWorkerClient);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerToContextConnection::DidFailHeartBeatCheck::name()) {
        IPC::handleMessage<Messages::WebSWServerToContextConnection::DidFailHeartBeatCheck>(decoder, this, &WebSWServerToContextConnection::didFailHeartBeatCheck);
        return;
    }
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit


#endif // ENABLE(SERVICE_WORKER)
