/*
 * Copyright (C) 2010-2019 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ArgumentCoders.h"
#include "Connection.h"
#include "ContentWorldShared.h"
#include "WebCompiledContentRuleListData.h"
#include "WebUserContentControllerDataTypes.h"
#include "WebUserContentControllerMessagesReplies.h"
#include <utility>
#include <wtf/Forward.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace WebKit {
enum class InjectUserScriptImmediately : bool;
}

namespace Messages {
namespace WebUserContentController {

static inline IPC::StringReference messageReceiverName()
{
    return IPC::StringReference("WebUserContentController");
}

class AddUserContentWorlds {
public:
    typedef std::tuple<const Vector<std::pair<WebKit::ContentWorldIdentifier, String>>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("AddUserContentWorlds"); }
    static const bool isSync = false;

    explicit AddUserContentWorlds(const Vector<std::pair<WebKit::ContentWorldIdentifier, String>>& worlds)
        : m_arguments(worlds)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RemoveUserContentWorlds {
public:
    typedef std::tuple<const Vector<WebKit::ContentWorldIdentifier>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RemoveUserContentWorlds"); }
    static const bool isSync = false;

    explicit RemoveUserContentWorlds(const Vector<WebKit::ContentWorldIdentifier>& worldIdentifiers)
        : m_arguments(worldIdentifiers)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class AddUserScripts {
public:
    typedef std::tuple<const Vector<struct WebKit::WebUserScriptData>&, WebKit::InjectUserScriptImmediately> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("AddUserScripts"); }
    static const bool isSync = false;

    AddUserScripts(const Vector<struct WebKit::WebUserScriptData>& userScripts, WebKit::InjectUserScriptImmediately immediately)
        : m_arguments(userScripts, immediately)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RemoveUserScript {
public:
    typedef std::tuple<const WebKit::ContentWorldIdentifier&, uint64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RemoveUserScript"); }
    static const bool isSync = false;

    RemoveUserScript(const WebKit::ContentWorldIdentifier& worldIdentifier, uint64_t identifier)
        : m_arguments(worldIdentifier, identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RemoveAllUserScripts {
public:
    typedef std::tuple<const Vector<WebKit::ContentWorldIdentifier>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RemoveAllUserScripts"); }
    static const bool isSync = false;

    explicit RemoveAllUserScripts(const Vector<WebKit::ContentWorldIdentifier>& worldIdentifiers)
        : m_arguments(worldIdentifiers)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class AddUserStyleSheets {
public:
    typedef std::tuple<const Vector<struct WebKit::WebUserStyleSheetData>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("AddUserStyleSheets"); }
    static const bool isSync = false;

    explicit AddUserStyleSheets(const Vector<struct WebKit::WebUserStyleSheetData>& userStyleSheets)
        : m_arguments(userStyleSheets)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RemoveUserStyleSheet {
public:
    typedef std::tuple<const WebKit::ContentWorldIdentifier&, uint64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RemoveUserStyleSheet"); }
    static const bool isSync = false;

    RemoveUserStyleSheet(const WebKit::ContentWorldIdentifier& worldIdentifier, uint64_t identifier)
        : m_arguments(worldIdentifier, identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RemoveAllUserStyleSheets {
public:
    typedef std::tuple<const Vector<WebKit::ContentWorldIdentifier>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RemoveAllUserStyleSheets"); }
    static const bool isSync = false;

    explicit RemoveAllUserStyleSheets(const Vector<WebKit::ContentWorldIdentifier>& worldIdentifiers)
        : m_arguments(worldIdentifiers)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class AddUserScriptMessageHandlers {
public:
    typedef std::tuple<const Vector<struct WebKit::WebScriptMessageHandlerData>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("AddUserScriptMessageHandlers"); }
    static const bool isSync = false;

    explicit AddUserScriptMessageHandlers(const Vector<struct WebKit::WebScriptMessageHandlerData>& scriptMessageHandlers)
        : m_arguments(scriptMessageHandlers)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RemoveUserScriptMessageHandler {
public:
    typedef std::tuple<const WebKit::ContentWorldIdentifier&, uint64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RemoveUserScriptMessageHandler"); }
    static const bool isSync = false;

    RemoveUserScriptMessageHandler(const WebKit::ContentWorldIdentifier& worldIdentifier, uint64_t identifier)
        : m_arguments(worldIdentifier, identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RemoveAllUserScriptMessageHandlers {
public:
    typedef std::tuple<const Vector<WebKit::ContentWorldIdentifier>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RemoveAllUserScriptMessageHandlers"); }
    static const bool isSync = false;

    explicit RemoveAllUserScriptMessageHandlers(const Vector<WebKit::ContentWorldIdentifier>& worldIdentifiers)
        : m_arguments(worldIdentifiers)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if ENABLE(CONTENT_EXTENSIONS)
class AddContentRuleLists {
public:
    typedef std::tuple<const Vector<std::pair<String, WebKit::WebCompiledContentRuleListData>>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("AddContentRuleLists"); }
    static const bool isSync = false;

    explicit AddContentRuleLists(const Vector<std::pair<String, WebKit::WebCompiledContentRuleListData>>& contentFilters)
        : m_arguments(contentFilters)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(CONTENT_EXTENSIONS)
class RemoveContentRuleList {
public:
    typedef std::tuple<const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RemoveContentRuleList"); }
    static const bool isSync = false;

    explicit RemoveContentRuleList(const String& name)
        : m_arguments(name)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(CONTENT_EXTENSIONS)
class RemoveAllContentRuleLists {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RemoveAllContentRuleLists"); }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

} // namespace WebUserContentController
} // namespace Messages
